/*
 * $Id: cia1.c,v 2.1 1996/07/29 20:44:09 ettore Exp ettore $
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README for copyright notice
 *
 *
 * Written by
 *   Vesa-Matti Puro  (vmp@lut.fi)
 *   Jarkko Sonninen  (sonninen@lut.fi)
 *   Jouko Valta      (jopi@stekt.oulu.fi)
 *   Andre Fachat     (a.fachat@physik.tu-chemnitz.de)
 *
 * Fixes by
 *   Ettore Perazzoli (ettore@comm2000.it)
 *
 * 
 * $Log: cia1.c,v $
 * Revision 2.1  1996/07/29 20:44:09  ettore
 * Precise timing for int_cia*: update_cia* can calculate
 * the new IRQ alarm according to the delay between the
 * alarm and the int_* call.
 * Timer A and B are now updated correctly.
 * Some other misc TOD and timer fixes.
 *
 * Revision 2.0  1996/04/18 17:06:39  fachat
 * Completely rewrote the timer code. Now it's using the new [un]setalrm
 * and setirq calls from interrupt.c. Also moved the VIC20 via stuff
 * to own files.
 *
 * Revision 1.11  1996/04/01 06:00:00  ettore
 * Bug fixes: timer & interrupts.
 *
 * Revision 1.10  1995/11/07  16:51:00  jopi
 * *** empty log message ***
 *
 * Revision 1.9  1995/06/28  19:40:56  jopi
 * VIA 6522 emulation
 * Control Bit Masks
 * Variable keyboard matrix sizes.
 *
 * Revision 1.8  1994/12/12  16:59:44  jopi
 * Defines for registers
 * C128 serial (slow mode) loop-back
 *
 * Revision 1.7  1994/06/16  17:19:26  jopi
 * Code reorganized and cleaned up
 *
 * Revision 1.6  1994/02/18  16:10:57  jopi
 * New debugger & arg parser and x128
 *
 * Revision 1.5  1994/01/26  16:08:37  jopi
 * X64 version 0.2 PL 1
 *
 * Revision 1.4  1993/11/10  01:55:34  jopi
 * reu, asm and disk directory fixed
 * REL_ADDR macro and 1541 made more portable
 *
 * Revision 1.3  93/06/21  13:38:13  jopi
 *  X64 version 0.2 PL 0
 * 
 * Revision 1.2  1993/06/13  08:19:36  sonninen
 * *** empty log message ***
 *
 *
 */

#include <stdio.h>
#include <time.h>
#include <string.h>

#include "config.h"
#include "vmachine.h"
#include "extern.h"
#include "cia.h"
#include "cpu.h"

#include "interrupt.h"

/* #define CIA1_TIMER_DEBUG */

/* global */

BYTE    cia1[16];
BYTE    keyarr[KBD_ROWS];

/* local functions */

static void update_cia1ta ( int, int );
static void update_cia1tb ( int, int );
static void check_tod1alarm ( void );

/*
 * Local variables
 */

static int   cia1int;	/* Interrupt Flag register for cia 1 */

static int cia1ta;	/* value of cia 1 timer A at last update */
static int cia1tb;	/* value of cia 1 timer B at last update */

static CLOCK cia1tau;	/* time when cia 1 timer A is updated */
static CLOCK cia1tbu;	/* time when cia 1 timer B is updated */

static int cia1ta_stopped = 1;	/* timer stopped */
static int cia1tb_stopped = 1;	/* timer stopped */

static int cia1sr_bits;	/* number of bits still to send * 2 */

static BYTE tod1alarm[4];
static BYTE tod1latch[4];
static char tod1stopped;
static char tod1latched;
static int tod1ticks = 100000;          /* approx. a 1/10 sec. */

#if 0
static char   tod1virt = 0;	/* Virtual or Real-Time clock */

static char   tod1mode = 0;	/* Stopped */
static char   tod1freq;		/* 50/60 Hz select for cia 1 TOD */
static long   tod1cycles = 0;	/* TOD Virtual Mode displacemant */
static time_t tod1rel;		/* TOD 1 difference to system clock */
static time_t tod1alarm;	/* TOD 1 alarm time */
#endif

/* ------------------------------------------------------------------------- */
/* CIA 1 */


void    reset_cia1(void)
{
	int i;
	for(i=0;i<16;i++) cia1[i] = 0;
	cia1ta_stopped = 0;
	cia1tb_stopped = 0;
	cia1ta = 0x10000;
	cia1tb = 0x10000;
	cia1tau = clk;
	cia1tbu = clk;
	cia1sr_bits = 0;

        memset(tod1alarm, 0, sizeof(tod1alarm));
        tod1latched = 0;
        tod1stopped = 0;
        setalarm(I_CIA1TOD, tod1ticks);
}


void    store_cia1(ADDRESS addr, BYTE byte)
{
    addr &= 0xf;
    
    if (debugflg)
	printf("store cia1[%x] %x\n", (int) addr, (int) byte);

    switch (addr) {

      case CIA_PRA: /* port A */
      case CIA_PRB: /* port B */
	cia1[addr] = byte;
	break;

	/*
	 * TOD clock is stopped by writing Hours, and restarted
	 * upon writing Tenths of Seconds.
	 *
	 * REAL:  TOD register + (wallclock - tod1rel)
	 * VIRT:  TOD register + (cycles - begin)/cycles_per_sec
	 */
      case CIA_TOD_TEN: /* Time Of Day clock 1/10 s */
      case CIA_TOD_HR:  /* Time Of Day clock hour */
      case CIA_TOD_SEC: /* Time Of Day clock sec */
      case CIA_TOD_MIN: /* Time Of Day clock min */
        if(cia1[CIA_CRB] & 0x80) tod1alarm[addr-CIA_TOD_TEN]=byte;
        else {
          if(addr == CIA_TOD_TEN) tod1stopped = 0;
          if(addr == CIA_TOD_HR) tod1stopped = 1;
          cia1[addr] = byte;
        }
	check_tod1alarm();
        break;

#if 0
      case CIA_TOD_TEN: /* Time Of Day clock 1/10 s */
	if (cia1[CIA_CRB] & 0x80) {  /* Set alarm */
	    tod1alarm = time(NULL);
	} else {
	    tod1mode = 1;  /* Run */
	    time(&tod1rel); /* seconds, use hrtime for better resolution */
	    tod1cycles = clk;

	    if (debugflg)
		printf("cia1 TOD set. time: ------  offset: %ld.\n", tod1rel);
	}
	cia1[addr] = byte;
	break;

      case CIA_TOD_HR:  /* Time Of Day clock hour */
	if (!(cia1[CIA_CRB] & 0x80))  /* Set TOD Hours */
	    tod1mode = 0;  /* Stop */

      case CIA_TOD_SEC: /* Time Of Day clock sec */
      case CIA_TOD_MIN: /* Time Of Day clock min */

	if (cia1[CIA_CRB] & 0x80) {  /* Set alarm */
	} else { 
	    cia1[addr] = byte;
	}
	break;
#endif

      case CIA_SDR: /* Serial Port output buffer */
	cia1[addr] = byte;
	if((cia1[CIA_CRA] & 0x40) && (cia1sr_bits < 16)) cia1sr_bits += 16;
	break;

	/* Interrupts */

      case CIA_ICR: /* Interrupt Control Register */
#if defined (CIA1_TIMER_DEBUG)
        printf ("Cia#1 set CIA_ICR: 0x%x\n", byte);
#endif
	if(byte & CIA_IM_SET) {
	  cia1[addr] |= byte;
	} else {
	  cia1[addr] &= ~byte;
	}
	setirq(I_CIA1TA, (cia1int & cia1[CIA_ICR] & CIA_IM_TA) ? 1 : 0 );
	setirq(I_CIA1TB, (cia1int & cia1[CIA_ICR] & CIA_IM_TB) ? 1 : 0 );
	setirq(I_CIA1SP, (cia1int & cia1[CIA_ICR] & CIA_IM_SP) ? 1 : 0 );
	setirq(I_CIA1TOD,(cia1int & cia1[CIA_ICR] & CIA_IM_TOD)? 1 : 0 );
	break;

	/* Control */

      case CIA_CRA: /* control register A */
#if defined (CIA1_TIMER_DEBUG)
        printf ("Cia#1 set CIA_CRA: 0x%x\n", byte);
#endif

	/* bit 7 tod frequency */
#if 0
	tod1freq = (byte & 0x80) ? 50 : 60;
#endif
	/* bit 6 serial port mode */

	/* bit 4 force load */
	if (byte & 0x10) {
	    update_cia1ta(1, 0);
	}
	/* bit 3 timer run mode */
	/* bit 2 & 1 timer output to PB6 */

	/* bit 0 start/stop timer */
	/* bit 5 timer count mode */
	if ((byte & 0x20) || !(byte & 1)) {
            /* stop timer */
#if defined (CIA1_TIMER_DEBUG)
	    printf ("Cia#1 Timer A stopped\n");
#endif
	    if(!cia1ta_stopped) {
	    	update_cia1ta(0, 0);	/* set cia1ta to actual value */
		cia1ta_stopped = 1;
		unsetalarm(I_CIA1TA);
	    }
	} else {
#if defined (CIA1_TIMER_DEBUG)
	    printf ("Cia#1 Timer A restarted, cia1ta=%d\n", cia1ta);
#endif
	    if(cia1ta_stopped) {
		cia1tau = clk;
		cia1ta_stopped = 0;
                update_cia1ta(1, 0);       /* EP 07/20/96 is this right? */
		setalarm(I_CIA1TA, cia1ta);
	    }
	}

	/* I haven't found anything about this in my specs, but obviously
	   not doing this breaks some programs... */
	if(!(byte&1)) {
	  setirq(I_CIA1TA,0);
	  cia1int &= ~CIA_IM_TA; 
	}

	cia1[addr] = byte & 0xef;	/* remove strobe */
	break;

      case CIA_CRB: /* control register B */
#if defined (CIA1_TIMER_DEBUG)
        printf ("Cia#1 set CIA_CRB: 0x%x\n", byte);
#endif
	/* bit 7 set alarm/tod clock */
	/* bit 4 force load */
	if (byte & 0x10) {
	    update_cia1tb(1, 0);
	}
	/* bit 3 timer run mode */
	/* bit 2 & 1 timer output to PB6 */
	/* bit 0 start/stop timer */
	/* bit 6, 5 timer count mode */
	if ((byte & 0x20) || !(byte & 1)) {
	    if(!cia1tb_stopped) {
#if defined (CIA1_TIMER_DEBUG)
 	    	printf ("Cia#1 Timer B stopped\n");
#endif
		update_cia1tb(0, 0);	/* set cia1tb to actual value */
		cia1tb_stopped = 1;
		unsetalarm(I_CIA1TB);
	    }
	} else {
	    if(cia1tb_stopped) {
#if defined (CIA1_TIMER_DEBUG)
	    	printf ("Cia#1 Timer B restarted\n");
#endif
		cia1tbu = clk;
		cia1tb_stopped = 0;
                update_cia1tb(1, 0);      /* EP 07/20/96 is this right? */
		setalarm(I_CIA1TB, cia1tb);
	    }
	}
        /* I haven't found anything about this in my specs, but obviously
           not doing this breaks some programs... */
	if(!(byte&1)) {
	  setirq(I_CIA1TB,0);
	  cia1int &= ~CIA_IM_TB; 
	}
	cia1[addr] = byte & 0xef;	/* remove strobe */
	break;
      default:
	cia1[addr] = byte;

    }  /* switch */
}


/* ------------------------------------------------------------------------- */

BYTE    read_cia1(ADDRESS addr)
{
    addr &= 0xf;

    if (debugflg)
	printf("read cia1[%x]\n", addr);

    switch (addr) {

      case CIA_PRA: /* port A */
	{
	    int     t, i;
	    BYTE    j, v = 1;
	    t = ~(cia1[CIA_PRB] & cia1[CIA_DDRB]);
#ifdef C128
	    t |= ((read_vic(0x2f) & 7) >> 8);
#endif
	    for (j = i = 0; i < KBD_ROWS; v <<= 1, i++)
		if (t & keyarr[i])
		    j |= v;
	    j = ~j;

	    if (debugflg) {
		printf("read port A %d\n", j);
		printf("a: %x b:%x  ca: %x cb: %x joy: %x\n",
		       (int) j, (int) cia1[CIA_PRB],
		       (int) cia1[CIA_PRA], (int) cia1[CIA_PRB], joy[2]);
	    }
	    return (((j & ~cia1[CIA_DDRA]) | (cia1[CIA_PRA] & cia1[CIA_DDRA])) & ~joy[2]);
	}

      case CIA_PRB: /* port B */
	{
	    int     t, i;
	    BYTE    j;
	    t = cia1[CIA_PRA] & cia1[CIA_DDRA];  /* intr() updates keyboard */

	    for (j = i = 0; i < KBD_ROWS; t >>= 1, i++)
		if (!(t & 1))
		    j |= keyarr[i];
	    j = ~j;

	    if (debugflg) {
		printf("read port B %d\n", j);
		printf("a: %x b:%x  ca: %x cb: %x joy: %x\n",
		       (int) cia1[CIA_PRA], (int) j,
		       (int) cia1[CIA_DDRA], (int) cia1[CIA_DDRB], joy[1]);
	    }
	    return (((j & ~cia1[CIA_DDRB]) | (cia1[CIA_PRB] & cia1[CIA_DDRB])) & ~joy[1]);
	}


	/* Timers */

      case CIA_TAL: /* timer A low */
	update_cia1ta(0, 0);
	return (cia1ta & 0xff);

      case CIA_TAH: /* timer A high */
	update_cia1ta(0, 0);
	return ((cia1ta >> 8) & 0xff);

      case CIA_TBL: /* timer B low */
	update_cia1tb(0, 0);
	return (cia1tb & 0xff);

      case CIA_TBH: /* timer B high */
	update_cia1tb(0, 0);
	return ((cia1tb >> 8) & 0xff);

	/*
	 * TOD clock is latched by reading Hours, and released
	 * upon reading Tenths of Seconds. The counter itself
	 * keeps ticking all the time.
	 * Also note that this latching is different from the input one.
	 */
      case CIA_TOD_TEN: /* Time Of Day clock 1/10 s */
      case CIA_TOD_SEC: /* Time Of Day clock sec */
      case CIA_TOD_MIN: /* Time Of Day clock min */
      case CIA_TOD_HR:  /* Time Of Day clock hour */
        if(!tod1latched) memcpy(tod1latch, cia1+CIA_TOD_TEN, sizeof(tod1latch));
        if(addr == CIA_TOD_TEN) tod1latched = 0;
        if(addr == CIA_TOD_HR) tod1latched = 1;
        return cia1[addr];

#if 0
      case CIA_TOD_TEN: /* Time Of Day clock 1/10 s */
	if (tod1mode == 1)  /* Running */
	    update_tod();
	    tod1mode = 1;  /* Run */
	return (cia1[addr]);

      case CIA_TOD_SEC: /* Time Of Day clock sec */
      case CIA_TOD_MIN: /* Time Of Day clock min */
	if (tod1mode == 1)  /* Running */
	    update_tod();
	return (cia1[addr]);

      case CIA_TOD_HR:  /* Time Of Day clock hour */
	if (tod1mode == 1)  /* Running */
	    update_tod();

	    tod1mode = 2;  /* Latched */
	return (cia1[addr]);
#endif

      case CIA_SDR: /* Serial Port Shift Register */
	return (cia1[addr]);

	/* Interrupts */

      case CIA_ICR: /* Interrupt Flag Register */
	{
	    BYTE    t = cia1int /*| 8*/;   /****  This is ugly trick for x128 ****/
	    if ( t & cia1[CIA_ICR] ) t |= 0x80; 
if(debugflg) 
	printf("read intfl gives cia1int=%02x -> %02x @ PC = %04x, sr_bits=%d\n",
					cia1int, t, PC, cia1sr_bits);
	    cia1int = 0;
	    setirq(I_CIA1TA, 0);
	    setirq(I_CIA1TB, 0);
	    setirq(I_CIA1SP, 0);
	    setirq(I_CIA1TOD,0);
	    return (t);
	}
    }  /* switch */

    return (cia1[addr]);
}


/* ------------------------------------------------------------------------- */

int    int_cia1ta(int offset)
{
    if (debugflg)
	printf("cia1 timer A event\n");

    if (cia1[CIA_CRA] & 8) {	/* one-shot mode */
#if defined (CIA1_TIMER_DEBUG)
        if(debugflg) printf ("Cia#1 Timer A event -- one-shot mode: next int won't happen\n");
#endif
	cia1ta = 0x10000;
	cia1ta_stopped = 1;
	unsetalarm(I_CIA1TA);
	cia1[CIA_CRA] &= ~1;
    } else {		/* continuous mode */
	update_cia1ta(1, offset);
    }
    if(cia1[CIA_CRA] & 0x40) {
      if(cia1sr_bits) {
	if(!(--cia1sr_bits)) {
	  cia1int |= CIA_IM_SP;
          if(cia1[CIA_ICR] & CIA_IM_SP) setirq(I_CIA1SP, 1);
	}
      }
    }
    cia1int |= CIA_IM_TA;
    if(cia1[CIA_ICR] & CIA_IM_TA) setirq(I_CIA1TA, 1);

    /* count timer A underflow pulses */
    if ((cia1[CIA_CRB] & 0x61) == 0x41)
	if (--cia1tb <= 0)
	    setalarm(I_CIA1TB, 0);	/*int_clk[I_CIA1TB] = clk;*/

    return 0; /*(cia1int & cia1[CIA_ICR] & 1) ? 1:0;*/	/* strobe irq */
}


/*
 * Timer B can run in 2 (4) modes
 * cia1[f] & 0x60 == 0x00   count system 02 pulses
 * cia1[f] & 0x60 == 0x40   count timer A underflows
 * cia1[f] & 0x60 == 0x20 | 0x60 count CNT pulses => counter stops
 */


int    int_cia1tb(int offset)
{
    if (debugflg)
	printf("cia1 timer B event\n");

    if (cia1[CIA_CRB] & 8) {	/* one-shot mode */
#if defined (CIA1_TIMER_DEBUG)
        printf ("Cia#1 Timer B event -- one-shot mode: next int won't happen\n");
#endif
	cia1tb = 0x10000;
	cia1tb_stopped = 1;
	unsetalarm(I_CIA1TB);
	cia1[CIA_CRB] &= ~1;	/* CRB run bit */
    } else {		/* continuous mode */
	update_cia1tb(1, offset);
    }
    cia1int |= 2;
    if(cia1[CIA_ICR] & 2) setirq(I_CIA1TB, 1);

    return 0; 
}

int int_cia1tod(void) {
        register int t, pm;

if (debugflg)
  printf("CIA1: TOD timer event (1/10 sec tick), tod=%02x:%02x,%02x.%x\n",
                cia1[CIA_TOD_HR], cia1[CIA_TOD_MIN], cia1[CIA_TOD_SEC],
                cia1[CIA_TOD_TEN]);

        /* set up new int */
        setalarm(I_CIA1TOD, tod1ticks);

        if(!tod1stopped) {
          /* inc timer */
          t = bcd2byte(cia1[CIA_TOD_TEN]);
          t++;
          cia1[CIA_TOD_TEN] = byte2bcd(t%10);
          if( t>= 10 ) {
            t = bcd2byte(cia1[CIA_TOD_SEC]);
            t++;
            cia1[CIA_TOD_SEC] = byte2bcd(t%60);
            if(t >= 60) {
              t = bcd2byte(cia1[CIA_TOD_MIN]);
              t++;
              cia1[CIA_TOD_MIN] = byte2bcd(t%60);
              if(t>= 60) {
                pm = cia1[CIA_TOD_HR] & 0x80;
                t = bcd2byte(cia1[CIA_TOD_HR]&0x1f);
                if(!t) pm ^= 0x80;      /* toggle am/pm on 0:59->1:00 hr */
                t++;
                t = t%12 | pm;
                cia1[CIA_TOD_HR] = byte2bcd(t);
              }
            }
          }
if (debugflg)
  printf("CIA1: TOD after event :tod=%02x:%02x,%02x.%x\n",
                cia1[CIA_TOD_HR], cia1[CIA_TOD_MIN], cia1[CIA_TOD_SEC],
                cia1[CIA_TOD_TEN]);
          /* check alarm */
	  check_tod1alarm();
#if 0
          if(!memcmp(tod1alarm, cia1+CIA_TOD_TEN, sizeof(tod1alarm))) {
            cia1int |= CIA_IM_TOD;
            if(cia1[CIA_ICR] & CIA_IM_TOD) setirq(I_CIA1TOD, 1);
          }
#endif
        }
        return 0;
}

static void check_tod1alarm(void) {
          if(!memcmp(tod1alarm, cia1+CIA_TOD_TEN, sizeof(tod1alarm))) {
            cia1int |= CIA_IM_TOD;
            if(cia1[CIA_ICR] & CIA_IM_TOD) setirq(I_CIA1TOD, 1);
          }
}

/* ------------------------------------------------------------------------- */

static void update_cia1ta(int force, int offset)
{
/*printf("update_cia1ta(force=%d, stopped=%d)\n",force, cia1ta_stopped);*/
    if(force) {
	cia1ta = cia1[CIA_TAL] + (cia1[CIA_TAH] << 8) - (offset & 0xffff);
	if(cia1ta <= 0) cia1ta += 0x10000;
	if(!cia1ta_stopped)
	    setalarm(I_CIA1TA, cia1ta);
    } else 
    if(!cia1ta_stopped) {
	cia1ta = (int)(cia1ta - clk + cia1tau) & 0xffff;
	if(cia1ta <= 0)
	    cia1ta += 0x10000;
    }
    cia1tau = clk;
}


static void update_cia1tb(int force, int offset)
{
    if (force) {
	cia1tb = cia1[CIA_TBL] + (cia1[CIA_TBH] << 8) - (offset & 0xffff);
	if(cia1tb <= 0) cia1tb += 0x10000;
	if(!cia1tb_stopped) 
	    setalarm(I_CIA1TB, cia1tb);
    } else 
    if (!cia1tb_stopped) {
	cia1tb = (int)(cia1tb - clk + cia1tbu) & 0xffff;
	if(cia1tb <= 0)
	    cia1tb += 0x10000;
    }
    cia1tbu = clk;
}

#if 0
static void update_tod(void)
{
    long   c;
    time_t t = 0;

    /*if (tod1mode != 1) return; */  /* Not Running */

    if (tod1virt) {
	c = clk - tod1cycles;
	t = (double)c / (tod1freq ? 9852.48 : 10227.30); /* CYCLES / 0.01 SEC */
    }
    else {
	time(&t);
	t -= tod1rel;
    }

    cia1[CIA_TOD_TEN] = 0;			/* Time Of Day clock 1/10 s */
    cia1[CIA_TOD_SEC] = (BYTE)(t/3600);		/* Time Of Day clock sec */
    cia1[CIA_TOD_MIN] = (BYTE)((t%3600)/60);	/* Time Of Day clock min */
    cia1[CIA_TOD_HR]  = (BYTE)(t%60);
}
#endif

int     show_keyarr(void)
{
    int     i, j;
    for (j = 0; j < KBD_ROWS; j++) {
	printf("%d:", j);
	for (i = 0x80; i; i >>= 1)
	    printf(" %d", keyarr[j] & i);
	printf("\n");
    }
    return (0);
}
