/*
 * $Id: cia2.c,v 2.1 1996/07/29 20:45:28 ettore Exp ettore $
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README for copyright notice
 *
 *
 * Written by
 *   Vesa-Matti Puro  (vmp@lut.fi)
 *   Jarkko Sonninen  (sonninen@lut.fi)
 *   Jouko Valta      (jopi@stekt.oulu.fi)
 *   Andre Fachat     (a.fachat@physik.tu-chemnitz.de)
 *
 * Fixes by
 *   Ettore Perazzoli (ettore@comm2000.it)
 *
 *
 * $Log: cia2.c,v $
 * Revision 2.1  1996/07/29 20:45:28  ettore
 * Precise timing for int_cia*: update_cia* can calculate
 * the new IRQ alarm according to the delay between the
 * alarm and the int* call.
 * Timer A and B are now updated correctly.
 * Some other misc TOD and timer fixes.
 *
 * Revision 2.0  1996/04/18 17:04:12  fachat
 * Completely rewrote the timer code. Now it's using the new [un]setalrm
 * and setirq calls from interrupt.c. Also moved the VIC20 via stuff
 * to own files.
 *
 * Revision 1.9  1996/04/01 06:02:00  ettore
 * Bug fixes: timer & interrupts.
 *
 * Revision 1.8  1995/11/07  16:51:00  jopi
 * *** empty log message ***
 *
 * Revision 1.7  1995/06/28  19:40:56  jopi
 * VIA 6522 emulation
 * Control Bit Masks
 *
 * Revision 1.6  1994/12/12  16:59:44  jopi
 * New VIC control
 *
 * Revision 1.5  1994/06/16  17:19:26  jopi
 * Code reorganized and cleaned up
 *
 * Revision 1.4  1993/11/10  01:55:34  jopi
 * reu, asm and disk directory fixed
 * REL_ADDR macro and 1541 made more portable
 *
 * Revision 1.3  93/06/21  13:38:16  jopi
 *  X64 version 0.2 PL 0
 * 
 * Revision 1.2  1993/06/13  08:19:36  sonninen
 * *** empty log message ***
 *
 *
 */

#include <stdio.h>
#include <time.h>
#include <string.h>

#include "config.h"
#include "vmachine.h"
#include "extern.h"
#include "cia.h"
#include "interrupt.h"

/* #define CIA2_TIMER_DEBUG */

/* global */

BYTE    cia2[16];

/* local functions */

static void update_cia2ta ( int, int );
static void update_cia2tb ( int, int );
static void check_tod2alarm ( void );

/*
 * Local variables
 */

static int cia2int;	/* Interrupt Flag register for cia 2 */

static int cia2ta;	/* value of cia 2 timer A at last update */
static int cia2tb;	/* value of cia 2 timer B at last update */

static CLOCK cia2tau;	/* time when cia 2 timer A is updated */
static CLOCK cia2tbu;	/* time when cia 2 timer B is updated */

static int cia2ta_stopped = 1;  /* timer stopped */
static int cia2tb_stopped = 1;  /* timer stopped */

static int cia2sr_bits;		/* number of bits still to send * 2 */

static BYTE tod2alarm[4];
static BYTE tod2latch[4];
static char tod2stopped;
static char tod2latched;
static int tod2ticks = 100000;          /* approx. a 1/10 sec. */

#if 0
static char   tod2virt = 0;	/* Virtual or Real-Time clock */

static char   tod2mode = 0;	/* Stopped */
static char   tod2freq;		/* 50/60 Hz select for cia 2 TOD */
static long   tod2cycles = 0;	/* TOD Virtual Mode displacemant */
static time_t tod2rel;		/* TOD 2 difference to system clock */
static time_t tod2alarm;	/* TOD 2 alarm time */
#endif

extern int vbank;	/* VIC bank */


/* ------------------------------------------------------------------------- */
/* CIA 2 */


void    reset_cia2(void)
{
	int i;
	for(i=0;i<16;i++) cia2[i] = 0;
        cia2ta_stopped = 0;
        cia2tb_stopped = 0;
        cia2ta = 0x10000;
        cia2tb = 0x10000;
        cia2tau = clk;
        cia2tbu = clk;
	cia2sr_bits = 0;

        memset(tod2alarm, 0, sizeof(tod2alarm));
        tod2latched = 0;
        tod2stopped = 0;
        setalarm(I_CIA2TOD, tod2ticks);
}

void    store_cia2(ADDRESS addr, BYTE byte)
{
    static BYTE cia2pa = 0;

    addr &= 0xf;
    
    if (debugflg)
	printf("store cia2[%x] %x\n", (int) addr, (int) byte);
    switch (addr) {

      case CIA_PRA: /* port A */
      case CIA_DDRA: /* data dir reg A */
	{
	    BYTE tmp;
	    cia2[addr] = byte;
	    tmp = cia2[CIA_PRA] | ~cia2[CIA_DDRA];
	    if(tmp != cia2pa) {
	      cia2pa = tmp;
	      vbank = ((cia2pa /*cia2[CIA_PRA]*/ & 3) ^ 3) << 14;
	      screen_init();
	    }
	}
	break;

      case CIA_PRB: /* port B */
	cia2[addr] = byte;
	break;

	/*
	 * TOD clock is stopped by writing Hours, and restarted
	 * upon writing Tenths of Seconds.
	 */
      case CIA_TOD_TEN: /* Time Of Day clock 1/10 s */
      case CIA_TOD_HR:  /* Time Of Day clock hour */
      case CIA_TOD_SEC: /* Time Of Day clock sec */
      case CIA_TOD_MIN: /* Time Of Day clock min */
        if(cia2[CIA_CRB] & 0x80) tod2alarm[addr-CIA_TOD_TEN]=byte;
        else {
          if(addr == CIA_TOD_TEN) tod2stopped = 0;
          if(addr == CIA_TOD_HR) tod2stopped = 1;
          cia2[addr] = byte;
        }
	check_tod2alarm();
        break;

#if 0

      case CIA_TOD_TEN: /* Time Of Day clock 1/10 s */
	if (cia2[CIA_CRB] & 0x80) {  /* Set alarm */
	    tod2alarm = (int)time(NULL);
	} else {
	    tod2mode = 1;  /* Run */
	    time(&tod2rel); /* seconds, use hrtime for better resolution */
	    tod2cycles = clk;

	    if (debugflg)
		printf("cia2 TOD set. time: ------  offset: %ld.\n", tod2rel);
	}
	cia2[addr] = byte;
	break;

      case CIA_TOD_HR:  /* Time Of Day clock hour */
	if (!(cia2[CIA_CRB] & 0x80))  /* Set TOD Hours */
	    tod2mode = 0;  /* Stop */

      case CIA_TOD_SEC: /* Time Of Day clock sec */
      case CIA_TOD_MIN: /* Time Of Day clock min */

	if (cia2[CIA_CRB] & 0x80) {  /* Set alarm */
	} else
	    cia2[addr] = byte;
	break;
#endif

      case CIA_SDR: /* Serial Port output buffer */
	cia2[addr] = byte;
        if((cia2[CIA_CRA] & 0x40) && (cia2sr_bits < 16)) cia2sr_bits += 16;
	break;


	/* Interrupts */

      case CIA_ICR: /* Interrupt Control Register */
#if defined (CIA2_TIMER_DEBUG)
        if(debugflg) printf ("Cia#2 set CIA_ICR: 0x%x\n", byte);
#endif
        if(byte & CIA_IM_SET) {
          cia2[addr] |= byte;
        } else {
          cia2[addr] &= ~byte;
        }
        setirq(I_CIA2TA, (cia2int & cia1[CIA_ICR] & CIA_IM_TA) ? 2 : 0 );
        setirq(I_CIA2TB, (cia2int & cia1[CIA_ICR] & CIA_IM_TB) ? 2 : 0 );
        setirq(I_CIA2SP, (cia2int & cia1[CIA_ICR] & CIA_IM_SP) ? 2 : 0 );
        setirq(I_CIA2TOD,(cia2int & cia1[CIA_ICR] & CIA_IM_TOD)? 2 : 0 );
        break;

	/* Control */

      case CIA_CRA: /* control register A */
#if defined (CIA2_TIMER_DEBUG)
        if(debugflg) printf ("Cia#2 set CIA_CRA: 0x%x\n", byte);
#endif

	/* bit 7 tod frequency */
#if 0
	tod2freq = (byte & 0x80) ? 50 : 60;
#endif

	/* bit 6 serial port mode */

	/* bit 4 force load */
	if (byte & 0x10) {
	    update_cia2ta(1, 0);
	}
	/* bit 3 timer run mode */
	/* bit 2 & 1 timer output to PB6 */

	/* bit 0 start/stop timer */
	/* bit 5 timer count mode */
	if ((byte & 0x20) || !(byte & 1)) {
	    /* stop timer */
#if defined (CIA2_TIMER_DEBUG)
	    if(debugflg) printf ("Cia#2 Timer A stopped\n");
#endif
            if(!cia2ta_stopped) {
                update_cia2ta(0, 0);       /* set cia2ta to actual value */
                cia2ta_stopped = 1;
                unsetalarm(I_CIA2TA);
            }
	} else {
#if defined (CIA2_TIMER_DEBUG)
	    if(debugflg) printf ("Cia#2 Timer A restarted\n");
#endif
            if(cia2ta_stopped) {
                cia2tau = clk;
                cia2ta_stopped = 0;
                update_cia2ta(1, 0);       /* EP 07/20/96 is this right? */
                setalarm(I_CIA2TA, cia2ta);
            }
	}

        /* I haven't found anything about this in my specs, but obviously
           not doing this breaks some programs... */
        if(!(byte&1)) {
	  setirq(I_CIA2TA,0);
          cia2int &= ~CIA_IM_TA;
	}

	cia2[addr] = byte & 0xef;	/* remove strobe */
	break;

      case CIA_CRB: /* control register B */
#if defined (CIA2_TIMER_DEBUG)
        if(debugflg) printf ("Cia#2 set CIA_CRB: 0x%x\n", byte);
#endif
	/* bit 7 set alarm/tod clock */
	/* bit 4 force load */
	if (byte & 0x10) {
	    update_cia2tb(1, 0);
	}
	/* bit 3 timer run mode */
	/* bit 2 & 1 timer output to PB6 */
	/* bit 0 start/stop timer */
	/* bit 6, 5 timer count mode */
	if ((byte & 0x20) || !(byte & 1)) {
            if(!cia2tb_stopped) {
#if defined (CIA2_TIMER_DEBUG)
                if(debugflg) printf ("Cia#2 Timer B stopped\n");
#endif
                update_cia2tb(0, 0);       /* set cia2tb to actual value */
                cia2tb_stopped = 1;
                unsetalarm(I_CIA2TB);
            }
	} else {
            if(cia2tb_stopped) {
#if defined (CIA2_TIMER_DEBUG)
                if(debugflg) printf ("Cia#2 Timer B restarted\n");
#endif
                cia2tbu = clk;
                cia2tb_stopped = 0;
		update_cia2tb(1, 0);       /* EP 07/20/96 is this right? */
                setalarm(I_CIA2TB, cia2tb);
            }
	}
        /* I haven't found anything about this in my specs, but obviously
           not doing this breaks some programs... */
        if(!(byte&1)) {
	  setirq(I_CIA2TB,0);
          cia2int &= ~CIA_IM_TB;
	}
	cia2[addr] = byte & 0xef;	/* remove strobe */
	break;

      default:
	cia2[addr] = byte;

    }  /* switch */
}


/* ------------------------------------------------------------------------- */

BYTE    read_cia2(ADDRESS addr)
{
    addr &= 0xf;
   
    if(debugflg) printf("read cia2[%x]\n",addr);
 
    switch (addr) {

      case CIA_PRA: /* port A */
	return ((cia2[CIA_PRA] & cia2[CIA_DDRA]) | (0x7f & ~cia2[CIA_DDRA]));

      case CIA_PRB: /* port B */
	return ((cia2[CIA_PRB] & cia2[CIA_DDRB]) | (0xff & ~cia2[CIA_DDRB]));


	/* Timers */

      case CIA_TAL: /* timer A low */
	update_cia2ta(0, 0);
	return (cia2ta & 0xff);

      case CIA_TAH: /* timer A high */
	update_cia2ta(0, 0);
	return ((cia2ta >> 8) & 0xff);

      case CIA_TBL: /* timer B low */
	update_cia2tb(0, 0);
	return (cia2tb & 0xff);

      case CIA_TBH: /* timer B high */
	update_cia2tb(0, 0);
	return ((cia2tb >> 8) & 0xff);

	/*
	 * TOD clock is latched by reading Hours, and released
	 * upon reading Tenths of Seconds. The counter itself
	 * keeps ticking all the time.
	 * Also note that this laching is different from the input one.
	 */
      case CIA_TOD_TEN: /* Time Of Day clock 1/10 s */
      case CIA_TOD_SEC: /* Time Of Day clock sec */
      case CIA_TOD_MIN: /* Time Of Day clock min */
      case CIA_TOD_HR:  /* Time Of Day clock hour */
        if(!tod2latched) memcpy(tod2latch, cia1+CIA_TOD_TEN, sizeof(tod2latch));
        if(addr == CIA_TOD_TEN) tod2latched = 0;
        if(addr == CIA_TOD_HR) tod2latched = 1;
        return cia1[addr];

#if 0

      case CIA_TOD_TEN: /* Time Of Day clock 1/10 s */
	if (tod2mode == 1)  /* Running */
	    update_tod();
	    tod2mode = 1;  /* Run */
	return (cia2[addr]);

      case CIA_TOD_SEC: /* Time Of Day clock sec */
      case CIA_TOD_MIN: /* Time Of Day clock min */
	if (tod2mode == 1)  /* Running */
	    update_tod();
	return (cia2[addr]);

      case CIA_TOD_HR:  /* Time Of Day clock hour */
	if (tod2mode == 1)  /* Running */
	    update_tod();

	    tod2mode = 2;  /* Latched */
	return (cia2[addr]);

      case CIA_SDR: /* Serial Port Shift Register */
	return (cia2[addr]);
#endif

	/* Interrupts */

      case CIA_ICR: /* Interrupt Flag Register */
	{
            BYTE    t = cia2int /*| 8*/;   /****  This is ugly trick for x128 ****/
            if ( t & cia2[CIA_ICR] ) t |= 0x80;
/*printf("read intfl gives cia2int=%02x -> %02x @ PC = %04x\n",cia2int, t, PC);*/
            cia2int = 0;
            setirq(I_CIA2TA, 0);
            setirq(I_CIA2TB, 0);
            setirq(I_CIA2SP, 0);
            setirq(I_CIA2TOD,0);
            return (t);
	}
    }  /* switch */

    return (cia2[addr]);
}


/* ------------------------------------------------------------------------- */

int    int_cia2ta(int offset)
{
     if (debugflg)
        printf("cia2 timer A event\n");

    if (cia2[CIA_CRA] & 8) {    /* one-shot mode */
#if defined (CIA2_TIMER_DEBUG)
        if(debugflg) printf ("Cia#2 Timer A event -- one-shot mode:"
						" next int won't happen!\n");
#endif
        cia2ta = 0x10000;
        cia2ta_stopped = 1;
        unsetalarm(I_CIA2TA);
	cia2[CIA_CRA] &= ~1;
    } else {            /* continuous mode */
        update_cia2ta(1, offset);
    }
    if(cia2[CIA_CRA] & 0x40) {
      if(cia2sr_bits) {
	if(!(--cia2sr_bits)) {
          cia2int |= CIA_IM_SP;
          if(cia2[CIA_ICR] & CIA_IM_SP) setirq(I_CIA2SP, 2);
	}
      }
    }
    cia2int |= CIA_IM_TA;
    if(cia2[CIA_ICR] & CIA_IM_TA) setirq(I_CIA2TA, 2);

    /* count timer A underflow pulses */
    if ((cia2[CIA_CRB] & 0x61) == 0x41)
        if (--cia2tb <= 0)
            setalarm(I_CIA2TB, 0);      /*int_clk[I_CIA2TB] = clk;*/

    return 0; /*(cia2int & cia2[CIA_ICR] & 1) ? 1:0;*/  /* strobe irq */
}


/*
 * Timer B can run in 2 (4) modes
 * cia2[f] & 0x60 == 0x00   count system 02 pulses
 * cia2[f] & 0x60 == 0x40   count timer A underflows
 * cia2[f] & 0x60 == 0x20 | 0x60 count CNT pulses => counter stops
 */


int    int_cia2tb(int offset)
{
     if (debugflg)
        printf("cia2 timer B event\n");

    if (cia2[CIA_CRB] & 8) {    /* one-shot mode */
#if defined (CIA2_TIMER_DEBUG)
        if(debugflg) printf ("Cia#2 Timer B event -- one-shot mode:"
			" next int won't happen!\n");
#endif
        cia2tb = 0x10000;
        cia2tb_stopped = 1;
        unsetalarm(I_CIA2TB);
	cia2[CIA_CRB] &= ~1;
    } else {            /* continuous mode */
        update_cia2tb(1, offset);
    }
    cia2int |= 2;
    if(cia2[CIA_ICR] & 2) setirq(I_CIA2TB, 2);

/*printf("cia2 tb int returns %d\n",!(cia2int & cia2[P_INTENAB] & 2));*/

    return 0; /* (cia2int & cia2[CIA_ICR] & 2) ? 1:0;*/
}

int int_cia2tod(void) {
        register int t, pm;

if (debugflg)
  printf("CIA2: TOD timer event (1/10 sec tick), tod=%02x:%02x,%02x.%x\n",
                cia2[CIA_TOD_HR], cia2[CIA_TOD_MIN], cia2[CIA_TOD_SEC],
                cia2[CIA_TOD_TEN]);

        /* set up new int */
        setalarm(I_CIA2TOD, tod2ticks);

        if(!tod2stopped) {
          /* inc timer */
          t = bcd2byte(cia2[CIA_TOD_TEN]);
          t++;
          cia2[CIA_TOD_TEN] = byte2bcd(t%10);
          if( t>= 10 ) {
            t = bcd2byte(cia2[CIA_TOD_SEC]);
            t++;
            cia2[CIA_TOD_SEC] = byte2bcd(t%60);
            if(t >= 60) {
              t = bcd2byte(cia2[CIA_TOD_MIN]);
              t++;
              cia2[CIA_TOD_MIN] = byte2bcd(t%60);
              if(t>= 60) {
                pm = cia2[CIA_TOD_HR] & 0x80;
                t = bcd2byte(cia2[CIA_TOD_HR]&0x1f);
                if(!t) pm ^= 0x80;      /* toggle am/pm on 0:59->1:00 hr */
                t++;
                t = t%12 | pm;
                cia2[CIA_TOD_HR] = byte2bcd(t);
              }
            }
          }
if (debugflg)
  printf("CIA2: TOD after event :tod=%02x:%02x,%02x.%x\n",
                cia2[CIA_TOD_HR], cia2[CIA_TOD_MIN], cia2[CIA_TOD_SEC],
                cia2[CIA_TOD_TEN]);
          /* check alarm */
	  check_tod2alarm();
#if 0
          if(!memcmp(tod2alarm, cia2+CIA_TOD_TEN, sizeof(tod2alarm))) {
            cia2int |= CIA_IM_TOD;
            if(cia2[CIA_ICR] & CIA_IM_TOD) setirq(I_CIA2TOD, 2);
          }
#endif
        }
        return 0;
}

void check_tod2alarm(void) {
          if(!memcmp(tod2alarm, cia2+CIA_TOD_TEN, sizeof(tod2alarm))) {
            cia2int |= CIA_IM_TOD;
            if(cia2[CIA_ICR] & CIA_IM_TOD) setirq(I_CIA2TOD, 2);
          }
}
 
/* ------------------------------------------------------------------------- */

static void update_cia2ta(int force, int offset)
{
/*printf("update_cia2ta(force=%d, stopped=%d)\n",force, cia2ta_stopped);*/
    if(force) {
        cia2ta = cia2[CIA_TAL] + (cia2[CIA_TAH] << 8) - (offset & 0xffff);
 	if(cia2ta <= 0) cia2ta += 0x10000;
        if(!cia2ta_stopped)
            setalarm(I_CIA2TA, cia2ta);
    } else
    if(!cia2ta_stopped) {
        cia2ta = (int)(cia2ta - clk + cia2tau) & 0xffff;
 	if(cia2ta <= 0) cia2ta += 0x10000;
    }
    cia2tau = clk;
}


static void update_cia2tb(int force, int offset)
{
    if (force) {
        cia2tb = cia2[CIA_TBL] + (cia2[CIA_TBH] << 8) - (offset & 0xffff);
 	if(cia2tb <= 0) cia2tb += 0x10000; 
        if(!cia2tb_stopped)
            setalarm(I_CIA2TB, cia2tb);
    } else
    if (!cia2tb_stopped) {
        cia2tb = (int)(cia2tb - clk + cia2tbu) & 0xffff;
 	if(cia2tb <= 0) cia2tb += 0x10000;
    }
    cia2tbu = clk;
}

#if 0
static void update_tod(void)
{
    long   c;
    time_t t = 0;

    /*if (tod2mode != 1) return; */  /* Not Running */

    if (tod2virt) {
	c = clk - tod2cycles;
	t = (double)c / (tod2freq ? 9852.48 : 10227.30); /* CYCLES / 0.01 SEC */
    }
    else {
	time(&t);
	t -= tod2rel;
    }

    cia2[CIA_TOD_TEN] = 0;			/* Time Of Day clock 1/10 s */
    cia2[CIA_TOD_SEC] = (BYTE)(t/3600);		/* Time Of Day clock sec */
    cia2[CIA_TOD_MIN] = (BYTE)((t%3600)/60);	/* Time Of Day clock min */
    cia2[CIA_TOD_HR]  = (BYTE)(t%60);
}
#endif
