/*
 * $Id: interrupt.c,v 2.1 1996/07/29 21:10:07 fachat Exp ettore $
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README for copyright notice
 *
 * This file contains interrupt handling
 *
 * Written by
 *   Jarkko Sonninen  (sonninen@lut.fi)
 *   Jouko Valta      (jopi@stekt.oulu.fi)
 *   Hannu Helminen   (dm@stekt.oulu.fi)
 *   Andre Fachat     (a.fachat@physik.tu-chemnitz.de)
 *
 * Patches by
 *   Ettore Perazzoli (ettore@comm2000.it)
 *
 *
 * $Log: interrupt.c,v $
 * Revision 2.1  1996/07/29 21:10:07  ettore
 * Adds the correct number of cycles for each interrupt.
 * NMI fixes.
 *
 * Revision 2.0  1996/04/21 10:30:49  fachat
 * Introduced new timer/alarm code. Interface to other routines is
 * via [un]setalarm and setirq. The IRQ line is now saved until
 * a cli/rti occurs, instead of being dropped when the interrupt
 * flag is set. The code is still compatible with the old
 * int_clk[] = (...); find_next_int(); sequence, which still works.
 *
 * Revision 1.5  1996/04/01  09:01:41  jopi
 * PET and VIC-20 interrupts
 *
 * Revision 1.4  1995/11/07  16:51:00  jopi
 * *** empty log message ***
 *
 * Revision 1.3  1995/04/01  07:54:09  jopi
 * X64 0.3 PL 0
 * Bug in interrupt vs. RTI fixed.
 *
 * Revision 1.2  1994/12/12  16:59:44  jopi
 * 16-bit portability
 *
 * Revision 1.1  1994/06/16  17:19:26  jopi
 * Initial revision
 *
 *
 */

/* 1996/04/10 a.fachat
 *
 * We need a new, clear concept for timer event and interrupt handling.
 * First, timer events occur not depending on the CPU interrupt flag.
 * Because of such events, the CPU IRQ or NMI line may be set. But
 * when scheduling a new timer event during a timer event, when the CPU
 * interrupt flag is set, this timer IRQ has to be saved. Otherwise
 * the IRQ is missed. 
 * Therefore we have two arrays, one for the independent timer events: 
 *		
 * 	int_clk[I_*]	0 	= event source disabled,
 *			> clk 	= event in the future
 *			<= clk	= event to be served now.
 * 
 *	setalarm(I_*, ticks_from_now)
 *	unsetalarm(I_*)		
 *				(un)schedule the timer events
 *
 *	find_next_int()		finds the next timer event and
 *				sets the CPU alarm to handle the next event.
 *
 * Then we need one array to save if an irq should be given to the CPU.
 *
 *	int_cpu[I_*]	0	= no irq pending
 *			1	= irq pending
 *			2	= nmi pending
 * 
 *	setirq(I_*, value)	sets the value for irq/nmi for the
 *				specified event type.
 *				It automagically sets the global irq/nmi flag
 *
 *	cpu_irq			global CPU irq/nmi flag - changed by setirq
 *				only!
 *
 * How are these functions called?
 *
 *	intr() is the routine that is called when a timer event is scheduled.
 *	It calls the appropriate timer irq routine. The timer event is
 *	automatically removed unless set by the routine again.
 *	The interrupt routine returns the irq value (as with setirq) if a 
 *	single interrupt should be 
 *	scheduled to the CPU, bypassing the irq/nmi flag (strobe irq's).
 *	The timer interrupt routines as well as simple read/stores to the
 *	I/O chips can schedule timer events via (un)setalarm and 
 *	set interrupts via setirq.
 *
 *	cpu_do_int() then checks the global irq/nmi flag and sets the PC
 *	and stack accordingly. NMI events are removed from the list.
 */

#include <stdio.h>
#include <time.h>

#include "config.h"
#include "macro.h"
#include "vmachine.h"
#include "extern.h"
#include "interrupt.h"

extern int halt;

/* global */

int     next_int = 0;
CLOCK   next_int_clk = 0;
CLOCK   cpu_tmp_clk = 0;
CLOCK   int_clk[NUMOFINT];

/*int 	cpu_check_int = 0;*/
int 	cpu_check_irq = 0;

/* irq bookkeeping */

static int int_cpu[NUMOFINT];
static int cpu_strobe_irq = 0;

static int nirq = 0;
static int nnmi = 0;
static int last_nnmi = 0;

/* 
 * These routines are exported to allow easy irq handling (a.fachat)
 * We should smoothly change to this interface, until int_clk can be 
 * made static, which allows hiding the timer code to the rest of
 * of vice.
 * 
 */

#define	set_cpu_check_irq()	(cpu_check_irq=nirq||nnmi||cpu_strobe_irq)

#if 0		/* [un]setalarm have become cpp macros... */
void unsetalarm( int irq ) {
/*printf("unsetalarm (irq=%d)\n",irq);*/
	int_clk[irq] = 0;
	cpu_check_int = 1; 	/*find_next_int();*/
}

void setalarm( int irq, int ticks ) {
	int_clk[irq] = clk + (ticks & 0xffff);
/*printf("setalarm (irq = %d, ticks = %d, clk = %ld)-> %ld\n", 
					irq, ticks, clk, int_clk[irq]);*/

	cpu_check_int = 1; 	/*find_next_int();*/
}
#endif

void setirq( int irq, int val ) {
	register int k;
	 
	if( int_cpu[irq] != val ) {	/* this alone should make it faster */
	  k = int_cpu[irq]^val;
	  if(k & 1) {			/* changed IRQ state */
	    if(val & 1) {
	      nirq ++;
/*if(debugflg) printf("new irq source %d, n=%d now\n",irq, nirq);*/
	    } else {
	      if(nirq)
	        nirq --;
	      else
		printf("Oops: irq bookkeeping messed up!\n");
/*if(debugflg) printf("removed irq source %d, n=%d now\n",irq, nirq);*/
	    }
	  }
	  if(k & 2) {			/* changed NMI state */
	    if(val & 2) {
	      nnmi ++;
/*if(debugflg) printf("new nmi source %d, n=%d now\n",irq, nnmi);*/
	    } else {
	      if(nnmi)
	        nnmi --;
	      else
		printf("Oops: nmi bookkeeping messed up!\n");
	      /* 
		when going down, set last_nnmi. This works because
		only 0->1 flank has to be detected, not 1->0 transition.
		Otherwise, when removing nmi, but not calling cpu_do_int()
		before the next nmi occurs,
		the transition is not detected because cpu_do_int() had
		no chance setting last_nnmi and the nmi is missed.
	      */
	      last_nnmi = nnmi; 
/*if(debugflg) printf("removed nmi source %d, n=%d now\n",irq, nnmi);*/
	    }
	  }	  
	  int_cpu[irq] = val;
	}
}

/* find next interrupt */

void    cpu_find_next_int(void)
{
    int     i;

    for (i = 0, next_int_clk = (CLOCK) -1; i < NUMOFINT; i++)
	if (int_clk[i] && int_clk[i] < next_int_clk) {
	    next_int_clk = int_clk[i];
	    next_int = i;
	}
    if (debugflg && next_int != I_SPECX)
	printf("%ld next int %d %ld int_clk[8]=%ld\n", 
			clk, next_int, next_int_clk, int_clk[8]);
/*    cpu_check_int = 0; */
}

void cpu_do_nmi(void) {
	cpu_strobe_irq = 2;
	PC = cpu_do_int(PC);
}

ADDRESS cpu_do_int(ADDRESS pc) {
    	BYTE    t;
/*if(debugflg && nnmi) printf("cpu_do_int(pc=%x, nnmi=%d, last_nnmi=%d)\n",
			pc, nnmi, last_nnmi);*/
	if((nnmi && !last_nnmi) || (cpu_strobe_irq /* | cpu_irq)*/ & 2)) {	/* nmi */
if(debugflg) { 
	int i;
	printf("Entering nmi routine (clk=%ld, nnmi=%d, strobe=%d, pc=%x)\n",
						clk, nnmi, cpu_strobe_irq,pc);
	printf("int_cpu[]=");
	for(i=0;i<NUMOFINT; i++) printf("%d ",int_cpu[i]);
	printf("\n");
/*
	halt = 1;
	setalarm(I_SPECX,0);
*/
}
          SET_BREAK(0);
          PUSH(UPPER(pc));
          PUSH(LOWER(pc));        /* Push return address into the stack. */
          t = GET_SR();
          PUSH(t);                /* Push status register into the stack. */
          SET_INTERRUPT(1);       /* Disable interrupts. */
          pc = LOAD_ADDR(0xFFFA); /* Jump using NMI vector. */
          clk += NMI_CYCLES;	/* EP Jun 7 1996 */
if(debugflg) printf("PC has become %x\n",pc);
	} else
	if(nirq || (cpu_strobe_irq /*| cpu_irq)*/ & 1)) {	/* irq */
          if (!IF_INTERRUPT()) {
if(debugflg) { 
	int i;
	printf("Entering interrupt routine (clk=%ld, nirq=%d, strobe=%d)\n",
						clk, nirq, cpu_strobe_irq);
	printf("int_cpu[]=");
	for(i=0;i<NUMOFINT; i++) printf("%d ",int_cpu[i]);
	printf("\n");
}
            SET_BREAK(0);
            PUSH(UPPER(pc));
            PUSH(LOWER(pc));        /* Push return address into the stack. */
            t = GET_SR();
            PUSH(t);                /* Push status register into the stack. */
            SET_INTERRUPT(1);       /* Disable interrupts. */
            pc = LOAD_ADDR((ADDRESS)65534); /* Jump using IRQ vector (0xfffe).*/
            clk += IRQ_CYCLES;	/* EP Jun 7 1996 */
	  }
	}
	cpu_strobe_irq = 0;	/* clear strobe irq */
	last_nnmi = nnmi;
    	set_cpu_check_irq();
	return pc;
}

void    restore(void)
{
    int_clk[I_RESTORE] = clk;
    cpu_find_next_int();
/*
#ifdef NEW_TIMER_CODE
    update_intr_alarm();
#endif
*/
}


/*
 * Serve interrupts. Called from CPU mainloop.
 */
/*
void    test_irq(CLOCK ck)
{
#ifdef NEW_TIMER_CODE
    if (ck >= next_alarm_clk)
	alarm_handler();
#else
    while (ck > next_int_clk)
	PC = intr(next_int, PC);
#endif
}
*/

/* interrupts */

/* This should be cleaned up when it's clear that everything is ok! */

ADDRESS intr(int sig, ADDRESS pc)
{
    BYTE    t;
    long offset;

    offset = clk - next_int_clk;

    if(debugflg && (sig!=I_SPECX)) {
	printf("intr: sig = %d, PC = %04x, clk = %ld, next_int_clk=%ld\n", 
				sig, pc, clk, next_int_clk);
	if((sig!=I_BRK) && offset)
	  printf("    : scheduled %ld tick(s) too late!\n", offset);
    }

    if((!next_int_clk) && clk) { cpu_find_next_int(); return pc; }

    int_clk[sig] = 0;

    switch (sig) {
      case I_RESET:
	clk = RESET_CYCLES;	/* EP Jun 9 1996 */

#ifdef PET
	reset_pia1();
	reset_pia2();
	reset_via();
	reset_crtc();
#else
	ram[0] = 0xff;
	ram[1] = 0xff;
#ifdef CBM64
	memconfig = 0x7;
#endif
#ifdef VIC20
	reset_via1();
	reset_via2();
#else
	cia2[2] = 0;
	cia2[0] = 255;
	reset_cia1();
	reset_cia2();
#endif
#if defined(CBM64) && defined(IEEE488)
	reset_tpi();
#endif
#endif /* pet */

	/* init_vic(); */

	init_memmaps();		/* Reset Memory Config */
	printf("** RESET **\n");
	pc = LOAD_ADDR(0xFFFC);	/* Jump using RESET vector. */
	break;

      case I_BRK:
	printf("** BREAK **\n");
	SET_BREAK(1);		/* Set break status. */
	PUSH(UPPER(pc));
	PUSH(LOWER(pc));	/* Push return address into the stack. */
	t = GET_SR();
	PUSH(t);		/* Push status register into the stack. */
	SET_INTERRUPT(1);	/* Disable interrupts. */
	pc = LOAD_ADDR((ADDRESS)65534);	/* Jump using BRK vector (0xfffe). */
	return (pc);

#ifndef PET
#ifdef VIC20

      case I_VIA1T1:
	x_get_keyarr();
	cpu_strobe_irq |= int_via1t1();
	break;
      case I_VIA1T2:
	cpu_strobe_irq |= int_via1t2();
	break;

#else /* !VIC20 -> C64, C128 */

      case I_CIA1SP:
	break;
      case I_CIA1TOD:
	cpu_strobe_irq |= int_cia1tod();
	break;
      case I_CIA1TB:
	cpu_strobe_irq |= int_cia1tb(offset);
	break;
      case I_CIA1TA:
	/*printf("TIMER IRQ\n");*/
#if !defined (C128) && !defined (PET)
	x_get_keyarr();
	cpu_strobe_irq |= int_cia1ta(offset);
	break;
#endif
      case I_CIA1FL:
      case I_LIGHT_PEN:
      case I_SS_COL:
      case I_SB_COL:
#endif /* VIC20 */

#else /* PET irqs follow */

      case I_VIAT1:
	cpu_strobe_irq |= int_viata();
	break;
      case I_VIAT2:
	cpu_strobe_irq |= int_viatb();
	break;
      case I_RASTER:		
#endif /* PET */

	cpu_strobe_irq |= 1;	/* strobe irq */
	break;

#ifndef PET
#ifdef VIC20

      case I_VIA2T1:
	cpu_strobe_irq |= int_via2t1();
	break;
      case I_VIA2T2:
	cpu_strobe_irq |= int_via2t2();
	break;
      case I_RESTORE:

#else
      case I_CIA2SP:
	break;
      case I_CIA2TOD:
	cpu_strobe_irq |= int_cia2tod();
	break;
      case I_CIA2TB:		/* Non-Maskable interrupts */
	cpu_strobe_irq |= int_cia2tb(offset);
	break;
      case I_CIA2TA:
	cpu_strobe_irq |= int_cia2ta(offset);
	break;

      case I_RESTORE:

#endif	/* vic20 */
#endif  /* pet */

	cpu_strobe_irq |= 2;	/* strobe NMI */
	break;

      case I_SPECX:
/*	int_clk[sig] = clk + 4 * CYCLES_PER_RFSH;  */
        int_clk[sig] = clk + CYCLES_PER_RFSH;
	x_loop();
#ifdef JOYSTICK
	joystick();
#endif
	pc = PC;	/* PC may be changed in x_loop() or mon() */
	break;

#ifndef PET
      case I_RASTER:
	int_raster();
#ifdef C128
	x_get_keyarr();
#endif
	break;
#endif

#ifdef TEEMU_SID 
      case I_SID:
	playsid();
	break;
#endif
	
      /* unsupported irqs follow */
      /* uuh, well, messed up somehow - someone should find all of them */
#ifndef PET
#ifndef VIC20
      case I_CIA2FL:
#else
      case I_VIA1SR:
      case I_VIA1FL:
#endif
#else	/* PET unsupported irqs */
      case I_VIASR:
      case I_VIAFL:
#endif
	printf("unsupported interrupt %d\n", sig);
	show();
	end64();
    }

    set_cpu_check_irq();
    cpu_find_next_int();
    return (pc);
}


/*
 * This routine is used to skip idle on ROM busy loop
 */
void    jump_to_next_alarm()
{
    clk = next_int_clk;
}

#if 0
#ifdef NEW_TIMER_CODE
/*
 * This is part kludge to get both timers usable
 */
/* static */ int new_alarm_number;


void    new_alarm_trap(int i)
{
    while (clk > next_int_clk)
	PC = intr(next_int, PC);
    set_alarm(new_alarm_number, next_int_clk+1, new_alarm_trap, 0);
}


void    update_intr_alarm(void)
{
    set_alarm(new_alarm_number, next_int_clk+1, new_alarm_trap, 0);
    find_next_alarm();
}

#endif /* NEW_TIMER_CODE */
#endif /* 0 */
