/* Copyright (C) 1993, 1992 Nathan Sidwell */
/* RCS $Id: lock.c,v 1.2 1995/11/07 16:51:00 jopi Exp jopi $ */
/*  file locking problems*/
/*
 * USELOCKFILE file locking as suggested by
 * Daniel Edward Lovinger <del+@cmu.edu>
 * With lockf (or flock), we just use the kernel's locking stuff to lock the
 * entire score file while we read, or update it. But some distributed
 * file systems don't support it and some are broken (SunOS 4.1).
 * USELOCKFILE uses uses open(O_CREAT | O_EXCL) to create
 * a lock file in the same directory as the xmris high score file, with
 * the name "xmris.lock".
 * The personal score files are either in the score file directory
 * with names "xmris-<name>", or in the user's home directory with name
 * ".xmris.scores".
 *
 * Some systems have flock (BSD), and some have lockf (SYSV).
 */


#include <sys/stat.h>
#ifdef USELOCKFILE
#include <fcntl.h>
#endif /* USELOCKFILE */

/* File locking*/
#ifdef SYSV
#define lock_file(stream) lockf(fileno(stream), F_LOCK, 0L)
#define unlock_file(stream) lockf(fileno(stream), F_ULOCK, 0L)
#else
#define lock_file(stream) flock(fileno(stream), LOCK_EX)
#define unlock_file(stream) flock(fileno(stream), LOCK_UN)
#endif /* SYSV */


/* static*/
static char     *personal_home = NULL;  /* personal in home dir */
#ifdef USELOCKFILE
static char     *locking_file = NULL;  /* lock file name */
static unsigned locks;      /* number of locks open */
#endif /* USELOCKFILE */
static HIGH_SCORE *CONST tables[] =
  {scoring.high, scoring.today, scoring.personal};


/*  prototypes*/
static unsigned file_changed PROTOARG((char CONST *, uid_t));
static FILE *get_lock PROTOARG((char CONST *, unsigned, uid_t));
static VOIDFUNC get_unlock PROTOARG((FILE *));


/* ------------------------------------------------------------------------- */

/*  void check_scores()*/
extern VOIDFUNC check_scores FUNCARGVOID
{
  if((score_file && file_changed(score_file, effective_uid)) ||
      (!personal_make && personal_file &&
      file_changed(personal_file, personal_uid)))
    load_check_expire_insert(0);
  return;
}


/*  void file_changed(name, uid)*/
static unsigned file_changed
FUNCARG((name, uid),
	char CONST *name
ARGSEP  uid_t       uid
)
/*
 * check if a score file has been changed since last I looked,
 * so that we pick up the new scores
 */
{
  static  time_t last_time[2];
  struct stat buffer;
  unsigned  changed;

  
  assert(name);

  if(!stat(name, &buffer))
    {
      changed = buffer.st_mtime != last_time[name == score_file];
      last_time[name == score_file] = buffer.st_mtime;
    }
  else
    changed = 0;

  return changed;
#endif /* TRANSPUTER */
}


/*  void get_lock(name, flag, uid)*/
static FILE *get_lock
FUNCARG((name, flag, uid),
	char CONST *name
ARGSEP  unsigned  flag
ARGSEP  uid_t     uid
)
/*
 * open and locks a high score file
 * flag & 1 == 0 -> "r+"
 * flag & 1 != 0 -> "w+"
 * flag & 2 inhibit error message
 * flag & 4 && effective_uid == real_uid set chmod 666
 * uid required to access
 */
{
  FILE    *stream;

#ifdef USELOCKFILE
  if(locking_file && !locks)
    /*  attempt exclusive file lock*/
    {
      unsigned  count;
      int       filed;
    
      for(count = 3; count;)
	{
	  filed = open(locking_file, O_CREAT | O_EXCL, 0666);
	  if(filed >= 0)
	    break;
	  if(errno == EINTR)
	    continue;
	  else if(errno == EEXIST)
	    {
	      sleep(1);
	      if(!file_changed(name, uid))
		count--;
	    }
	  else
	    {
	      perror(locking_file);
	      break;
	    }
	}
      if(filed >= 0)
	close(filed);
    }
#endif /* USELOCKFILE */


  stream = fopen(name, flag & 1 ? "w+" : "r+");
  if(!stream && !(flag & 2))
    perror(name);
#ifdef USELOCKFILE
  if(stream)
    locks++;
  else if(locking_file && !locks)
    {
      unlink(locking_file);
    }
#else
  if(stream)
    /*  get lock on the file*/
    while(lock_file(stream))
      {
	if(errno == EINTR)
	  continue;
      }
    

#endif /* USELOCKFILE */
#ifndef TRANSPUTER
  if(stream && flag & 4 && effective_uid == real_uid)
    fchmod(fileno(stream), 0666);
#endif /* TRANSPUTER */
  return stream;
}


/*  void get_unlock(stream)*/
static VOIDFUNC get_unlock
FUNCARG((stream),
	FILE    *stream
)
/*
 * unlock and close the high score file
 */
{
  fflush(stream);
#ifdef USELOCKFILE
  fclose(stream);
  locks--;
  if(locking_file && locks)
    {
      unlink(locking_file);
    }
#else
  rewind(stream);
  unlock_file(stream);
  fclose(stream);
#endif /* USELOCKFILE */
  return;
}


/* ------------------------------------------------------------------------- */

/*  void init_scores()*/
extern VOIDFUNC init_scores FUNCARGVOID
/*
 * check that we have the wherewithall to deal with
 * high scores
 */
{
  char CONST *user;
  char CONST *home;
  char      *reallife;
  size_t    dirlen;
  
  user = NULL;
  home = NULL;
  reallife = NULL;


  dirlen = data.dir ? strlen(data.dir) : 0;
  if(dirlen && data.dir[dirlen - 1] == '/')
    dirlen--;

#ifdef USELOCKFILE
  /*  lock file?*/
  if(dirlen)
    {
      locking_file = malloc(dirlen + 12);
      if(locking_file)
	{
	  strcpy(locking_file, data.dir);
	  strcpy(&locking_file[dirlen], "/xmris.lock");
	}
    }
  

#endif /* USELOCKFILE */
  /*  score directory?*/
  if(dirlen)
    {
      size_t    nlen;
  
      score_file = malloc(dirlen + 14);
      if(score_file)
	{
	  strcpy(score_file, data.dir);
	  strcpy(&score_file[dirlen], "/xmris.score");
	}
      nlen = strlen(user);
      personal_file = malloc(dirlen + 8 + nlen);
      if(personal_file)
	{
	  strcpy(personal_file, data.dir);
	  strcpy(&personal_file[dirlen], "/xmris-");
	  strcpy(&personal_file[dirlen + 7], user);
	}
    }
  

  /*  personal file*/
  {
    if(!home)
      home = getenv("HOME");
    if(home)
      {
	size_t    length;
	
	length = strlen(home);
	personal_home = malloc(length + 15);
	if(personal_home)
	  {
	    strcpy(personal_home, home);
	    strcpy(&personal_home[length], "/.xmris.score");
	  }
      }
  }

  return;
}



/*  void load_check_expire_insert(insert)*/
static VOIDFUNC load_check_expire_insert
FUNCARG((insert),
	unsigned  insert
)
/* this does most of the high score file manipulation
 * it loads up the high score files
 * checks the integrity of the personal score vs high score
 * optionally inserts a new score into the tables
 * optionally expires scores
 * write the files back
 */
{
  FILE      *score_stream;
  FILE      *personal_stream;
  FILE      *home_stream;
  unsigned  do_score, do_personal;
  
  score_stream = personal_stream = home_stream = NULL;
  do_score = do_personal = 0;
  /*  score_file?*/
  if(score_file)
    {
      score_stream = get_lock(score_file, 2, effective_uid);
      if(score_stream)
	do_score = merge_scores(score_stream);
      else
	do_score = 1;
    }
  

      .
      .
      .


  /*  write personal?*/
  if(do_personal)
    {
      if(!personal_stream && score_stream && personal_file)
	personal_stream = get_lock(personal_file, 3, personal_uid);
      if(personal_stream)
	{
	  write_personal(personal_stream);
	  if(personal_home)
	    unlink(personal_home);
	  personal_home = NULL;
	}
      else
	{
	  if(!home_stream)
	    home_stream = get_lock(personal_home, 1, real_uid);
	  if(home_stream)
	    write_personal(home_stream);
	  else
	    personal_home = NULL;
	  personal_uid = real_uid;
	  personal_file = personal_home;
	  personal_home = NULL;
	}
      personal_make = 0;
    }
  

  if(personal_stream)
    get_unlock(personal_stream);
  if(home_stream)
    get_unlock(home_stream);
  if(score_stream)
    get_unlock(score_stream);
  if(do_score && score_file)
    file_changed(score_file, effective_uid);
  if(do_personal && personal_file)
    file_changed(personal_file, personal_uid);
  return;
}



/*  unsigned merge_scores(stream)*/
static unsigned merge_scores
FUNCARG((stream),
	FILE      *stream
)
/*
 * merges the high score file into the current high score
 * table. If the score file is corrupted, then we lose the
 * merged entries. The file must have been locked appropriately.
 * returns non-zero if the file is unreadable
 */
{
  unsigned long check;
  unsigned  failed;
  time_t    now;
  HIGH_SCORE  new;
  
  clearerr(stream);

  failed = ferror(stream) || check != new.stamp;

  if(failed)
    {
      fprintf(stderr, "%s:The high score file '%s' has been corrupted.\n",
	  myname, score_file);
    }
  return failed;
}
