/*
 * $Id: raster.c,v 1.2 1996/07/29 22:41:06 ettore Exp ettore $
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README file for copyright notice.
 *
 *
 * Versatile raster emulator.
 *
 *
 * Written by
 *   Ettore Perazzoli (ettore@comm2000.it)
 *
 *
 * $Log: raster.c,v $
 * Revision 1.2  1996/07/29 22:41:06  ettore
 * VIC-II specific improvements:
 *   bad lines;
 *   correct sprite drawing;
 *   timing;
 *   illegal lines fetch 0x3fff.
 * Explicitly requests 8 bit depth when creating the XImage.
 * Slightly faster: PUTMONO* and PUTMULTI* use parameters instead of
 * global variables.
 * Calls sigpause() after each frame, even if it is has not been
 * recalculated.
 * VINTAGE_DISPLAY option.
 *
 * Revision 1.1  1996/04/01 05:50:08  ettore
 * Initial revision
 *
 */


#define _RASTER_C

#if defined( __hpux)
#define _INCLUDE_HPUX_SOURCE
#endif
#include <sys/time.h>
#if defined( __hpux)
#undef _INCLUDE_HPUX_SOURCE
#endif
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <string.h>

#include "config.h"
#include "extern.h"
#include "raster.h"
#include "vmachine.h"
#include "video_x.h"
#include "rasterutil.h"
#include "x64.h"

#if defined( MITSHM )
#include <X11/extensions/XShm.h>
#include <sys/ipc.h>
#include <sys/shm.h>
static XShmSegmentInfo xshm_info;
#endif

extern Display *display;
extern int screen;
extern GC gc;
extern char *wintitle;

#define SCR_PER_TCALC        75
#define tvtodouble( tv )     ( (tv).tv_sec + (tv).tv_usec / 1000000.0 )

#if defined( NEEDS_NormalGetTextData)
#define NEEDS_GetCharData
#endif
#if defined( NEEDS_NormalGetBitmapData)
#define NEEDS_GetData
#endif
#if defined( NEEDS_NormalGetColorData)
#define NEEDS_GetData
#endif
#if defined( NEEDS_NormalGetBgDataFromRegister)
#define NEEDS_GetData
#endif

/* #define RASTER_DEBUG_PUTIMAGE_CALLS */
/* #define RASTER_DEBUG_MITSHM */
/* #define DEBUG_ARRANGEVIDEOIMAGE */

static int use_mitshm = 0;
static int havetosleep = 1;

static void RefreshAll (VideoStruct * v);


/*

   This is a simple alarm handler used to keep the speed of screen updates fixed.

 */

static void
sigalrm_handler (void)
{
  havetosleep = 0;
}


/*

   AllocSpriteLineStruct(): allocates memory for sprite information in a SpriteStruct

 */

static SpriteLineStruct *
AllocSpriteLineStruct (int num)
{
  SpriteLineStruct *new;

  if (!num)
    return NULL;
  new = (SpriteLineStruct *) malloc (sizeof (SpriteLineStruct) * num);

  if (new)
    new->visible = 0;
  return new;
}


/*

   AllocLineStruct(): allocates the structure for line information.

 */

static LineStruct *
AllocLineStruct (int ncols, int numlines, int numsprites)
{
  LineStruct *new;
  int i;

  new = (LineStruct *) malloc (sizeof (LineStruct) * numlines);
  if (!new)
    return NULL;

  for (i = 0; i < numlines; i++)
    {
      if (numsprites)
	new[i].sprites = AllocSpriteLineStruct (numsprites);
      new[i].fgdata = (BYTE *) malloc (sizeof (BYTE) * ncols);
      new[i].bgdata = (BYTE *) malloc (sizeof (BYTE) * ncols);
      new[i].colordata1 = (BYTE *) malloc (sizeof (BYTE) * ncols);
      new[i].colordata2 = (BYTE *) malloc (sizeof (BYTE) * ncols);
      new[i].colordata3 = (BYTE *) malloc (sizeof (BYTE) * ncols);
      if ((!(new[i].sprites) && numsprites) || !(new[i].fgdata) ||
	  !(new[i].bgdata) || !(new[i].colordata1) || !(new[i].colordata1) ||
	  !(new[i].colordata1))
	return NULL;
      new[i].numsprites = new[i].sprmask = 0;
      new[i].bordcol = -1;
    }

  return new;
}


/*

   AllocVideo(): allocates and (partially) initializes a VideoStruct. It does not open the
   window, though.
   It also installs the alarm handler to keep the speed of the emulation fixed.

 */

static VideoStruct *
AllocVideo (int scrcols, int scrlines, int bordwidth, int bordheight, int xpix, int ypix,
	    int tcols, int tlines, int numcolors, int refrate, int numsprites, int speed_index,
	    char *title, int asleep)
{
  VideoStruct *new;
  PIXEL *data;
  int i;
#if defined( MITSHM )
  int m1, m2, pixmap_flag;
#endif

  new = (VideoStruct *) malloc (sizeof (VideoStruct));

  if (!new)
    return NULL;

  new->scrcols = scrcols;
  new->scrlines = scrlines;
  new->refrate = new->refcount = refrate;
#if defined( MITSHM )
  if (!XShmQueryVersion (display, &m1, &m2, &pixmap_flag))
    fprintf (stderr, "warning: MITSHM extension not supported\n");
  else if (mitshm)
    {
      use_mitshm = 1;
#if defined( RASTER_DEBUG_MITSHM)
      printf ("AllocVideo(): allocating XImage with MITSHM, %d x %d pixels... ",
	      scrcols * PIXEL_XSIZE, scrlines * PIXEL_YSIZE);
#endif
      new->scrimg = XShmCreateImage (display, visual, depth, ZPixmap, NULL, &xshm_info,
				     scrcols * PIXEL_XSIZE, scrlines * PIXEL_YSIZE);
      if (!new->scrimg)
	{
	  fprintf (stderr, "error: can't allocate XImage with XShm.\n");
	  return NULL;
	}
#if defined( RASTER_DEBUG_MITSHM)
      printf ("done.\nAllocVideo(): shmgetting %ld bytes... ",
	      (long) new->scrimg->bytes_per_line * new->scrimg->height);
#endif
      xshm_info.shmid = shmget (IPC_PRIVATE, new->scrimg->bytes_per_line *
				new->scrimg->height, IPC_CREAT | 0777);
      if (xshm_info.shmid == -1)
	{
	  fprintf (stderr, "error: can't get shared memory.\n");
	  return NULL;
	}
#if defined( RASTER_DEBUG_MITSHM)
      printf ("done, id = 0x%lx.\nAllocVideo(): getting address... ",
	      (unsigned long) xshm_info.shmid);
#endif
      if (!(xshm_info.shmaddr = new->scrimg->data = shmat (xshm_info.shmid, 0, 0)))
	{
	  fprintf (stderr, "error: can't get shared memory address.\n");
	  return NULL;
	}
#if defined( RASTER_DEBUG_MITSHM)
      printf ("0x%lx OK.\n", (unsigned long) xshm_info.shmaddr);
#endif
      xshm_info.readOnly = 0;
      if (!XShmAttach (display, &xshm_info))
	{
	  fprintf (stderr, "error: can't attach shared memory.\n");
	  return NULL;
	}
      new->scrp = (PIXEL *) new->scrimg->data;
    }
#endif

  if (!use_mitshm)
    {
      /*printf ("warning: using non-mitshm video emulation\n\n"); */
      data = (PIXEL *) malloc (scrcols * PIXEL_XSIZE * scrlines * PIXEL_YSIZE);
      if (!data)
	return NULL;
      new->scrimg = XCreateImage (display, visual, depth, ZPixmap, 0, (char *) data,
				  scrcols * PIXEL_XSIZE, scrlines * PIXEL_YSIZE, 8, 0);
      new->scrp = (PIXEL *) new->scrimg->data;
    }

  new->rasterline = new->oldclk = new->xsmooth = new->ysmooth = new->brdgrow_xl = new->asleep =
    new->brdgrow_xr = new->brdgrow_yl = new->brdgrow_yu = new->blank = new->memptr =
    new->ycounter = 0;
  new->beenpaused = 1;
  new->scount = -1;
  new->maxsprites = numsprites;
  new->xpix = xpix;
  new->ypix = ypix;
  new->tcols = tcols;
  new->tlines = tlines;
  new->bordheight = bordheight;
  new->bordwidth = bordwidth;
  new->speed_index = speed_index;
  new->asleep = asleep;

  new->wintitle = (char *) malloc (strlen (title) + 1);
  new->display = display;
  if (!new->wintitle)
    return NULL;
  strcpy (new->wintitle, title);

  new->refstruct.num_elem = 0;
  if (!(new->lineinfo = AllocLineStruct (tcols, scrlines, numsprites)))
    return NULL;

  if (numsprites)
    if (!(new->sprites = (SpriteStruct *) malloc (sizeof (SpriteStruct) * numsprites)))
      return NULL;

  for (i = 0; i < numsprites; i++)
    memset (&(new->sprites[i]), 0, sizeof (SpriteStruct));

  if (!(new->colors = (PIXEL *) malloc (sizeof (PIXEL) * numcolors)))
    return NULL;

  new->numcolors = numcolors;

  /* install the alarm handler for fixed speed */

  new->s_a_handler = sigalrm_handler;
  if (new->speed_index)
    if (SetVideoSpeed (new, new->speed_index))
      fprintf (stderr, "Warning: can't set timer for custom speed emulation\n");

  return new;
}


/*

   FreeVideo() is written to detach shared memory. NOTICE: it does not really free
   dinamically allocated VideoStructs.

 */

static void
FreeVideo (VideoStruct * v)
{
  if (!v)
    return;
#if defined( MITSHM )
  if (use_mitshm)
    {
      XShmDetach (display, &xshm_info);
      XDestroyImage (v->scrimg);
      if (shmdt (xshm_info.shmaddr))
	fprintf (stderr, "WARNING! can't release shared memory!\n");
      shmctl (xshm_info.shmid, IPC_RMID, 0);
    }
  else
#endif
    XDestroyImage (v->scrimg);
}


/*

   ArrangeVideoImage(): re-arranges the screen values to fit winwidth x winheight.

 */

static void
ArrangeVideoImage (VideoStruct * v, int winwidth, int winheight)
{
#if defined( DEBUG_ARRANGEVIDEOIMAGE)
  printf ("ArrangeVideoImage: win %ld width %d height %d\n", v->win, winwidth, winheight);
#endif
  v->fscanline = v->bordheight - ((winheight / PIXEL_YSIZE) - v->ypix) / 2 + 1;
  if (v->fscanline > (v->bordheight + 1))
    v->fscanline = v->bordheight;
  if (v->fscanline < 0)
    v->fscanline = 0;
  v->lscanline = v->fscanline + (winheight / PIXEL_YSIZE);
  v->xpos = ((winwidth / PIXEL_XSIZE) - v->xpix) / 2;
  v->xoffs = v->bordwidth - v->xpos;
#if defined( DEBUG_ARRANGEVIDEOIMAGE)
  if (v->xoffs < 0)
    {
      printf ("ArrangeVideoImage(): invalid xoffs %d\n", v->xoffs);
      v->xoffs = 0;
    }
#endif
  v->winwidth = winwidth;
  v->winheight = winheight;
}


/*

   OpenVideoWindow(): opens a window for the VideoStruct. Returns nonzero on error.
   (if the VideoStruct has the `asleep' flag set, it does not map the window).

 */

static int
OpenVideoWindow (Display * disp, int scr_num, VideoStruct * v, char *win_name, char *icon_name,
		 char *res_class, char *res_name, char *icon_data[], char *cnames[],
		 Colormap * cmap_return, Atom * wm_delwin)
{
  v->winwidth = (v->xpix + v->bordwidth) * PIXEL_XSIZE;
  v->winheight = (v->ypix + v->bordheight / 2) * PIXEL_YSIZE;
  v->win = create_window (disp, scr_num, win_name, icon_name, res_class, res_name,
			  v->winwidth, v->winheight, v->xpix * PIXEL_XSIZE, v->ypix * PIXEL_YSIZE,
			  (v->xpix + v->bordwidth * 2) * PIXEL_XSIZE,	/* max width */
			  (v->ypix + v->bordheight * 2) * PIXEL_YSIZE,	/* max height */
			  icon_data, !v->asleep, wm_delwin);
  if (!v->win)
    return 1;
  else
    {
      ArrangeVideoImage (v, v->winwidth, v->winheight);
      alloc_colors (disp, scr_num, v->win, v->numcolors, cnames, cmap_return, v->colors);
      memset (v->scrimg->data, v->colors[0], v->scrcols * PIXEL_XSIZE * v->scrlines *
	      PIXEL_YSIZE);
      RefreshAll (v);
      return 0;
    }
}


/*

   Here come the routines for fetching screen data from memory...

 */

#if defined( NEEDS_GetData)
INLINE int
GetData (BYTE * dest, BYTE * src, int length, int srcstep, int *xs, int *xe,
	 int no_check)

{
  int i = 0;

  if (no_check)
    {
      *xs = 0;
      *xe = length - 1;
      for (; i < length; i++, src += srcstep)
	dest[i] = src[0];
      return 1;
    }
  else
    {
      int x = 0;

      for (; dest[i] == src[0] && i < length; i++, src += srcstep)
	;
      if (i < length)
	{
	  if (*xs > i)
	    *xs = i;
	  for (; i < length; i++, src += srcstep)
	    if (dest[i] != src[0])
	      {
		dest[i] = src[0];
		x = i;
	      }
	  if (*xe < x)
	    *xe = x;
	  return 1;
	}
      else
	return 0;
    }
}
#endif /* NEEDS_GetData */


/*

   GetDataHiLo fetches the high and low nibbles separately.

 */

#if defined( NEEDS_GetDataHiLo)
INLINE int
GetDataHiLo (BYTE * desthi, BYTE * destlo, BYTE * src, int length, int srcstep, int *xs,
	     int *xe, int no_check)

{
  int i = 0;

  if (no_check)
    {
      *xs = 0;
      *xe = length - 1;
      for (; i < length; i++, src += srcstep)
	{
	  desthi[i] = HI_NIBBLE (src[0]);
	  destlo[i] = LO_NIBBLE (src[0]);
	}
      return 1;
    }
  else
    {
      int x = 0;
      BYTE b;

      for (; desthi[i] == HI_NIBBLE (src[0]) && destlo[i] == LO_NIBBLE (src[0]) &&
	   i < length; i++, src += srcstep)
	;
      if (i < length)
	{
	  if (*xs > i)
	    *xs = i;
	  for (; i < length; i++, src += srcstep)
	    if (desthi[i] != (b = HI_NIBBLE (src[0])))
	      {
		desthi[i] = b;
		x = i;
	      }
	    else if (destlo[i] != (b = LO_NIBBLE (src[0])))
	      {
		destlo[i] = b;
		x = i;
	      }
	  if (*xe < x)
	    *xe = x;
	  return 1;
	}
      else
	return 0;
    }
}
#endif /* defined( NEEDS_GetDataHiLo) */


/*

   GetCharData() takes the data as char codes and consequently reads the actual values for
   screen drawing from the character memory.

 */

#if defined( NEEDS_GetCharData)
INLINE int
GetCharData (BYTE * dest, BYTE * src, BYTE * charmem, int length, int l, int *xs, int *xe,
	     int no_check)

{
  int i = 0;

  if (no_check)
    {
      *xs = 0;
      *xe = length - 1;
      for (; i < length; i++, src++)
	dest[i] = GET_CHAR_DATA (charmem, src[0], l);
      return 1;
    }
  else
    {
      BYTE b;

      for (; dest[i] == GET_CHAR_DATA (charmem, src[0], l) && i < length; i++, src++)
	;
      if (i < length)
	{
	  *xs = *xe = i;
	  for (; i < length; i++, src++)
	    if (dest[i] != (b = GET_CHAR_DATA (charmem, src[0], l)))
	      {
		dest[i] = b;
		*xe = i;
	      }
	  return 1;
	}
      else
	return 0;
    }
}
#endif /* defined( NEEDS_GetCharData) */


/*

   GetSpriteData gets the data for sprites defined in a VideoStruct and puts them in a
   SpriteLineStruct, setting the other values too. xs and xe are expressed in pixels here.

 */

INLINE int
GetSpriteData (VideoStruct * v, LineStruct * ll, int *xs, int *xe)

{
  int rr = 0, i, r, sxe, sxs, sxe1, sxs1, n = 0, msk;
  BYTE *data;
  SpriteStruct spr;
  SpriteLineStruct *l = ll->sprites;
  SpriteLineStruct *sprl;

  ll->numsprites = v->maxsprites;
  ll->sprmask = 0;
  for (msk = 1, i = 0; i < v->maxsprites; i++, msk <<= 1)
    {
      spr = v->sprites[i];
      sprl = &(l[i]);
      r = 0;
      if (spr.dma_flag)
	{
	  ll->sprmask |= msk;
	  data = v->vbankbase + (v->spriteptrbase[i] << 6) + spr.memptr;
	  sxe = spr.x + (spr.xexp ? 48 : 24);
	  sxs = spr.x;
	  if (spr.x != sprl->x)
	    {
	      if (sprl->visible)
		{
		  sxe1 = sprl->x + (sprl->xexp ? 48 : 24);
		  sxs1 = sprl->x;
		  n++;
		  if (sxs1 < sxs)
		    sxs = sxs1;
		  if (sxe1 > sxe)
		    sxe = sxe1;
		}
	      sprl->x = spr.x;
	      r = 1;
	    }
	  sprl->visible = 1;
	  if (spr.xexp != sprl->xexp)
	    {
	      sprl->xexp = spr.xexp;
	      r = 1;
	    }
	  if (spr.multicolor != sprl->multicolor)
	    {
	      sprl->multicolor = spr.multicolor;
	      r = 1;
	    }
	  if (v->sprmc1 != sprl->c1)
	    {
	      sprl->c1 = v->sprmc1;
	      r = 1;
	    }
	  if (v->sprmc2 != sprl->c2)
	    {
	      sprl->c2 = v->sprmc2;
	      r = 1;
	    }
	  if (spr.color != sprl->c3)
	    {
	      sprl->c3 = spr.color;
	      r = 1;
	    }
	  if (spr.priority != sprl->priority)
	    {
	      sprl->priority = spr.priority;
	      r = 1;
	    }
	  if (sprl->data[0] != data[0])
	    {
	      sprl->data[0] = data[0];
	      r = 1;
	    }
	  if (sprl->data[1] != data[1])
	    {
	      sprl->data[1] = data[1];
	      r = 1;
	    }
	  if (sprl->data[2] != data[2])
	    {
	      sprl->data[2] = data[2];
	      r = 1;
	    }
	  if (r)
	    {
	      *xs = *xs < sxs ? *xs : sxs;
	      *xe = *xe > sxe ? *xe : sxe;
	      rr = 1;
	    }
	}
      else if (sprl->visible)
	{
	  sprl->visible = 0;
	  sxe = sprl->x + (sprl->xexp ? 24 : 48);
	  *xs = *xs < sprl->x ? *xs : sprl->x;
	  *xe = *xe < sxe ? sxe : *xe;
	  rr = 1;
	}
    }
  if (*xe >= v->scrcols)
    *xe = v->scrcols - 1;
  return rr;
}


/*

   The following functions get the screen data. When initializing the raster emulator, for each
   video mode you can choose to use them or to write your own ones (eg for particular modes,
   such as extended text mode in VIC-II).

 */

#if defined( NEEDS_NormalGetTextData)
static int
NormalGetTextData (VideoStruct * v, LineStruct * l, int *xs, int *xe, int no_check)
{
#if defined (__VIC_II__)
  if (v->ycounter > 7)
    return GetData (l->fgdata, v->vbankbase + 0x3fff, v->tcols, 0, xs, xe, no_check);
  else
    return GetCharData (l->fgdata, v->vbuf, v->chargenmem, v->tcols,
			v->ycounter, xs, xe, no_check);
#else
  return GetCharData (l->fgdata, v->screenmem + v->memptr, v->chargenmem, v->tcols,
		      v->ycounter, xs, xe, no_check);
#endif
}
#endif

#if defined( NEEDS_NormalGetBitmapData)
static int
NormalGetBitmapData (VideoStruct * v, LineStruct * l, int *xs, int *xe, int no_check)
{
#if defined (__VIC_II__)
  if (v->ycounter > 7)
    return GetData (l->fgdata, v->vbankbase + 0x3fff, v->tcols, 0, xs, xe, no_check);
  else
#endif
    return GetData (l->fgdata, v->bitmapmem + v->memptr * 8 + v->ycounter, v->tcols, 8,
		    xs, xe, no_check);
}
#endif

#if defined( NEEDS_NormalGetColorData)
static int
NormalGetColorData (VideoStruct * v, LineStruct * l, int *xs, int *xe,
		    int no_check)
{
#if defined (__VIC_II__)
  if (v->ycounter > 7)
    return GetData (l->colordata1, (BYTE *) "", v->tcols, 0, xs, xe, no_check);
  else
#endif
    return GetData (l->colordata1, v->colormem + v->memptr, v->tcols, 1, xs, xe, no_check);
}
#endif

#if defined( NEEDS_NormalGetBgDataFromRegister)
static int
NormalGetBgDataFromRegister (VideoStruct * v, LineStruct * l, int *xs, int *xe)
{
  return GetData (l->bgdata, &(v->bgcol), v->tcols, 0, xs, xe, 0);
}
#endif

#if defined( NEEDS_NullGetFunct)
static int
NullGetFunct (VideoStruct * v, LineStruct * l, BYTE * addr, int *xs, int *xe)
{
  return 0;
}
#endif


/*

   The following functions draw the raster line (background, foreground, sprites) from xs to
   xe. xs and xe are expressed in character coordinates in DrawMulticolorLine, DrawMonoLine and
   in DrawBackgroundLine, in pixel coordinates in DrawSprites.

 */

INLINE void
DrawBgLine (PIXEL * p, LineStruct * l, int xs, int xe, PIXEL * colors)
{
  int i;
  BYTE *colorp = &(l->bgdata[0]);

  p += xs * CHAR_XSIZE;
  for (i = xs; i <= xe; i++, p += CHAR_XSIZE)
    p[0] = p[1] = p[2] = p[3] = p[4] = p[5] = p[6] = p[7] =
#if defined( DOUBLE_X ) || defined( TRIPLE_X ) || defined( QUAD_X )
      p[8] = p[9] = p[10] = p[11] = p[12] = p[13] = p[14] = p[15] =
#endif
#if defined( TRIPLE_X ) || defined( QUAD_X )
      p[16] = p[17] = p[18] = p[19] = p[20] = p[21] = p[22] = p[23] =
#endif
#if defined( QUAD_X )
      p[24] = p[25] = p[26] = p[27] = p[28] = p[29] = p[30] = p[31] =
#endif
      colors[colorp[i]];
}

static void
DrawHiresLine (PIXEL * p, LineStruct * l, int xs, int xe, PIXEL * colors)
{
  BYTE *colorp, *pixelp;
  int i;

  colorp = &(l->colordata1[0]);
  pixelp = &(l->fgdata[0]);
  p += xs * CHAR_XSIZE;
  for (i = xs; i <= xe; i++, p += CHAR_XSIZE)
    {
      register BYTE byte = pixelp[i];
      register PIXEL fg = colors[colorp[i]];
      PUTMONO (byte, fg);
    }
}

static void
DrawMulticolorLine (PIXEL * p, LineStruct * l, int xs, int xe, PIXEL * colors)
{
  BYTE *colorp1, *colorp2, *colorp3, *pixelp;
  int i;

  colorp1 = &(l->colordata1[0]);
  colorp2 = &(l->colordata2[0]);
  colorp3 = &(l->colordata3[0]);
  pixelp = &(l->fgdata[0]);
  p += xs * CHAR_XSIZE;
  for (i = xs; i <= xe; i++, p += CHAR_XSIZE)
    {
      register PIXEL c1 = colors[colorp1[i]];
      register PIXEL c2 = colors[colorp2[i]];
      register PIXEL c3 = colors[colorp3[i]];
      register BYTE byte = pixelp[i];
      PUTMULTI (byte, c1, c2, c3);
    }
}

INLINE int
DrawSprites (PIXEL * pp, SpriteLineStruct * s, int pr, int n, int xs, int xe, PIXEL * colors)
{
  register int i, size;
  SpriteLineStruct spr;
  PIXEL *p;
  int numdrawn = 0;

  for (i = n - 1; i >= 0; i--)
    {
      spr = s[i];
      if (spr.visible)
	{
	  numdrawn++;
	  size = s[i].xexp ? 24 : 48;
	  if ((spr.x + size) >= (xs) && (spr.x + size) <= (xe) && (spr.priority == (pr)))
	    {
	      p = &(pp[spr.x * PIXEL_XSIZE]);
	      if (spr.multicolor)
		{
		  register PIXEL c1 = colors[spr.c1];
		  register PIXEL c2 = colors[spr.c3];
		  register PIXEL c3 = colors[spr.c2];
		  register BYTE byte = spr.data[0];
		  if (spr.xexp)
		    {
		      PUTMULTI2x (byte, c1, c2, c3);
		      byte = spr.data[1];
		      p += 2 * CHAR_XSIZE;
		      PUTMULTI2x (byte, c1, c2, c3);
		      byte = spr.data[2];
		      p += 2 * CHAR_XSIZE;
		      PUTMULTI2x (byte, c1, c2, c3);
		    }
		  else
		    {
		      PUTMULTI (byte, c1, c2, c3);
		      byte = spr.data[1];
		      p += CHAR_XSIZE;
		      PUTMULTI (byte, c1, c2, c3);
		      byte = spr.data[2];
		      p += CHAR_XSIZE;
		      PUTMULTI (byte, c1, c2, c3);
		    }
		}
	      else
		{
		  register PIXEL fg = colors[spr.c3];
		  register BYTE byte = spr.data[0];
		  if (spr.xexp)
		    {
		      PUTMONO2x (byte, fg);
		      byte = spr.data[1];
		      p += 2 * CHAR_XSIZE;
		      PUTMONO2x (byte, fg);
		      byte = spr.data[2];
		      p += 2 * CHAR_XSIZE;
		      PUTMONO2x (byte, fg);
		    }
		  else
		    {
		      PUTMONO (byte, fg);
		      byte = spr.data[1];
		      p += CHAR_XSIZE;
		      PUTMONO (byte, fg);
		      byte = spr.data[2];
		      p += CHAR_XSIZE;
		      PUTMONO (byte, fg);
		    }
		}
	    }
	}
    }
  return numdrawn;
}


INLINE int
DoSpriteDma (VideoStruct * v)
{
  int i, d;
  int cycles = 0;
  SpriteStruct *s = v->sprites;

  for (i = 0; i < v->maxsprites; i++)
    {
      if ((s[i].y == (v->rasterline & 0xff)) && s[i].visible)
	{
	  s[i].dma_flag = 1;
	  s[i].memptr = 0;
	  s[i].memptr_inc = 3;
	  cycles += 2;
	}
      else
	{
	  if (s[i].yexp)
	    s[i].memptr_inc = (s[i].memptr_inc == 3) ? 0 : 3;
	  if (s[i].dma_flag)
	    {
	      if ((s[i].memptr += s[i].memptr_inc) >= 63)
		s[i].dma_flag = 0;
	      else
		cycles += 2;
	    }
	}
    }
  return cycles;
}


#if defined( NEEDS_NullDrawFunct)
static void
NullDrawFunct (VideoStruct * v, PIXEL * p, int xs, int xe)
{
}
#endif


/*

   SetVideoModes(): allocates the videomodes and sets the functions for foreground and background
   screen data. Returns 0 on success.

 */

static int
SetVideoModes (VideoStruct * v, int n,...)
{
  va_list l;
  int i;

  va_start (l, n);

  v->numvmodes = n;
  if (!(v->vmodedesc = (VideoModeStruct *) malloc (sizeof (VideoModeStruct) * n)))
    return 1;
  for (i = 0; i < n; i++)
    {
      v->vmodedesc[i].getfgdata = va_arg (l, datafunct *);
      v->vmodedesc[i].getcolordata = va_arg (l, datafunct *);
      v->vmodedesc[i].getbgdata = va_arg (l, datafunct *);
      switch (va_arg (l, int))
	{
	case MULTICOLOR_VMODE:
	  v->vmodedesc[i].drawline = DrawMulticolorLine;
	  break;
	case HIRES_VMODE:
	  v->vmodedesc[i].drawline = DrawHiresLine;
	  break;
	case SPECIAL_VMODE:
	  v->vmodedesc[i].drawline = va_arg (l, drawfunct *);
	  break;
	default:
	  return 1;
	}
    }
  va_end (l);
  return 0;
}


/*

   DisplaySpeed() displays the average speed rate in the window title bar.

 */

static void
DisplaySpeed (VideoStruct * v, double value)
{
  static char speedstring[80];
  int intvalue = (int) value;

  if ((value - intvalue) > .5)
    intvalue++;
  sprintf (speedstring, "%s at %d%% speed", v->wintitle, intvalue);
  XSetStandardProperties (display, v->win, speedstring, NULL, None, NULL, 0, NULL);
}


/*

   AddRefreshLine() updates a RefreshStruct so that it includes line y and interval
   [ xs ; xe ] too.

 */

INLINE void
AddRefreshLine (RefreshStruct * refstruct, int y, int xs, int xe)
{
  int num_elem;

  num_elem = refstruct->num_elem;

#ifdef RASTER_DEBUG_PUTIMAGE_CALLS
  printf ("AddRefreshLine(): y = %d, xs = %d, xe = %d\n", y, xs, xe);
#endif

  if (!num_elem)
    {
      refstruct->ys = refstruct->ye = y;
      refstruct->xs = xs;
      refstruct->xe = xe;
      refstruct->num_elem = 1;
    }
  else
    {
      refstruct->xs = xs < refstruct->xs ? xs : refstruct->xs;
      refstruct->xe = xe < refstruct->xe ? refstruct->xe : xe;
      refstruct->ye = y;
    }

#ifdef RASTER_DEBUG_PUTIMAGE_CALLS
  printf ("AddRefreshLine(): current refstruct has ys = %d, ye = %d, xs = %d, xe = %d\n",
	  refstruct->ys, refstruct->ye, refstruct->xs, refstruct->xe);
#endif
}


/*

   DoRefresh() makes the window refresh according to refstruct.

 */

INLINE void
DoRefresh (VideoStruct * v)
{
  RefreshStruct r = v->refstruct;
  int x, y, xx, yy;
  int w, h;

  if (!r.num_elem)
    return;

  x = r.xs;
  y = r.ys;
  xx = r.xs - v->xoffs;
  yy = r.ys - v->fscanline;
  w = r.xe - r.xs + 1;
  h = r.ye - r.ys + 1;
  if (xx < 0)
    {
      x -= xx;
      w += xx;
      xx = 0;
    }
  if (yy < 0)
    {
      y -= yy;
      h += yy;
      yy = 0;
    }

#if PIXEL_XSIZE > 1  /* (for silly compilers) */
  x *= PIXEL_XSIZE;
  w *= PIXEL_XSIZE;
  xx *= PIXEL_XSIZE;
#endif
#if PIXEL_YSIZE > 1  /* (for silly compilers) */
  yy *= PIXEL_YSIZE;
  y *= PIXEL_YSIZE;
  h *= PIXEL_YSIZE;
#endif

#if defined( MITSHM )

  if (use_mitshm)
    {
#if defined( RASTER_DEBUG_PUTIMAGE_CALLS )
      printf ("XShmPutImage %d %d %d %d %d %d \n", x, y, xx, yy, w, h);
#endif
      XShmPutImage (display, v->win, gc, v->scrimg, x, y, xx, yy, w, h, False);
      XFlush (display);
    }
  else
#endif
    {
#if defined( RASTER_DEBUG_PUTIMAGE_CALLS )
      printf ("XPutImage %d %d %d %d %d %d \n", x, y, xx, yy, w, h);
#endif
      XPutImage (display, v->win, gc, v->scrimg, x, y, xx, yy, w, h);
      XFlush (display);
    }

  v->refstruct.num_elem = 0;
}


/*

   RefreshAll() provides unconditioned global refresh

 */

static void
RefreshAll (VideoStruct * v)
{
#if defined( MITSHM )
  if (use_mitshm)
    {
#if defined( RASTER_DEBUG_PUTIMAGE_CALLS )
      printf ("(total)XShmPutImage %d %d %d %d %d %d \n",
	    v->xoffs * PIXEL_XSIZE, v->fscanline * PIXEL_YSIZE, 0, 0, v->winwidth, v->winheight);
#endif
      XShmPutImage (v->display, v->win, gc, v->scrimg, v->xoffs * PIXEL_XSIZE,
		    v->fscanline * PIXEL_YSIZE, 0, 0, v->winwidth, v->winheight,
		    False);
    }
  else
#endif
    {
#if defined( RASTER_DEBUG_PUTIMAGE_CALLS )
      printf ("(total) XPutImage %d %d %d %d %d %d \n", v->xoffs,
	      v->fscanline, 0, 0, v->winwidth, v->winheight);
#endif
      XPutImage (display, v->win, gc, v->scrimg, v->xoffs * PIXEL_XSIZE, v->fscanline *
		 PIXEL_YSIZE, 0, 0, v->winwidth, v->winheight);
    }
}


/*

   Raster() actually emulates the raster drawing a line. It decides whether it's time to
   re-calculate a line from the value of curr_clk; returns -1 if no new line has been
   calculated (and the raster counter incremented), the number of the raster line otherwise.

 */

extern CLOCK clk;

INLINE int
Raster (VideoStruct * v)
{
  PIXEL *pos;
  int line, offs, xs, xe, xs1, xe1, r, rr, xs2, xe2, cl;
  LineStruct *l;
  VideoModeStruct vmd;
  PIXEL *colors = v->colors;
  static int blank = 0;

  v->oldclk += CYCLES_PER_LINE;

  if (v->maxsprites)
    clk += DoSpriteDma (v);

#if defined (__VIC_II__)
  if (v->rasterline >= 0x30)
    {
      if (v->rasterline == 0x30)
	blank = v->blank;
      /* ``Bad line'' emulation:
       * As far as I know, the 1st algorithm should the right one...
       * The PC64 Algorithm #1 is not implemented since I haven't seen any 
       * programs that actually need it.
       */
#if 1 /* PC64 Algorithm #2 */
      v->ycounter++;
      if ((v->rasterline & 7) == (v->ysmooth + 3) && v->rasterline <= 0xf7 &&
	  v->memptr < 960 && !blank)
	{
	  if (v->ycounter > 7)
	    v->memptr += v->tcols;
	  memcpy (v->vbuf, v->screenmem + v->memptr, 40);
	  clk += v->tcols /* + 3 */ ;
	  v->ycounter = 0;
	}
#else /* PC64 Algorithm #3 */
      v->ycounter++;
      if (v->ycounter == 8)
	v->memptr += 40;
      if ((v->rasterline & 7) == (v->ysmooth + 3) && v->rasterline <= 0xf7)
	{
	  if ((v->rasterline >> 3) == 6)
	    v->memptr = 0;
	  v->ycounter = 0;
	  memcpy (v->vbuf, v->screenmem + v->memptr, 40);
	  clk += v->tcols + 3;
	}
#endif
    }
#endif /* defined (__VIC_II__) */

  if (!v->refcount && !v->asleep)
    {
      if (v->rasterline >= v->fscanline && v->rasterline <= v->lscanline)
	{
	  l = &(v->lineinfo[v->rasterline]);
	  offs = v->bordwidth + v->xsmooth;
	  pos = v->scrp + offs * PIXEL_XSIZE;
	  xs = xs1 = v->scrcols;
	  r = 0;
	  xe = xe1 = 0;
	  vmd = v->vmodedesc[v->vmode];
	  if (((v->rasterline <= (v->bordheight + v->brdgrow_yu) ||
		v->rasterline > (v->bordheight + v->ypix - v->brdgrow_yl)) &&
	       !v->openvborders) || blank)
	    {						   /* blank line */
	      if (!l->blank || l->bordcol != v->bordcol)
		{
		  l->blank = 1;
		  l->bordcol = v->bordcol;
		  DrawBlankLine (v, v->scrp, (xs = 0), (xe = v->scrcols - 1), colors);
		  AddRefreshLine (&(v->refstruct), v->rasterline, xs, xe);
#if defined (DOUBLE_Y)
		  memcpy (v->scrp + v->scrimg->bytes_per_line, v->scrp, v->scrimg->bytes_per_line);
#elif defined (TRIPLE_Y)
		  memcpy (v->scrp + v->scrimg->bytes_per_line, v->scrp, v->scrimg->bytes_per_line);
		  memcpy (v->scrp + v->scrimg->bytes_per_line * 2, v->scrp,
			  v->scrimg->bytes_per_line);
#endif
		}
	    }
	  else
	    {
	      line = v->rasterline - v->bordheight - v->ysmooth - 1;
	      if (l->n != line || l->xsmooth != v->xsmooth || l->vmode != v->vmode || l->blank)
		{
		  r = 1;
		  l->blank = 0;
		  l->n = line;
		  l->xsmooth = v->xsmooth;
		  l->vmode = v->vmode;
		  xs = 0;
		  xe = v->scrcols - 1;
		  if (v->maxsprites)
		    GetSpriteData (v, l, &xs, &xe);
		  vmd.getfgdata (v, l, &xs1, &xe1, 1);
		  vmd.getbgdata (v, l, &xs1, &xe1, 1);
		  vmd.getcolordata (v, l, &xs1, &xe1, 1);
		  l->bordcol = v->bordcol;
		  xs1 = 0;
		  xe1 = v->tcols - 1;
		}
	      else
		{
		  if (v->maxsprites)
		    rr = r = GetSpriteData (v, l, &xs, &xe);
		  else
		    rr = r = 0;
		  r |= vmd.getfgdata (v, l, &xs1, &xe1, rr);
		  r |= vmd.getbgdata (v, l, &xs1, &xe1, rr);
		  r |= vmd.getcolordata (v, l, &xs1, &xe1, rr);
		  if (l->bordcol != v->bordcol || l->brdgrow_xl != v->brdgrow_xl ||
		      l->brdgrow_xr != v->brdgrow_xr)
		    {
		      l->bordcol = v->bordcol;
		      xs = 0;
		      xe = v->scrcols - 1;
		      r = 1;
		      if (l->brdgrow_xl != v->brdgrow_xl || l->brdgrow_xr != v->brdgrow_xr)
			{
			  l->brdgrow_xl = v->brdgrow_xl;
			  l->brdgrow_xr = v->brdgrow_xr;
			  xs1 = 0;
			  xe1 = v->tcols - 1;
			}
		      r = 1;
		    }
		  else
		    {
		      xs2 = offs + xs1 * 8;
		      xe2 = offs + (xe1 + 1) * 8 - 1;
		      if (xs < xe)
			{
			  xs = xs < xs2 ? xs : xs2;
			  xe = xe > xe2 ? xe : xe2;
			}
		      else
			{
			  xs = xs2;
			  xe = xe2;
			}
		    }
		}

	      if (r)   /* this line needs to be updated... */
		{
		  AddRefreshLine (&(v->refstruct), v->rasterline, xs, xe);
		  DrawBgLine (pos, l, xs1, xe1, colors);
		  if (v->xsmooth > 0 && xs <= v->bordwidth + v->xsmooth)
		    memset (v->scrp + v->bordwidth * PIXEL_XSIZE,
			    colors[(int) v->bgcol], v->xsmooth * PIXEL_XSIZE);
		  DrawSprites (v->scrp, l->sprites, 1, v->maxsprites, 0, v->scrcols - 1, colors);

#if defined (__VIC_II__)
		  if (v->ycounter > 7)
		    DrawHiresLine (pos, l, xs1, xe1, colors);
		  else
#endif
		    vmd.drawline (pos, l, xs1, xe1, colors);
		  DrawSprites (v->scrp, l->sprites, 0, v->maxsprites, 0, v->scrcols - 1, colors);
		  DrawBlankLine (v, v->scrp, 0, v->bordwidth + v->brdgrow_xl - 1, colors);
		  DrawBlankLine (v, v->scrp, v->bordwidth + v->xpix - v->brdgrow_xr,
				 xe = v->scrcols - 1, colors);
#if PIXEL_YSIZE > 1 && !defined (VINTAGE_DISPLAY)
		  {
		    int j = 1, k = (xe - xs) * PIXEL_XSIZE;
		    PIXEL *src = v->scrp + xs * PIXEL_XSIZE;
		    PIXEL *dest = v->scrp + v->scrimg->bytes_per_line + xs * PIXEL_XSIZE;

		    for (; j < PIXEL_YSIZE; j++, dest += v->scrimg->bytes_per_line)
		      memcpy (dest, src, k);
		  }
#endif
		} /* if (r) */

#if !defined (__VIC_II__) && !defined (__VDC__)

	      /*
	         I don't know how bad lines are handled by the other video chips, so I
	         just make the indexes work in the most obvious way...
	         The VDC uses the old style memory fetching for now, so ycounter and
	         memptr are just not used.
	       */

	      if ((++v->ycounter) > 7)
		{
		  v->ycounter = 0;
		  v->memptr += v->tcols;
		  /* The following should be correct for the VIC-II */
		  /* But how many cycles do CRTC and VIC-I need? */
		  clk += v->tcols + 3;
		}
#endif
	    }
	}
    }

  v->scrp += v->scrimg->bytes_per_line * PIXEL_YSIZE;

  cl = v->rasterline;
  v->rasterline++;

  if (cl == (v->bordheight + v->ypix - 4))
    v->openvborders = 0;

  if (v->rasterline == v->scrlines)
    {
      if (havetosleep && v->speed_index && !v->asleep)
	sigpause (0);
      havetosleep = 1;
      if (!v->refcount)
	{
	  DoRefresh (v);
	  v->refcount = v->refrate - 1;
	}
      else
	v->refcount--;
      if ((++v->scount) == (v->refrate * SCR_PER_TCALC))
	{
	  double deltatime, framesec, speedindex;
	  struct timeval tv;
	  gettimeofday (&tv, NULL);
	  deltatime = -v->time;
	  v->time = tvtodouble (tv);
	  deltatime += v->time;
	  framesec = ((double) (v->refrate * SCR_PER_TCALC)) / deltatime;
	  speedindex = (double) framesec / ((double) RFSH_PER_SEC) * 100.0;
	  if (!v->beenpaused)
	    DisplaySpeed (v, speedindex);
	  else
	    v->beenpaused = 0;
	  v->scount = 0;
	}
      else if (!v->scount)
	{
	  struct timeval tv;
	  gettimeofday (&tv, NULL);
	  v->time = tvtodouble (tv);
	}
      v->scrp = (PIXEL *) v->scrimg->data;
      v->rasterline = 0;
#if defined (__VIC_II__)
      v->memptr = -v->tcols;
      v->ycounter = 8;
#else
      v->memptr = 0;
      v->ycounter = 0;
#endif
      blank = v->blank;
    }
  return cl;
}
