/*
 * $Id: sprcoll.c,v 1.2 1996/07/29 22:51:14 ettore Exp ettore $
 *
 * This file contains the code required for sprite-sprite and sprite-background
 * collision register emulation.
 *
 *
 * Written by
 *   Ettore Perazzoli (ettore@comm2000.it)
 *
 *
 * $Log: sprcoll.c,v $
 * Revision 1.2  1996/07/29 22:51:14  ettore
 * Much faster implementation.
 *
 * Revision 1.1  1996/04/01 08:25:07  jopi
 * Initial revision
 *
 *
 */


#include <X11/Xlib.h>
#include <stdio.h>
#include <stdlib.h>

#include "config.h"
#include "vicii.h"
#include "sprcoll.h"


static int sscoll_table [1 << NUM_SPRITES][1 << (NUM_SPRITES + 1)];
static sprmask_t multi_convtable [256]; 	/* Multicolor -> Mono */
static sprmask_t exp_convtable [256][2];     	/* Normal -> doubled */


/*

   some functions for debugging pourposes only...

 */

void
print_binary (BYTE value)
{
  BYTE msk;

  for (msk = 1 << 7; msk; msk >>= 1)
    putchar (value & msk ? 'H' : '-');
}

void
print_huge_mask (sprmask_t * mask, int size)
{
  int i;

  for (i = 0; i < size; i++)
    print_binary (mask[i]);
}


/* ------------------------------------------------------------------------------------------ */

static void
double_mask (sprmask_t * sprmask, int size)
{
  int i, j;
  sprmask_t src[MASK_SIZE], srcmsk, destmsk;

  for (i = 0; i < size; i++)
    src[i] = sprmask[i];

  for (j = -1, i = 0; i < size; i++)
    {
      sprmask[++j] = 0;
      for (destmsk = 3 << 6, srcmsk = 1 << 7; srcmsk > 8; srcmsk >>= 1, destmsk >>= 2)
	if (src[i] & srcmsk)
	  sprmask[j] |= destmsk;
      sprmask[++j] = 0;
      for (destmsk = 3 << 6; srcmsk; srcmsk >>= 1, destmsk >>= 2)
	if (src[i] & srcmsk)
	  sprmask[j] |= destmsk;
    }
}


/*

   convert_multi () transforms the mask so that each couple of bits is turned into a couple
   of '1' bits when it represents a coloured pixel in multicolor mode, allowing correct
   sprite-sprite collision with multicolor sprites, too.

 */

static void
convert_multi (sprmask_t * dest, BYTE * src, int size)
{
  int i, msk;

  for (i = 0; i < size; i++, dest++, src++)
    {
      *dest = 0;
      for (msk = 3 << 6; msk; msk >>= 2)
	if (*src & msk)
	  *dest |= msk;
    }
}


void
init_sprcoll (void)
{
  int i, msk1, msk2, s1, s2, n;

  for (i = 0; i < (1 << NUM_SPRITES); i++)
    {
      for (n = 0, msk1 = 1 << (NUM_SPRITES - 1), s1 = NUM_SPRITES - 1; s1 >= 1; s1--, msk1 >>= 1)
	for (msk2 = msk1 >> 1, s2 = s1 - 1; s2 >= 0; s2--, msk2 >>= 1)
	  if ((i & msk1) && (i & msk2))
	    {
	      sscoll_table[i][n] = s1;
	      sscoll_table[i][n + 1] = s2;
	      n += 2;
	    }
      sscoll_table[i][n] = -1;
    }
  /* Build the conversion tables */
  for (i = 0; i <= 0xff; i++)
    {
      sprmask_t k = (sprmask_t)i;
      convert_multi (&(multi_convtable[i]), &k, 1);
      exp_convtable[i][0] = k;
      double_mask (exp_convtable[i], 1);
    }
}
	
    
INLINE void
build_mask (BYTE * src, int sprsize, sprmask_t * sprmask_return, int sh_right)
{
  int i;
  BYTE tempmask, savemask;
  int sh_left = 8 - sh_right;

  savemask = 0xff >> sh_left;
  tempmask = 0;
  for (i = 0; i < sprsize; i++, src++, sprmask_return++)
    {
      *sprmask_return = tempmask | (*src >> sh_right);
      tempmask = (*src & savemask) << sh_left;
    }
  *sprmask_return = tempmask;
}


/*
   
   init_sprite_masks () initializes the sprite structures with the proper startcol, endcol, size
   and ready (0 for now) values.
   
 */

void
init_sprite_masks (VideoStruct * v, LineStruct * l, sprmask_struct * sprmask_return)
{
  int i;
  SpriteLineStruct * s = l->sprites;
  
  for (i = 0; i < NUM_SPRITES; i++, s++, sprmask_return++)
    if (s->visible)
      {
	sprmask_return->relx = s->x - (v->bordwidth + l->xsmooth);
	if (sprmask_return->relx > 0)
	  sprmask_return->startcol = sprmask_return->relx >> 3;
	else
	  sprmask_return->startcol = -((-sprmask_return->relx) >> 3) - 1;
	sprmask_return->size = s->xexp ? 8 : 4;
	sprmask_return->endcol = sprmask_return->startcol + sprmask_return->size - 1;
	sprmask_return->ready = 0;
	sprmask_return->visible = 1;
      }
    else
      sprmask_return->visible = 0;
}


/*

   build_sprite_mask () builds the mask for the specified sprite.

 */

INLINE void
build_sprite_mask (SpriteLineStruct * s, sprmask_struct * sprmask_return)
{
  if (sprmask_return->ready)
    return;			/* the mask has already been built */
   
  if ((sprmask_return->visible = s->visible))
    {
      sprmask_t tmpmsk[6];
      if (s->multicolor)
	{
	  int z;
	  for (z = 0; z < 3; z++)
	    tmpmsk[z] = multi_convtable[(int)s->data[z]];
	}
      else
	{
	  tmpmsk[0] = s->data[0];
	  tmpmsk[1] = s->data[1];
	  tmpmsk[2] = s->data[2];
	}
      if (s->xexp)
	{
	  int z;
	  tmpmsk[4] = tmpmsk[2];
	  tmpmsk[2] = tmpmsk[1];
	  for (z = 0; z < 6; z += 2)
	    {
	      sprmask_t t = tmpmsk[z];
	      tmpmsk[z] = exp_convtable[(int)t][0];
	      tmpmsk[z+1] = exp_convtable[(int)t][1];
	    }
	}
      build_mask (tmpmsk, s->xexp ? 6 : 3, sprmask_return->mask,
		  sprmask_return->relx < 0 ? ((-sprmask_return->relx) % 8) :
		  (sprmask_return->relx % 8));
    }

  sprmask_return->ready = 1;
  return;
}


/*

   check_ss_collisions () checks the sprite-sprite collisions.

 */

void
check_ss_collisions (LineStruct * l, sprmask_struct * s, sprmask_t * ss_collmask_return)
{
  int i, j, k, msk1, msk2, length;
  sprmask_t *p1, *p2;
  int mask, n;
  
  mask = l->sprmask;
	
  for (n = 0; sscoll_table[mask][n] != -1; n += 2)
    {
      i = sscoll_table[mask][n];
      j = sscoll_table[mask][n + 1];
      msk1 = 1 << i;
      msk2 = 1 << j;
      if (!(*ss_collmask_return & (msk1 | msk2)))
	{
 	  if (s[j].startcol <= s[i].endcol && s[i].endcol <= s[j].endcol)
	    {
	      length = s[i].endcol - s[j].startcol + 1;
	      p1 = s[j].mask;
	      p2 = s[i].mask + s[j].startcol - s[i].startcol;
	    }
	  else if (s[j].startcol <= s[i].startcol && s[i].startcol <= s[j].endcol)
	    {
	      length = s[j].endcol - s[i].startcol + 1;
	      p1 = s[i].mask;
	      p2 = s[j].mask + s[i].startcol - s[j].startcol;
	    }
	  else
	    continue;
	  build_sprite_mask (&(l->sprites[i]), &s[i]);
	  build_sprite_mask (&(l->sprites[j]), &s[j]);
	  for (k = 0; k < length; k++, p1++, p2++)
	    if (*p1 & *p2)
	      {
		*ss_collmask_return |= msk1 | msk2;
		break;
	      }
	}
    }
}


/*

   check_sb_collisions () checks the sprite-to-background collisions.
   
 */


void
check_sb_collisions (VideoStruct * v, LineStruct * l, sprmask_struct * sprmasks,
		     sprmask_t * sb_collmask_return)
{
  int i, n, s, j, x;
  sprcollmask_t msk;
  BYTE *data;

  if ((v->rasterline < (v->bordheight + v->ysmooth)) ||
      (v->rasterline >= (v->bordheight + v->ypix)))
    return;

  data = l->fgdata;
  for (i = 0, msk = 1; i < 8; i++, sprmasks++, msk <<= 1)
    if (sprmasks->visible && sprmasks->endcol >= 0 && sprmasks->startcol < v->tcols &&
	!(*sb_collmask_return & msk))
      {
	build_sprite_mask (&(l->sprites[i]), sprmasks);
	n = sprmasks->size;
	s = 0;
	if (sprmasks->startcol < 0)
	  {
	    n += sprmasks->startcol;
	    s -= sprmasks->startcol;
	    x = 0;
	  }
	else
	  x = sprmasks->startcol;
	if (sprmasks->endcol >= v->tcols)
	  n -= sprmasks->endcol - (v->tcols - 1);
	for (j = 0; j < n; j++, s++, x++)
	  if (data[x] & sprmasks->mask[s])
	    {
	      *sb_collmask_return |= msk;
	      break;
	    }
      }
}
