/*
 * $Id: symbol.c,v 1.2 1994/12/12 16:59:44 jopi Exp jopi $
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README for copyright notice
 *
 * This file contains functions to handle assembly source code.
 * 
 * Written by 
 *   Jouko Valta (jopi@stekt.oulu.fi)
 *
 * $Log: symbol.c,v $
 * Revision 1.2  1994/12/12  16:59:44  jopi
 * Compability
 *
 * Revision 1.1  1994/08/30  15:22:05  jopi
 * Symbol handling routines from an old version of asm.c.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>
#include <assert.h>  /* only for convenience */

#include "asm.h"
#include "extern.h"


int    asm_init(char *file);
int    asm_command(char *line, int mode);
static int asm_directives(ADDRESS address, char *line, int mode);
static int asm_add_symbol(ADDRESS address, char *line, int mode, int pass);
ADDRESS  asm_eval_argument(/* symtab, int cnt, */ char *s, int mode);
/*int     parse_argstr(char *, char *, int, int, int);*/

typedef struct {
  ADDRESS       val;
  unsigned char type;
  char          name[SYMBOL_MAX_CHARS +1];	/* default 7 characters */
} LABEL;


/*
 * To avoid unnecessary memory mismanagement the value of
 * labeltop is never decreased.
 */

static int labeltop = 0;
static LABEL *symbol = 0;
static int labelcnt =0;


/*
 * This function implements a simple MOS6502 cross-assembler.
 * The interpreter selects default base according to the
 * value of hexflg whereas assembler always uses decimal
 * as default for files.
 */

int    asm_init(char *file)
{
    FILE *fopen(), *infile;
    char  line[256];
    int   err, errors = 0;
    int   mode = 0;
    ADDRESS addr = 0;
    int level = 0;  /* include level */
    int pass = 0;

    /*
     * For better reliability realloc() is not used at the first time.
     */

    if (symbol == NULL) {
	labeltop = 64;
	symbol = (LABEL *)malloc(labeltop * sizeof(LABEL));
	assert(symbol);
    }

    if (*file) {
	printf("Loading source file %s\n", file);
	if (NULL == (infile = fopen(file, "r"))) {
	    printf("Loading failed.\n");
	    perror(file);
	}
	else {
	    printf("Assembler\n");
	    mode = MODE_ASM | MODE_INF | MODE_SYMBOL;   /* hexmode off */

	    /* clear symtab */
	    for (labelcnt = labeltop ;labelcnt >0; labelcnt--)
		symbol[labelcnt].type = 0;  /* Note: label 0 is always 'EQU' */
	}
    }
    else {
	printf ("No source file\n");
	return (-1);
    }

    for (;;) {
	/*line = read_line(addr, mode & MODE_INF);*/

	if (fgets(line, 255, infile) == NULL) {
	    if (mode & MODE_INF) {
		mode &= MODE_SYMBOL;
		level--;
	    }
	    else  break;
	}
	else {
	    if (((mode & MODE_INF) || asm_command(line, mode) ) &&
		(err = interpret_line(line, mode)) <= 0) {
		printf("**** %.4X SYNTAX ERROR (%d): %s ****\n", addr, err, line);
		errors++;
	    }
	}
    } /* for */

    printf("%2d errors.\n", errors);
    return (errors);
}


/*
 * Direct mode commands
 */

int    asm_command(char *line, int mode)
{
    static char asmfile[64];
    char buf[SYMBOL_MAX_CHARS+1];

    if (0 == line[0]) return 0;  /* No input */

    if (strchr("=%+#&$", line[0]) && line[1] ) {	/* Base conversion. */
	if(line[0] == '=') line++;

	/* labels enabled */
	show_bases(line, mode | MODE_SYMBOL );		/* Base conversions. */

    } else if ('?' == line[0]) {
	printf("\n ? %% # + & $ *= =  ASM  EQU  SYM  END\n");

    } else if (0 == strncasecmp("ASM", line, 3)) {
	if (line[3] != 0)
	    sscanf( line+3, "%s", asmfile);   /* save filename for later use */
	asm_init(asmfile);
    }

    else if (0 == strncasecmp("SYM", line, 3)) {
	int i;

	if (sscanf(line+3, "%s%n", buf, &i) < 1) {
	    for (i = 0; i < labelcnt;)
		printf("%.4X  %-7s%s",
		       symbol[i].val & 0xffff,
		       symbol[i].name, (++i % 4) ? "\t" : "\n");

	    printf("\n %d labels.\n\n", labelcnt);
	}
/*
	else if (i == 1 && *buf == '+') mode |= MODE_SYMBOL;
	else if (i == 1 && *buf == '-') mode &= ~MODE_SYMBOL;
*/
	else for (i = 0; i < labelcnt; i++)
	    if (!strcasecmp(buf, symbol[i].name)) {

		if (symbol[i].type & SYMBOL_SET)
		    printf("  %.4X\n\n", symbol[i].val & 0xffff);
		else
		    printf("Undefined symbol\n");

		break;
	    }
    }
    else if (0 == strncasecmp("END", line, 3)) {

    }
    else return(1);  /* command not found */
    return(0);      /* ready */
}


/*
 * This routine handtes all directives in non.interactive mode.
 * All directives begin with '.', except for '*' and 'EQU'.
 */

static int asm_directives(ADDRESS address, char *line, int mode)
{
    int     i = 0;
    int     len = 0;
    static int skip = 0;

    while (*line && isspace(*line)) line++;

    if (0 == strncasecmp(line, ".else", 5)) {   /* sisakkaisissa voi olla skipissa */
	skip ^= 1; line += 5;
	while(*(++line) && isspace(*line));
    }
    else if (('>' == *line) || !strncasecmp(line, ".endif", 6)) {
	skip = 0;
	while(*(++line) && isspace(*line));
    }


/* tama kuuluu interpret_line'en */
    if (skip || *line)
	return(0); /* There is no more input. */


    if (*line == '*'  || !strncasecmp(line, "EQU", 3)) {
	char *p = line;

	if (*p != '*') p += 2;
	while (*(++p) && (*p == '=' || *p == ' ')); 
	address = asm_eval_argument(line, mode);
	printf("Assembling from $%.4X\n", address);

	strcpy(symbol[0].name, "EQU");    /* saved for Monitor */
	symbol[0].val = address;
	symbol[0].type = SYMBOL_WALL;
    }

    /*
     * Second iteration is required when the first word turns out to be a label.
     */

    if ('.' == *line) {	/* ASM directives */

	if (0 == strncasecmp(line+1, "byt", 3)) {
	    if ((*(line += 4) & 0xdf) == 'E') line++;
	    len = parse_argstr(line, &ram[address], 1, 256, mode);
	    address += len;
	}
	else if (0 == strncasecmp(line+1, "wor", 3)) {
	    if ((*(line += 4) & 0xdf) == 'D') line++;
	    len = parse_argstr(line, &ram[address], 2, 32, mode);
	    address += len;
	}
	else if (0 == strncasecmp(line+1, "ife", 3)) {
	}
	else if (0 == strncasecmp(line+1, "ifn", 3)) {
	}
	else if (0 == strncasecmp(line+1, "end", 3)) {
	    /* Make the programmer happy  (-: */

	}
	else
	    printf("%.4X Warning: Unrecognised directive in %s\n", address, line);
	return(0);
    }

    return(0);

}	/* end of asm_directives */


/*
 * This routine tests for label declarations on program line, and adds it
 * to symbol table.
 */

static int asm_add_symbol(ADDRESS address, char *line, int mode, int pass)
{
    int     i;
    int     len = 0;
    int val;
    char *pb, *p;

    /*
     * It has to be label if it survived this far. So, identify it.
     * First check earlier declarations and create new entry.
     */

    for (pb= p= line, len = 0; (len++ < SYMBOL_MAX_CHARS) && *(++p) > ' '; );

    if (*p > ' ') return E_BAD_IDENTIFIER;

    for (i = 0; i++ < labelcnt;)
	if (!strncasecmp(line, symbol[i].name, len) &&
	    (symbol[i].type & SYMBOL_SET))  {
	    printf("ERROR: symbol redef: %s\n", symbol[i].name);
	    return E_SYMBOL_REDEF;
	}
    if (i >= labeltop) {
	int j;
	symbol =
	    (LABEL *)realloc(symbol, (labeltop += 32) * sizeof(LABEL));
	assert(symbol);
	for(j=i; j < labeltop; j++)
	    symbol[j].val = 0;
    }

    line += len;
    while (*(++line) && isspace(*line));
    if (tstflg)
	printf("LABEL: %s =  --%s--\n", pb, line);

    if ('=' == *line) {
	/* else second try is required */
	val = asm_eval_argument(line+1, mode);
	symbol[i].val = val;
    }
    else
	symbol[i].val = address;

    symbol[i].type= ((val < 0x100) ? SYMBOL_BALL : SYMBOL_WALL);
    sscanf(pb, "%s", symbol[i].name);
    labelcnt++;

    return(0);
}	/* end of asm_add_symbol */


#define BYTL 1
#define BYTH 2

ADDRESS  asm_eval_argument(/* symtab, int cnt, */ char *s, int mode)
{
    int n;
    ADDRESS value;
    int cnt = labelcnt;
    int part = BYTL | BYTH;

    /* identify labels and return its value
     * String is a label if 'mode' is set and the string begins
     * with an alphabetic character
     */

    if (*s == '>') {
	part = BYTH;
	s++;
    }
    else if (*s == '<') {
	part = BYTL;
	s++;
    }

    if ((mode & MODE_SYMBOL) && isalpha(*s)) {
	int found = 0;
	printf("label found %s\n", s);
	for (n = 0; n < cnt; n++)
	    if (strncmp(symbol[n].name, s, SYMBOL_MAX_CHARS)) {
		value = symbol[n].val;

		if (!symbol[n].type)
		    return (E_SYMBOL_UNDEFINED);

		found++;
		break;
	    }
	if (!found)
	    return (E_SYMBOL_UNDEFINED);
    } /* symbol */

    /*
       else
       sconv(...);
       */

    if (tstflg)
	printf("value= %ld\n", value);

    switch (part) {
      case BYTL:
	return (value & 0xff);
      case BYTH:
	return (value>>8 & 0xff);
      default:
	return (value);
    }

#if 0
    if (!cerrno && *s && !isspace(s))
	switch(*s) {
	  case '+':
	  case '-':
	  case '*':
	  case '/':
	  case '(':
	  default: /* error */
	}
#endif

    return (value);
}
