/*
 * $Id: timer.c,v 1.5 1994/06/16 17:19:26 jopi Exp jopi $
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README for copyright notice
 *
 *
 * Written by
 *   Teemu Rantanen (tvr@cs.hut.fi)
 *
 * $Log: timer.c,v $
 * Revision 1.5  1994/06/16  17:19:26  jopi
 * Code reorganized and cleaned up
 *
 * Revision 1.4  1993/11/10  01:55:34  jopi
 * reu, asm and disk directory fixed
 * REL_ADDR macro and 1541 made more portable
 *
 * Revision 1.3  93/06/21  13:39:37  jopi
 *  X64 version 0.2 PL 0
 *
 * Revision 1.2  1993/06/13  08:21:50  sonninen
 * *** empty log message ***
 *
 *
 */

/*
 * timer.c -- Experimental alarm code. Dynamic approach. Should be fast too.
 */

#include <stdio.h>
#include <stdlib.h>

#include "macro.h"
#include "extern.h"

#ifdef NEW_TIMER_CODE

#define NR_ALARMS 20

CLOCK next_alarm_clk;

typedef struct timer_slot_s
{
    int			  in_use;	/* alarm in use */
    struct timer_slot_s  *next;		/* next alarm to happen */
    CLOCK		  alarm;	/* When ? */
    void	        (*func)(int);	/* Call func on alarm */
    int			  parameter;	/* Passed to callfunc */
} timer_slot_t;


/*
 * Table of alarms
 */
static timer_slot_t alarms[NR_ALARMS];


/*
 * List of alarms to happen
 */
static timer_slot_t *first_alarm = NULL;


static void cancel_alarm(timer_slot_t *p);


/*
 * Allocate new alarm.
 */

int     new_alarm(void)
{
    static int alarms_given = 0;

    if (alarms_given == NR_ALARMS)
    {
	printf("out of alarms\n");
	exit(1);
    }
    return alarms_given++;
}


/*
 * Update alarm clock
 */

void    find_next_alarm(void)
{
    next_alarm_clk = first_alarm ? first_alarm->alarm : (CLOCK) (-1);
}


/*
 * When alarm happens, handle all alarms
 */

void    alarm_handler(void)
{
    timer_slot_t *p;
    
    /*
     * Call alarms. Note that if callbackfunction sets alarm <= clk,
     * it will be called again.
     */
    while (first_alarm && clk >= first_alarm->alarm)
    {
	p = first_alarm;
	p->in_use = 0;
	first_alarm = p->next;
	(*(p->func))(p->parameter);
    }

    /*
     * Next alarm
     */
    find_next_alarm();
}


/*
 * Set alarm to happen. Note that if set_alarm is not called from
 * alarm-callback, find_next_alarm must be called.
 */

void    set_alarm(int alarm, CLOCK value, void (*func)(int), int parameter)
{
    register timer_slot_t *current;
    register timer_slot_t *p;
    register timer_slot_t *prev;

    current = &alarms[alarm];

    /*
     * Remove from list first. This happens only if alarm is active and time
     * is changed
     */
    if (current->in_use)
    {
	if (value == current->alarm)
	    return;
	cancel_alarm(current);
    }

    current->in_use = 1;
    current->alarm = value;
    current->func = func;
    current->parameter = parameter;
    
    prev = NULL;
    p = first_alarm;

    /*
     * If value is same, call alarms in order they are set
     */
    while(p && (value >= p->alarm))
    {
	prev = p;
	p = p->next;
    }

    if (!prev)
    {
	current->next = first_alarm;
	first_alarm = current;
#if 0
	/*
	 * This slows things down 10%. Remember to call find_next_alarm() if
	 * you set alarms outside of alarm callback
	 */
	find_next_alarm();
#endif
	return;
    }

    current->next = p;
    prev->next = current;
}


/*
 * Remove alarm from alarm queue.
 */

void    clear_alarm(int alarm)
{
    register timer_slot_t *p;

    p = &alarms[alarm];

    if (!p->in_use)
	return;

    cancel_alarm(p);
}


/*
 * Remove alarm from list. This will be called quite seldom.
 */

static void cancel_alarm(timer_slot_t *current)
{
    register timer_slot_t *p = first_alarm;
    register timer_slot_t *prev = NULL;

    /*
     * This is the only way...
     */
    while (p)
    {
	if (!p->in_use)
	{
	    printf("alarm list corrupted\n");
	    abort();
	}

	if (p == current)
	{
	    current->in_use = 0;

	    if (prev)
	    {
		prev->next = p->next;
	    }
	    else
	    {
		first_alarm = p->next;
		find_next_alarm();
	    }
	    return;
	}

	prev = p;
	p = p->next;
    }

    printf("cancel_alarm: alarm not on list\n");
    abort();
}



#ifdef MAINTEST
CLOCK clk;

static int  foo = 0;

void    exit_me(i)
int i;
{
    printf("foo = %d\n", foo);
    exit(i);
}


void    setagain(i)
int i;
{
    foo++;
    set_alarm(i, clk + (4+i), setagain, i);
}

/*
 * gcc -O6 timer.c -o timer -DMAINTEST
 * time ./timer 400
 */
int     main(argc, argv)
     int argc;
     char *argv[];
{
    int i;
    for (i=0; i < NR_ALARMS - 1; i++)
	set_alarm(i, random()%50, setagain, i);

    set_alarm(NR_ALARMS-1, atoi(argv[1])*1024, exit_me, 0);
    find_next_alarm();

    while(1)
    {
	clk = next_alarm_clk;
	alarm_handler();
    }
}
#endif

#endif /* NEW_TIMER_CODE */
