/* $Id: via.c,v 2.1 1996/07/29 23:11:55 fachat Exp ettore $
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README for copyright notice
 *
 *
 * Written by
 *   Vesa-Matti Puro (vmp@lut.fi)
 *   Jarkko Sonninen (sonninen@lut.fi)
 *   Jouko Valta     (jopi@stekt.oulu.fi)
 *   Andre Fachat    (a.fachat@physik.tu-chemnitz.de)
 *
 *
 * $Log: via.c,v $
 * Revision 2.1  1996/07/29 23:11:55  fachat
 * PET sound stuff added.
 *
 * Revision 2.0  1996/04/18 17:12:05  fachat
 * Nearly completely rewritten.
 * All CIA specific stuff removed to have cleaner PET i/o handling.
 *
 * Revision 1.1  1996/04/01 08:25:07  jopi
 * Initial revision
 *
 */


#include <stdio.h>
#include <time.h>

#include "vmachine.h"
#include "extern.h"
#include "via.h"
#include "parallel.h"

#include "cpu.h"
#include "interrupt.h"
#include "config.h"

#ifdef SOUND
#include "sndif.h"
#endif


/* global */

BYTE    via[16];
BYTE    keyarr[KBD_ROWS];

/* local functions */

static int viadebug = 0;

static void update_viata ( int );
static void update_viatb ( int );

/*
 * Local variables
 */

static int viaifr;	/* Interrupt Flag register for via */
static int viaier;	/* Interrupt Flag register for via */

static int viata;	/* value of via timer 1 at last update */
static int viatb;	/* value of via timer 2 at last update */

static CLOCK viatau;	/* time when via timer 1 is updated */
static CLOCK viatbu;	/* time when via timer 2 is updated */


/* ------------------------------------------------------------------------- */
/* VIA 1 */


/* 
 * according to Rockwell, all internal registers are cleared, except
 * for the Timer (1 and 2, counter and latches) and the shift register.
 */ 

void    reset_via(void)
{
    int i;
    if(viadebug) printf("VIA1: reset\n");   
    /* clear registers */ 
    for(i=0;i<4;i++) via[i]=0;
    for(i=11;i<16;i++) via[i]=0;

/* 
 * The following line doesn't appear in the specs...
 * If it's for the PET, it's useless. A PET uses the Rasterline 
 * IRQ.
 */
    /*via[VIA_ACR] = 0x40;*/  /* timer T1 run mode */

    /* set IEC output lines */
    par_set_atn(0);
    par_set_nrfd(0);

    /* disable vice interrupts */
    setirq(I_VIAT1, 0); unsetalarm(I_VIAT1);
    setirq(I_VIAT2, 0); unsetalarm(I_VIAT2);
    setirq(I_VIASR, 0); unsetalarm(I_VIASR);
    setirq(I_VIAFL, 0); unsetalarm(I_VIAFL);

    viaier = 0;
    viaifr = 0;
}


void    store_via(ADDRESS addr, BYTE byte)
{
    if (debugflg)
	printf("store via[%x] %x\n", (int) addr, (int) byte);


    switch (addr) {

	/* these are done with saving the value */

      case VIA_PRA: /* port A */
      case VIA_PRA_NHS: /* port A, no handshake */
      case VIA_DDRA:
      case VIA_DDRB:
    	via[addr] = byte;
	break;

      case VIA_T2LL:
    	via[addr] = byte;
#ifdef SOUND
	snd_pet_T(byte);
#endif
	break;

      case VIA_SR:
    	via[addr] = byte;
#ifdef SOUND
	snd_pet_sample(byte);
#endif
	break;

	/* Port registers */

      case VIA_PRB: /* port B */
    	via[addr] = byte;
	byte &= via[VIA_DDRB];
	par_set_nrfd(!(byte & 0x02));
	par_set_atn(!(byte & 0x04));
	break;

	/* Timers */

      case VIA_T1CL:
      case VIA_T1LL:
	via[VIA_T1LL] = byte;
	break;

      case VIA_T1CH: /* Write timer A high */
	if(pardebug) printf("Write timer A high: %d\n",byte);
    	via[VIA_T1LH] = byte;
	/* load counter with latch value */
	via[VIA_T1CL] = via[VIA_T1LL];
	via[VIA_T1CH] = via[VIA_T1LH];
	/* Clear T1 interrupt */
	viaifr &= ~VIA_IM_T1;
	setirq(I_VIAT1, 0);
	update_viata(1);
	break;

      case VIA_T1LH: /* Write timer A high order latch */
	via[addr] = byte;
	/* Clear T1 interrupt */
	viaifr &= ~VIA_IM_T1;
	setirq(I_VIAT1, 0);
	break;

      case VIA_T2CH: /* Write timer 2 high */
	via[VIA_T2CH] = byte;
	via[VIA_T2CL] = via[VIA_T2LL]; /* bogus, both are identical */
	update_viatb(1);
	/* Clear T2 interrupt */
	viaifr &= ~VIA_IM_T2;
	setirq(I_VIAT2, 0);
	break;

	/* Interrupts */

      case VIA_IFR: /* 6522 Interrupt Flag Register */
	viaifr &= ~byte;
	if(!viaifr & VIA_IM_T1) setirq(I_VIAT1, 0);
	if(!viaifr & VIA_IM_T2) setirq(I_VIAT2, 0);
	break;

      case VIA_IER: /* Interrupt Control Register */
	if (byte & VIA_IM_IRQ) {
	    /* set interrupts */
	    if ((byte & VIA_IM_T1) && (viaifr & VIA_IM_T1)) setirq(I_VIAT1, 1);
	    if ((byte & VIA_IM_T2) && (viaifr & VIA_IM_T2)) setirq(I_VIAT2, 1);
	    viaier |= byte & 0x7f;
	}
	else {
	    /* clear interrupts */
	    if( byte & VIA_IM_T1 ) setirq(I_VIAT1, 0);
	    if( byte & VIA_IM_T2 ) setirq(I_VIAT2, 0);
	    viaier &= ~byte;
	}
	break;

	/* Control */

      case VIA_ACR:
	if(viadebug) 
	  printf("VIA1: write %02x to ACR at PC=%04x, previous value = %02x\n",
			byte, PC, via[VIA_ACR]);
    	via[addr] = byte;
	/* bit 7 timer 1 output to PB7 */
	/* bit 6 timer 1 run mode -- default seems to be continuous */

	/* bit 5 timer 2 count mode */
	if (byte & 32) {
/* TODO */
/*	    update_viatb(0);*/	/* stop timer if mode == 1 */
	}

	/* bit 4, 3, 2 shift register control */
#ifdef SOUND
	snd_pet((byte & 0x3c)==0x10);
#endif

	/* bit 1, 0  latch enable port B and A */
	break;

      case VIA_PCR:
    	via[addr] = byte;
	if(viadebug) printf("VIA1: write %02x to PCR\n",byte);
	/* bit 7, 6, 5  CB2 handshake/interrupt control */
	/* bit 4  CB1 interrupt control */

	/* bit 3, 2, 1  CA2 handshake/interrupt control */
	/* bit 0  CA1 interrupt control */

	crtc_set_char( byte & 2 ); /* hack for switching PET charrom */
	break;

      default:
    	via[addr] = byte;
	break;
    }  /* switch */
}


/* ------------------------------------------------------------------------- */

BYTE    read_via(ADDRESS addr)
{
    if (debugflg)
	printf("read via[%d]\n", addr);

    switch (addr) {

      case VIA_PRA: /* port A */
      case VIA_PRA_NHS: /* port A, no handshake */
	{
	    BYTE    j = 255;
	    /* nothing to read yet */
	    if (debugflg) {
		printf("read port A %d\n", j);
		printf("a: %x b:%x  ca: %x cb: %x joy: %x\n",
		       (int) j, (int) via[VIA_PRB],
		       (int) via[VIA_DDRA], (int) via[VIA_DDRB], joy[2]);
	    }
	    return ((j & ~via[VIA_DDRA]) | (via[VIA_PRA] & via[VIA_DDRA]));
	}

      case VIA_PRB: /* port B */
	{
	    BYTE    j;
	    /* read parallel IEC interface line states */
	    j = 255 - (par_nrfd ? 64:0) - (par_ndac ? 1:0) - (par_dav ? 128:0);
	    /* vertical retrace */
	    j -= crtc_offscreen() ? 32:0;
	    if (0 /*pardebug*/) {
		printf("read port B %d\n", j);
		printf("a: %x b:%x  ca: %x cb: %x joy: %x\n",
		       (int) via[VIA_PRA], (int) j,
		       (int) via[VIA_DDRA], (int) via[VIA_DDRB], joy[1]);
	    }
	    return ((j & ~via[VIA_DDRB]) | (via[VIA_PRB] & via[VIA_DDRB]));
	}

	/* Timers */

      case VIA_T1CL: /* timer A low */
	viaifr &= ~VIA_IM_T1;
	setirq(I_VIAT1, 0);
	return ((viata - clk + viatau) & 0xff);

      case VIA_T1CH: /* timer A high */
	return (((viata - clk + viatau) >> 8) & 0xff);

      case VIA_T2CL: /* timer B low */
	viaifr &= ~VIA_IM_T2;
	setirq(I_VIAT2, 0);
	return ((viatb - clk + viatbu) & 0xff);

      case VIA_T2CH: /* timer B high */
	return (((viatb - clk + viatbu) >> 8) & 0xff);

      case VIA_SR: /* Serial Port Shift Register */
	return (via[addr]);

	/* Interrupts */

      case VIA_IFR: /* Interrupt Flag Register */
	{
	    BYTE    t = viaifr;
	    if (viaifr & viaier /*[VIA_IER]*/)
		t |= 0x80;
	    if(viadebug || pardebug) printf("Read Intflag VIA gives %x\n",t);
	    return (t);
	}

      case VIA_IER: /* 6522 Interrupt Control Register */
	    return (viaier /*[VIA_IER]*/ | 0x80);

      case VIA_ACR:
	    if(viadebug) printf("read ACR gives %02x\n",via[addr]);
	    return via[addr];
    }  /* switch */

    return (via[addr]);
}


/* ------------------------------------------------------------------------- */

int    int_viata(void)
{
    if (debugflg || pardebug)
	printf("via timer A interrupt\n");

    if (!(via[VIA_ACR] & 0x40)) { /* one-shot mode */
	if(viadebug || pardebug) printf("one-shot mode via timer A!\n");
	viata = 0;
	viatau = clk;
	unsetalarm(I_VIAT1); 	/*int_clk[I_VIAT1] = 0;*/
    }
    else {		/* continuous mode */
	/* load counter with latch value */
	via[VIA_T1CL] = via[VIA_T1LL];
	via[VIA_T1CH] = via[VIA_T1LH];
	update_viata(1);
    }
    viaifr |= VIA_IM_T1; 
    if(viaier /*[VIA_IER]*/ & VIA_IM_T1 ) setirq(I_VIAT1, 1);

    return 0; /*(viaier & VIA_IM_T1) ? 1:0;*/
}


/*
 * Timer B is always in one-shot mode
 */


int    int_viatb(void)
{
    if (debugflg)
	printf("via timer B interrupt\n");

    viatb = 0;
    viatbu = clk;
    unsetalarm(I_VIAT2);	/*int_clk[I_VIAT2] = 0;*/

    viaifr |= VIA_IM_T2;
    if( viaier/*[VIA_IER]*/ & VIA_IM_T2 ) setirq(I_VIAT2, 1);

    return 0; /*(viaier & VIA_IM_T2) ? 1:0;*/
}


/* ------------------------------------------------------------------------- */

static void update_viata(int force)
{
    if(force) {
       if(viadebug) 
	  printf("update via timer A : latch=%d, counter =%d, clk = %ld\n",
      		via[VIA_T1CL] + (via[VIA_T1CH] << 8),
      		via[VIA_T1LL] + (via[VIA_T1LH] << 8),
		clk);

      viata = via[VIA_T1CL] + (via[VIA_T1CH] << 8);
      viatau = clk;
      setalarm(I_VIAT1, viata);
    }
}


static void update_viatb(int force)
{
    if(force) {
      viatb = via[VIA_T2CL] + (via[VIA_T2CH] << 8);
      viatbu = clk;
      setalarm(I_VIAT2, viatb);
    }
}

int     show_keyarr(void)
{
    int     i, j;
    for (j = 0; j < KBD_ROWS; j++) {
        printf("%d:", j);
        for (i = 0x80; i; i >>= 1)
            printf(" %d", keyarr[j] & i);
        printf("\n");
    }
    return (0);
}


