/*
 * $Id: vicii.c,v 1.2 1996/07/29 23:29:46 ettore Exp ettore $
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README files for copyright notice.
 *
 *
 * VIC-II video chip emulation.
 *
 *
 * Written by
 *   Ettore Perazzoli (ettore@comm2000.it)
 *
 *
 * $Log: vicii.c,v $
 * Revision 1.2  1996/07/29 23:29:46  ettore
 * Lots of bugs fixed.
 * Better interrupt handling, using setalarm() and setirq().
 * RMW bug implemented in $D019.
 * Uses the new sprite collision checking routines.
 * Delays in reading/writing rasterlines implemented.
 * Complete debugging of the VIC-II register accesses.
 *
 * Revision 1.1  1996/04/01 05:49:29  ettore
 * Initial revision
 *
 */


#define _VICII_C

#include "extern.h"
#include "config.h"
#include "vicii.h"
#define __VIC_II__
#include "raster.c"
#include "vmachine.h"
#include "sprcoll.h"
#include "rasterutil.h"
#include "interrupt.h"
#include "extern.h"		/* rmw_flag */

#include <stdlib.h>
#include <stdio.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>

#if defined (C128)
#include "vdc.h"
#endif


/* Debugging options */

/* #define VIC_II_VMODE_DEBUG */
/* #define VIC_II_RASTER_DEBUG */
/* #define VIC_II_INTERRUPT_DEBUG */
/* #define DEBUG_SS_COLLISIONS */
/* #define DEBUG_SB_COLLISIONS */
/* #define VIC_II_REGISTERS_DEBUG */


#if defined (C128) && defined (USE_VDC)
extern VideoStruct *v_vdc;
#endif


/* global variables */

BYTE vic[64];
BYTE colorRam[0x400];
int vbank = 0;
BYTE *scrbase; 
BYTE *charbase;
BYTE *bitmapbase;


/* sprite collision masks */

static sprcollmask_t ss_collmask = 0;	/* sprite-sprite */
static sprcollmask_t sb_collmask = 0;	/* sprite-background */


/* the following enable/disable the collision registers */

#if defined (CHECK_SS_COLL)
int check_ss_coll = 1;
#else
int check_ss_coll = 1;
#endif

#if defined (CHECK_SB_COLL)
int check_sb_coll = 1;
#else
int check_sb_coll = 0;
#endif


VideoStruct *v_vic_ii = NULL;

#if defined (C128)
extern BYTE mmu[];
#endif


/* local Variables */

static int videoint = 0;	/* interrupt register for vic */
static CLOCK rasteru;		/* time when next raster count reaches raster register */
static ADDRESS scraddr;		/* screen start address */
static int rasterintline = 0;   /* line for raster IRQ */


/*

   screen_init(): initializes the VIC-II memory pointers.

 */

void
screen_init (void)
{
  ADDRESS addr;

  scraddr = vbank + ((vic[0x18] & 0xf0) << 6);

  if ((scraddr & 0x7000) != 0x1000)
    {
      scrbase = ram + scraddr - vbank;
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tVideo memory at $%04X\n", scrbase - ram + vbank);
#endif
    }
  else
    {
      scrbase = char_rom + (scraddr & 0x800);
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tVideo memory at Character ROM + $%04X\n", scraddr & 0x800);
#endif
    }

  addr = ((vic[0x18] & 0xe) << 10);
  bitmapbase = ram + (addr & 0xe000);
  addr += vbank;

#if defined (VIC_II_REGISTERS_DEBUG)
  printf ("\tBitmap memory at $%04X\n", bitmapbase - ram + vbank);
#endif 

  if ((addr & 0x7000) != 0x1000)
    {
      charbase = ram + addr;
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tUser-defined character set at $%04X\n", addr);
#endif
    }
  else
    {
      charbase = char_rom + (addr & 0x0800);
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tStandard %s character set enabled\n", addr & 0x800 ?
	      "Lower Case" : "Upper Case");
#endif
    }

  v_vic_ii->screenmem = scrbase + vbank;
  v_vic_ii->colormem = colorRam;
  v_vic_ii->bitmapmem = bitmapbase + vbank;
  v_vic_ii->chargenmem = charbase;
  v_vic_ii->vbankbase = ram + vbank;
  v_vic_ii->spriteptrbase = &v_vic_ii->screenmem[0x3f8];
}


/*

   update_raster(): updates the raster variable so that the next raster interrupt is generated
   correctly.

 */

INLINE void
update_raster (int flag)
{
  rasteru = v_vic_ii->oldclk - VIC_FETCH + (1 + rasterintline - v_vic_ii->rasterline) *
    CYCLES_PER_LINE;
  if (rasteru < (v_vic_ii->oldclk - VIC_FETCH))
    {
      rasteru += CYCLES_PER_LINE * SCR_LINES;
    }
  if (flag)
    {
      setalarmclk (I_RASTER, rasteru);
#if defined (NEW_TIMER_CODE)
      update_intr_alarm ();
#endif
    }

#if defined (VIC_II_RASTER_DEBUG) || defined (VIC_II_INTERRUPT_DEBUG)
  printf ("VIC: update_raster(%d), rasteru = %ld, line = $%04x, vic[0x1a] & 1 = %d\n",
	  flag, rasteru, rasterintline, vic[0x1a] & 1);
#endif
}


/*

   The following functions are used to read the screen data (see raster.c for more info)

 */

int
VicII_GetHiresBitmapColorData (VideoStruct * v, LineStruct * l, int *xs, int *xe, int no_check)
{
  if (v->ycounter > 7)
    {
      BYTE tmpcol = v->bgcol;
      return GetDataHiLo (l->colordata1, l->bgdata, &tmpcol, v->tcols, 0, xs, xe, no_check);
    }
  else
    return GetDataHiLo (l->colordata1, l->bgdata, v->vbuf, v->tcols, 1, xs, xe, no_check);
}

int
VicII_GetHiresBitmapBgData (VideoStruct * v, LineStruct * l, int *xs, int *xe)
{
  return 0;			/* no need to get background data (see prev. function) */
}

int
VicII_GetMCBitmapColorData (VideoStruct * v, LineStruct * l, int *xs, int *xe, int no_check)
{
  int r;

  if (v->ycounter > 7)
    return VicII_GetHiresBitmapColorData (v, l, xs, xe, no_check);
  else
    {
      r = GetDataHiLo (l->colordata1, l->colordata2, v->vbuf, v->tcols, 1, xs,
		       xe, no_check);
      return GetData (l->colordata3, v->colormem + v->memptr, v->tcols, 1, xs, xe, r);
    }
}

int
VicII_GetMCTextColorData (VideoStruct * v, LineStruct * l, int *xs, int *xe, int no_check)
{
  BYTE *addr = v->colormem + v->memptr;

  if (v->ycounter <= 7)
    {
      if (vic[0x22] != l->colordata1[0] || vic[0x23] != l->colordata2[0])
        {
          no_check = 1;
          *xs = 0;
          *xe = v->tcols - 1;
          l->colordata1[0] = vic[0x22];	/* we only use the first byte for MCTextMode */
          l->colordata2[0] = vic[0x23];
        }
      return GetData (l->colordata3, addr, v->tcols, 1, xs, xe, no_check);
    }
  else
    return VicII_GetHiresBitmapColorData(v, l, xs, xe, no_check);
}

int
VicII_GetExtTextData (VideoStruct * v, LineStruct * l, int *xs, int *xe, int no_check)
{
  static BYTE fdata[40], bdata[40];
  BYTE b, *addr = v->vbuf;
  int i;

  for (i = 0; i < 40; i++)
    {
      fdata[i] = (b = addr[i]) & 0x3f;
      bdata[i] = b >> 6;
    }

  no_check = GetData (l->bgdata, bdata, v->tcols, 1, xs, xe, no_check);
  return GetCharData (l->fgdata, fdata, v->chargenmem, v->tcols, v->ycounter, xs, xe, no_check);
}

int
VicII_GetExtTextColorData (VideoStruct * v, LineStruct * l, int *xs, int *xe, int no_check)
{
  int i;

  if (v->ycounter > 7)
    return VicII_GetHiresBitmapColorData (v, l, xs, xe, no_check);
  else
    {
      if (no_check || vic[0x21] != l->colordata2[0] || vic[0x22] != l->colordata2[1]
	  || vic[0x23] != l->colordata2[2] || vic[0x24] != l->colordata2[3])
	{
	  l->colordata2[0] = vic[0x21];
	  l->colordata2[1] = vic[0x22];
	  l->colordata2[2] = vic[0x23];
	  l->colordata2[3] = vic[0x24];
	  no_check = 1;
	}
      for (i = 0; i < v->tcols; i++)
	l->bgdata[i] = l->colordata2[l->bgdata[i]];
      return NormalGetColorData (v, l, xs, xe, no_check);
    }
}

int
VicII_GetExtTextBgData (VideoStruct * v, LineStruct * l, int *xs, int *xe, int no_check)
{
  return 0;			/* we already did this... */
}

int
VicII_GetBlackData (VideoStruct * v, LineStruct * l, int *xs, int *xe, int no_check)
{
  return GetData (l->fgdata, (BYTE *) "", v->tcols, 0, xs, xe, no_check);
}

int
VicII_GetBlackColorData (VideoStruct * v, LineStruct * l, int *xs, int *xe, int no_check)
{
  return GetDataHiLo (l->colordata1, l->bgdata, (BYTE *) "",  v->tcols, 0, xs, xe, no_check);
}

int
VicII_GetBlackBgData (VideoStruct * v, LineStruct * l, int *xs, int *xe, int no_check)
{
  return 0;
}

void
VicII_DrawMCTextLine (PIXEL * p, LineStruct * l, int xs, int xe, PIXEL * colors)
{
  BYTE *colorp, *pixelp;
  int i;
  register PIXEL c1, c2, c3;

  c1 = colors[vic[0x22]];
  c2 = colors[vic[0x23]];
  colorp = &(l->colordata3[0]);
  pixelp = &(l->fgdata[0]);
  p += xs * CHAR_XSIZE;
  for (i = xs; i <= xe; i++, p += CHAR_XSIZE)
    {
      register BYTE byte = pixelp[i];
      if (colorp[i] & 0x8)
	{
	  c3 = colors[colorp[i] & 0x7];
	  PUTMULTI(byte, c1, c2, c3);
	}
      else
	{
	  register PIXEL fg = colors[colorp[i]];
	  PUTMONO(byte, fg);
	}
    }
}


/*

   VicII_SetVMode(): sets the video mode according to the values in the registers.

 */

void
VicII_SetVMode (void)
{
#if defined (VIC_II_VMODE_DEBUG)
  printf ("\t");
#endif
  if ((vic[0x11] & 0x40) && (vic[0x16] & 0x10))
    {
#if defined (VIC_II_VMODE_DEBUG)
      printf ("Black Screen");
#endif
      v_vic_ii->vmode = VIC_II_BLACK_MODE;
    }
  else
    {
      if (vic[0x11] & 0x20)
	if (vic[0x16] & 0x10)
	  {
	    v_vic_ii->vmode = VIC_II_MULTICOLOR_BITMAP_MODE;
#if defined (VIC_II_VMODE_DEBUG)
	    printf ("Multicolor Bitmap");
#endif
	  }
	else
	  {
	    v_vic_ii->vmode = VIC_II_HIRES_BITMAP_MODE;
#if defined (VIC_II_VMODE_DEBUG)
	    printf ("Hires Bitmap");
#endif
	  }
      else if (vic[0x11] & 0x40)
	{
	  v_vic_ii->vmode = VIC_II_EXTENDED_TEXT_MODE;
#if defined (VIC_II_VMODE_DEBUG)
	  printf ("Extended Text");
#endif
	}
      else
	{
	  if (vic[0x16] & 0x10)
	    {
	      v_vic_ii->vmode = VIC_II_MULTICOLOR_TEXT_MODE;
#if defined (VIC_II_VMODE_DEBUG)
	      printf ("Multicolor Text");
#endif
	    }
	  else
	    {
	      v_vic_ii->vmode = VIC_II_NORMAL_TEXT_MODE;
#if defined (VIC_II_VMODE_DEBUG)
	      printf ("Normal Text");
#endif
	    }
	}
#if defined (VIC_II_VMODE_DEBUG)
      printf (" Mode enabled at rasterline $%04X.\n", v_vic_ii->rasterline);
#endif
    }
}


/*

   store_vic(): stores a value in a VIC-II register.

 */

void
store_vic (ADDRESS addr, BYTE value)
{
  addr &= 0x3f;

#if defined (VIC_II_REGISTERS_DEBUG)
  printf ("VIC: WRITE $D0%02X at rasterline $%04X\n", addr, v_vic_ii->rasterline);
#endif

  switch (addr)
    {
    case 0x0:			/* $D000: Sprite #0 X position LSB */
    case 0x2:			/* $D002: Sprite #1 X position LSB */
    case 0x4:			/* $D004: Sprite #2 X position LSB */
    case 0x6:			/* $D006: Sprite #3 X position LSB */
    case 0x8:			/* $D008: Sprite #4 X position LSB */
    case 0xa:			/* $D00a: Sprite #5 X position LSB */
    case 0xc:			/* $D00c: Sprite #6 X position LSB */
    case 0xe:			/* $D00e: Sprite #7 X position LSB */
      vic[addr] = value;
      v_vic_ii->sprites[addr >> 1].x &= 0x100;
      v_vic_ii->sprites[addr >> 1].x |= value;
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tSprite #%d X position LSB: $%02X\n", addr >> 1, value);
#endif
      break;

    case 0x1:			/* $D001: Sprite #0 Y position */
    case 0x3:			/* $D003: Sprite #1 Y position */
    case 0x5:			/* $D005: Sprite #2 Y position */
    case 0x7:			/* $D007: Sprite #3 Y position */
    case 0x9:			/* $D009: Sprite #4 Y position */
    case 0xb:			/* $D00B: Sprite #5 Y position */
    case 0xd:			/* $D00D: Sprite #6 Y position */
    case 0xf:			/* $D00F: Sprite #7 Y position */
      vic[addr] = value;
      v_vic_ii->sprites[addr >> 1].y = value + 1;
#if defined (VIC_II_REGISTERS_DEBUG) 
      printf ("\tSprite #%d Y position: $%02X\n", addr >> 1, value);
#endif
      break;
	  
    case 0x10:			/* $D010: Sprite X position MSB */
      if (value != vic[0x10])
	{
	  int i;
	  BYTE b;
	  vic[0x10] = value;
	  for (i = 0, b = 0x01; i < 8; b <<= 1, i++)
	    v_vic_ii->sprites[i].x = (v_vic_ii->sprites[i].x & 0xff) |
	      (value & b ? 0x100 : 0);
	}
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tSprite X position MSBs: $%02X\n", value);
#endif
      return;

    case 0x11:			/* $D011: video mode, Y scroll, 24/25 line mode and raster MSB */
      {
	unsigned long tmp = (value & 0x7) - 3;
	unsigned int r = rasterintline;
#if defined (VIC_II_REGISTERS_DEBUG)
	printf ("\tControl register: $%02X\n", value);
#endif
	vic[0x11] = value;
	rasterintline = (rasterintline & 0xff) | ((value & 0x80) << 1);
#if defined (VIC_II_REGISTERS_DEBUG)
	printf ("\tRaster interrupt line set to $%04X\n", rasterintline);
#endif
	if (rasterintline != r)
	  update_raster (1);
#if defined (VIC_II_REGISTERS_DEBUG)
	if (rasterintline >= SCR_LINES)
	  printf ("VIC: mmmh... strange raster interrupt line $%04X\n", rasterintline);
#endif
	v_vic_ii->ysmooth = (int)tmp;
#if defined (VIC_II_REGISTERS_DEBUG)
	printf ("\tY smooth scroll value: $%04X\n", value & 0x7);
	printf ("\t%s line mode enabled\n", value & 0x8 ? "25" : "24");
#endif
	v_vic_ii->blank = !(value & 0x10);
	if (value & 0x8)
	  v_vic_ii->brdgrow_yu = v_vic_ii->brdgrow_yl = 0;
	else
	  /* ``open vertical borders'' effect */
	  {
	    if (!v_vic_ii->brdgrow_yu)
	      if (v_vic_ii->rasterline >= 0xf8 && v_vic_ii->rasterline <= 0xfb)
		{
		  v_vic_ii->openvborders = 1;
#if defined (VIC_II_REGISTERS_DEBUG)
		  printf ("\tThe vertical borders are now open!\n");
#endif
		}
	    v_vic_ii->brdgrow_yu = v_vic_ii->brdgrow_yl = 4;
	  }
	VicII_SetVMode ();
      }
      return;

    case 0x12:			/* $D012: Raster line compare */
      {
	unsigned int r = rasterintline;
	vic[0x12] = value;
	rasterintline = (rasterintline & 0x100) | value;
	if (r != rasterintline)
	  update_raster (1);
#if defined (VIC_II_REGISTERS_DEBUG)
	printf ("\tRaster interrupt line set to $%04X\n", rasterintline);
#endif
#if defined (VIC_II_REGISTERS_DEBUG)
	if (rasterintline >= SCR_LINES)
	  printf ("VIC: mmmh... strange raster interrupt line $%04X\n", rasterintline);
#endif
#if defined (VIC_II_REGISTERS_DEBUG)
	printf ("\tRaster compare register: $%02X\n", value);
#endif
      }
      return;

    case 0x13:			/* $D013: Light Pen X */
    case 0x14:			/* $D014: Light Pen Y */
      return;
	  
    case 0x15:			/* $D015: Sprite Enable */
      if (value != vic[0x15])
	{
	  int i;
	  BYTE b;
	  vic[0x15] = value;
	  for (i = 0, b = 0x01; i < 8; b <<= 1, i++)
	    v_vic_ii->sprites[i].visible = value & b ? 1 : 0;
	}
#if defined (VIC_II_REGISTERS_DEBUG)
	printf ("\tSprite Enable register: $%02X\n", value);
#endif
      return;

    case 0x16:			/* $D016 */
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tControl register: $%02X\n", value);
#endif
      vic[0x16] = value;
      v_vic_ii->xsmooth = value & 0x7;
      if (value & 0x8)
	{
	  v_vic_ii->brdgrow_xr = v_vic_ii->brdgrow_xl = 0;
#if defined (VIC_II_REGISTERS_DEBUG)
          printf ("\t40 column mode enabled\n");
#endif
	}
      else
	{
#if defined (VIC_II_REGISTERS_DEBUG)
	  printf ("\t38 column mode enabled\n");
#endif
	  v_vic_ii->brdgrow_xr = 9;
	  v_vic_ii->brdgrow_xl = 7;
	}
      VicII_SetVMode ();
      return;

    case 0x17:			/* $D017: Sprite Y-expand */
      if (value != vic[0x17])
	{
	  int i;
	  BYTE b;
	  vic[0x17] = value;
	  for (i = 0, b = 0x01; i < 8; b <<= 1, i++)
	    v_vic_ii->sprites[i].yexp = value & b ? 1 : 0;
	}
#if defined (VIC_II_REGISTERS_DEBUG)
	printf ("\tSprite Y Expand register: $%02X\n", value);
#endif
      return;

    case 0x18:			/* $D018: Video and char matrix base address */
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tMemory register: $%02X\n", value);
#endif
      vic[0x18] = value;
      screen_init ();
      return;

    case 0x19:			/* $D019: IRQ flag register */
      if (rmw_flag)		/* (emulates the Read-Modify-Write bug) */
	vic[0x19] = 0;
      else
	{
	  vic[0x19] &= ~(value | 0x80);
	  if (vic[0x19] & vic[0x1a])
	    vic[0x19] |= 0x80;
	}
      videoint = vic[0x19];

      /* update the IRQ line accordingly... */
      /* the VIC IRQ line is a AND of the internal collision and raster IRQ lines */
      if (videoint & 0x7)
	{
	  setirq (I_RASTER, 1);
	  videoint |= 0x80;
	}
      else
	{
	  setirq (I_RASTER, 0);
	  videoint &= 0x7f;
	}
      
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tIRQ flag register: $%02X\n", vic[0x19]);
#endif
      return;

    case 0x1a:			/* $D01A: IRQ mask register */
      /* light pen interrupt not implemented */
      if (value & 8)
	int_clk[I_LIGHT_PEN] = 0;
      else
	int_clk[I_LIGHT_PEN] = 0;

#if defined (NEW_TIMER_CODE)
      update_intr_alarm ();
#endif
      vic[0x1a] = value & 0xf;

      if (vic[0x1a] & videoint)
	{
	  videoint |= 0x80;
	  setirq (I_RASTER, 1);
	}
      else
	{
	  videoint &= 0x7f;
	  setirq (I_RASTER, 0);
	}
      
#if defined (VIC_II_REGISTERS_DEBUG) 
      printf ("\tIRQ mask register: $%02X\n", vic[0x1a]);
#endif
      return;

    case 0x1b:			/* $D01B: Sprite priority */
      if (value != vic[0x1b])
	{
	  int i;
	  BYTE b;
	  vic[0x1b] = value;
	  for (i = 0, b = 0x01; i < 8; b <<= 1, i++)
	    v_vic_ii->sprites[i].priority = value & b ? 1 : 0;
	}
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tSprite priority register: $%02X\n", value);
#endif
      return;

    case 0x1c:			/* $D01C: Sprite Multicolor select */
      if (value != vic[0x1c])
	{
	  int i;
	  BYTE b;
	  vic[0x1c] = value;
	  for (i = 0, b = 0x01; i < 8; b <<= 1, i++)
	    v_vic_ii->sprites[i].multicolor = value & b ? 1 : 0;
	}
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tSprite Multicolor Enable register: $%02X\n", value);
#endif
      return;

    case 0x1d:			/* $D01D: Sprite X-expand */
      if (value != vic[0x1d])
	{
	  int i;
	  BYTE b;
	  vic[0x1d] = value;
	  for (i = 0, b = 0x01; i < 8; b <<= 1, i++)
	    v_vic_ii->sprites[i].xexp = value & b ? 1 : 0;
	}
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tSprite X Expand register: $%02X\n", value);
#endif
      return;

    case 0x1e:			/* $D01E: Sprite-sprite collision */
    case 0x1f:			/* $D01F: Sprite-background collision */
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\t(collision register, Read Only)\n");
#endif
      return;
	  
    case 0x20:			/* $D020: Border color */
      v_vic_ii->bordcol = vic[0x20] = value & 0xf;
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tBorder color register: $%02X\n", value);
#endif
      return;

    case 0x21:			/* $D021: Background #0 color */
      v_vic_ii->bgcol = vic[0x21] = value & 0xf;
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tBackground #0 color register: $%02X\n", value);
#endif
      return;

    case 0x22:			/* $D022: Background #1 color */
    case 0x23:			/* $D023: Background #2 color */
    case 0x24:			/* $D024: Background #3 color */
      vic[addr] = value & 0xf;
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tBackground color #%d register: $%02X\n", addr - 0x21, value);
#endif
      return;
	  
    case 0x25:			/* $D025: Sprite multicolor register #0 */
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tBackground multicolor register #0: $%02X\n", value);
#endif
      v_vic_ii->sprmc1 = vic[0x25] = value & 0xf;
      return;

    case 0x26:			/* $D026: Sprite multicolor register #1 */
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tBackground multicolor register #1: $%02X\n", value);
#endif
      v_vic_ii->sprmc2 = vic[0x26] = value & 0xf;
      return;

    case 0x27:			/* $D027: Sprite #0 color */
    case 0x28:			/* $D028: Sprite #1 color */
    case 0x29:			/* $D029: Sprite #2 color */
    case 0x2a:			/* $D02A: Sprite #3 color */
    case 0x2b:			/* $D02B: Sprite #4 color */
    case 0x2c:			/* $D02C: Sprite #5 color */
    case 0x2d:			/* $D02D: Sprite #6 color */
    case 0x2e:			/* $D02E: Sprite #7 color */
      v_vic_ii->sprites[addr - 0x27].color = vic[addr] = value & 0xf;
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tSprite #%d color register: $%02X\n", addr - 0x27, value);
#endif
      return;

    case 0x2f:			/* $D02F: Unused */
    case 0x30:			/* $D030: Unused */
    case 0x31:			/* $D031: Unused */
    case 0x32:			/* $D032: Unused */
    case 0x33:			/* $D033: Unused */
    case 0x34:			/* $D034: Unused */
    case 0x35:			/* $D035: Unused */
    case 0x36:			/* $D036: Unused */
    case 0x37:			/* $D037: Unused */
    case 0x38:			/* $D038: Unused */
    case 0x39:			/* $D039: Unused */
    case 0x3a:			/* $D03A: Unused */
    case 0x3b:			/* $D03B: Unused */
    case 0x3c:			/* $D03C: Unused */
    case 0x3d:			/* $D03D: Unused */
    case 0x3e:			/* $D03E: Unused */
    case 0x3f:			/* $D03F: Unused */
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\t(unused)\n");
#endif
      return;
    }
}


/*

   read_vic(): reads from one VIC-II register.

 */

BYTE
read_vic (ADDRESS addr)
{
  addr &= 0x3f;

#if defined (VIC_II_REGISTERS_DEBUG)
  printf ("VIC: READ $D0%02X at rasterline $%04X:\n", addr, v_vic_ii->rasterline);
#endif

  switch (addr)
    {
    case 0x0:			/* $D000: Sprite #0 X position LSB */
    case 0x2:			/* $D002: Sprite #1 X position LSB */
    case 0x4:			/* $D004: Sprite #2 X position LSB */
    case 0x6:			/* $D006: Sprite #3 X position LSB */
    case 0x8:			/* $D008: Sprite #4 X position LSB */
    case 0xa:			/* $D00a: Sprite #5 X position LSB */
    case 0xc:			/* $D00c: Sprite #6 X position LSB */
    case 0xe:			/* $D00e: Sprite #7 X position LSB */
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tSprite #%d X position LSB: $%02X\n", addr >> 1, vic[addr]);
#endif
      return vic[addr];
      
    case 0x1:			/* $D001: Sprite #0 Y position */
    case 0x3:			/* $D003: Sprite #1 Y position */
    case 0x5:			/* $D005: Sprite #2 Y position */
    case 0x7:			/* $D007: Sprite #3 Y position */
    case 0x9:			/* $D009: Sprite #4 Y position */
    case 0xb:			/* $D00B: Sprite #5 Y position */
    case 0xd:			/* $D00D: Sprite #6 Y position */
    case 0xf:			/* $D00F: Sprite #7 Y position */
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tSprite #%d Y position: $%02X\n", addr >> 1, vic[addr]);
#endif
      return vic[addr];

    case 0x10:			/* $D010: Sprite X position MSB */
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tSprite X position MSB: $%02X\n", vic[addr]);
#endif
      return vic[addr];
      
    case 0x11:		        /* $D011: video mode, Y scroll, 24/25 line mode and raster MSB */
    case 0x12:			/* $D012: Raster line compare */
      { /* ...a special thank to Lutz Sammer for this! */
	int tmp = v_vic_ii->rasterline;
	if (clk < (v_vic_ii->oldclk + CYCLES_PER_LINE - VIC_FETCH - VIC_FETCH_ADD))
	  --tmp;
	if (tmp < 0)
	    tmp += SCR_LINES;
#if defined (VIC_II_RASTER_DEBUG) || defined (VIC_II_REGISTERS_DEBUG)
	printf ("\tRaster Line register %s value = $%04X\n",
		(addr == 0x11 ? "(highest bit) " : ""), tmp);
#endif
	if (addr == 0x11)
	  return ((vic[addr] & 0x7f) |
		  ((tmp & 0x100) >> 1));
	else
	  return ((tmp) & 0xff);
      }

    case 0x13:			/* $D013: Light Pen X */
    case 0x14:			/* $D014: Light Pen Y */
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\t(Light pen not implemented)\n");
#endif
      return 0xff;

    case 0x15:			/* $D015: Sprite Enable */
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tSprite Enable register: $%02X\n", vic[addr]);
#endif
      return vic[addr];

    case 0x16:			/* $D016 */
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\t$D016 Control register read: $%02X\n", vic[addr]);
#endif
      return vic[addr];

    case 0x17:			/* $D017: Sprite Y-expand */
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tSprite Y Expand register: $%02X\n", vic[addr]);
#endif
      return vic[addr];

    case 0x18:			/* $D018: Video and char matrix base address */
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tVideo memory address register: $%02X\n", vic[addr]);
#endif
      return vic[addr];

    case 0x19:			/* $D019: IRQ flag register */
#if defined (VIC_II_INTERRUPT_DEBUG) && !defined (VIC_II_REGISTERS_DEBUG)
      printf ("VIC: read interrupt register: $%02X\n", videoint | 0x70);
#endif
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tInterrupt register: $%02X\n", videoint | 0x70);
#endif
      return videoint | 0x70;

    case 0x1a:			/* $D01A: IRQ mask register  */
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tMask register: $%02X\n", vic[addr] | 0xf0);
#endif
      return vic[addr] | 0xf0;

    case 0x1b:			/* $D01B: Sprite priority */
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tSprite Priority register: $%02X\n", vic[addr]);
#endif
      return vic[addr];

    case 0x1c:			/* $D01C: Sprite Multicolor select */
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tSprite Multicolor Enable register: $%02X\n", vic[addr]);
#endif
      return vic[addr];

    case 0x1d:			/* $D01D: Sprite X-expand */
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tSprite X Expand register: $%02X\n", vic[addr]);
#endif
      return vic[addr];

    case 0x1e:			/* $D01E: Sprite-sprite collision */
      vic[addr] = ss_collmask;
      ss_collmask = 0;
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tSprite-sprite collision mask: $%02X\n", vic[addr]);
#endif
#if defined (DEBUG_SS_COLLISIONS)
      printf ("VIC: ss_collmask reset by $D01E read at line 0x%X.\n", v_vic_ii->rasterline);
#endif
      return vic[addr];

    case 0x1f:			/* $D01F: Sprite-background collision */
      vic[addr] = sb_collmask;
      sb_collmask = 0;
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tSprite-background collision mask: $%02X\n", vic[addr]);
#endif
#if defined (DEBUG_SB_COLLISIONS)
      printf ("VIC: sb_collmask reset by $D01F read at line 0x%X.\n", v_vic_ii->rasterline);
#endif
      return vic[addr];

    case 0x20:			/* $D020: Border color */
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tBorder Color register: $%02X\n", vic[addr]);
#endif
      return vic[addr];

    case 0x21:			/* $D021: Background #0 color */
    case 0x22:			/* $D022: Background #1 color */
    case 0x23:			/* $D023: Background #2 color */
    case 0x24:			/* $D024: Background #3 color */
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tBackground Color #%d register: $%02X\n", addr - 0x21, vic[addr]);
#endif
      return vic[addr];
	  
    case 0x25:			/* $D025: Sprite multicolor register #0 */
    case 0x26:			/* $D026: Sprite multicolor register #1 */
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tMulticolor register #%d: $%02X\n", addr - 0x22, vic[addr]);
#endif
      return vic[addr];
      
    case 0x27:			/* $D027: Sprite #0 color */
    case 0x28:			/* $D028: Sprite #1 color */
    case 0x29:			/* $D029: Sprite #2 color */
    case 0x2a:			/* $D02A: Sprite #3 color */
    case 0x2b:			/* $D02B: Sprite #4 color */
    case 0x2c:			/* $D02C: Sprite #5 color */
    case 0x2d:			/* $D02D: Sprite #6 color */
    case 0x2e:			/* $D02E: Sprite #7 color */
#if defined (VIC_II_REGISTERS_DEBUG)
      printf ("\tSprite #%d color: $%02X\n", addr - 0x22, vic[addr]);
#endif
      return vic[addr];

#if defined (C128)
    case 0x2f:			/* $D02F: Additional KBD rows */
      return (vic[47] | 0xf8);
#else
    case 0x2f:			/* $D02F: Unused */
#endif

    case 0x30:			/* $D030: Unused */
    case 0x31:			/* $D031: Unused */
    case 0x32:			/* $D032: Unused */
    case 0x33:			/* $D033: Unused */
    case 0x34:			/* $D034: Unused */
    case 0x35:			/* $D035: Unused */
    case 0x36:			/* $D036: Unused */
    case 0x37:			/* $D037: Unused */
    case 0x38:			/* $D038: Unused */
    case 0x39:			/* $D039: Unused */
    case 0x3a:			/* $D03A: Unused */
    case 0x3b:			/* $D03B: Unused */
    case 0x3c:			/* $D03C: Unused */
    case 0x3d:			/* $D03D: Unused */
    case 0x3e:			/* $D03E: Unused */
    case 0x3f:			/* $D03F: Unused */
      return vic[addr];

    default:
      return ((addr < 0x31) ? vic[addr] : 0xff);	/* VIC IIe size */
    }
}


/*

   store_colorram(): stores a value in the VIC-II color RAM.

 */

void
store_colorram (ADDRESS addr, BYTE value)
{
  colorRam[addr & 0x3ff] = value & 0xf;
}


/*

   read_colorram(): reads a value from the VIC-II color RAM.

 */

BYTE
read_colorram (ADDRESS addr)
{
  return colorRam[addr & 0x3ff];
}


/*

   int_raster(): VIC-II raster interrupt.

 */

void
int_raster (void)
{
  rasteru += SCR_LINES * CYCLES_PER_LINE;
  setalarmclk (I_RASTER, rasteru);
  videoint |= 0x1;
  if (vic[0x1a] & 0x1)
    {
      setirq (I_RASTER, 1);
      videoint |= 0x80; 
#if defined (VIC_II_RASTER_DEBUG)
      printf ("VIC: raster interrupt requested at line $%04X\n", v_vic_ii->rasterline);
#endif
    }
#if defined (NEW_TIMER_CODE)
  update_intr_alarm ();
#endif
}


/*

   VicII_Init(): Initializes the VIC-II emulation.

 */

Window
VicII_Init (Display * disp, int scr_num, char *win_name, char *icon_name, char *res_class,
	    char *res_name, int *width, int *height, char *icon_data[], Colormap * cmap_return,
	    int exact_speed, int asleep, int refrate, Atom * wm_delwin)
{
  v_vic_ii = AllocVideo (SCR_COLS, SCR_LINES, BORD_WIDTH, BORD_HEIGHT,
			 XPIX, YPIX, XSIZE, YSIZE, VIC_II_NUMCOLORS, refrate,
			 VIC_II_NUM_SPRITES, exact_speed, win_name, asleep);
  if (!v_vic_ii)
    {
      fprintf (stderr, "fatal error: can't initialize Vic-II emulation.\n");
      return -1;
    }
  if (OpenVideoWindow (disp, scr_num, v_vic_ii, win_name, win_name, res_class, res_name,
		       icon_data, vic_ii_cnames, cmap_return, wm_delwin))
    {
      fprintf (stderr, "fatal error: can't open window for Vic-II emulation.\n");
      return -1;
    }
  if (SetVideoModes (v_vic_ii, 6,
  /* VIC_II_NORMAL_TEXT_MODE */
		     NormalGetTextData, NormalGetColorData,
		     NormalGetBgDataFromRegister, HIRES_VMODE,
  /* VIC_II_MULTICOLOR_TEXT_MODE */
		     NormalGetTextData, VicII_GetMCTextColorData,
		     NormalGetBgDataFromRegister, SPECIAL_VMODE, VicII_DrawMCTextLine,
  /* VIC_II_EXTENDED_TEXT_MODE */
		     VicII_GetExtTextData, VicII_GetExtTextColorData,
		     VicII_GetExtTextBgData, HIRES_VMODE,
  /* VIC_II_HIRES_BITMAP_MODE */
		     NormalGetBitmapData, VicII_GetHiresBitmapColorData,
		     VicII_GetHiresBitmapBgData, HIRES_VMODE,
  /* VIC_II_MULTICOLOR_BITMAP_MODE */
		     NormalGetBitmapData, VicII_GetMCBitmapColorData,
		     NormalGetBgDataFromRegister, MULTICOLOR_VMODE,
  /* VIC_II_BLACK_MODE */
		     VicII_GetBlackData, VicII_GetBlackColorData,
		     VicII_GetBlackBgData, HIRES_VMODE))
    {
      fprintf (stderr, "fatal error: can't initialize Vic-II emulation.\n");
      return -1;
    }
  v_vic_ii->vmode = VIC_II_NORMAL_TEXT_MODE;
  store_vic (0x18, 0x15);
  VicII_SetVMode ();
  screen_init ();
  RefreshAll (v_vic_ii);
  *width = v_vic_ii->winwidth;
  *height = v_vic_ii->winheight;
  init_sprcoll ();
  return v_vic_ii->win;
}


/*

   VicII_ArrangeWindow(): centers the image at the center of the window.

 */

void
VicII_ArrangeWindow (int width, int height)
{
  ArrangeVideoImage (v_vic_ii, width, height);
  RefreshAll (v_vic_ii);
}


void
VicII_Free (void)
{
  FreeVideo (v_vic_ii);
}


void
VicII_Pause (void)
{
  PauseVideo (v_vic_ii);
}


void
VicII_Restart (void)
{
  RestartVideo (v_vic_ii);
}


/*

   This one performs the collision checking too...

 */

void
DoRaster (void)
{
  static sprmask_struct sprmasks[8];
  int line;

  if ((line = Raster (v_vic_ii)) != -1)
    if (check_ss_coll || check_sb_coll)
      {
	LineStruct * l = &(v_vic_ii->lineinfo[line]);
	init_sprite_masks (v_vic_ii, l, sprmasks);
	if (check_ss_coll)
	  {
	    int zf = ss_collmask == 0 ? 1 : 0;
	    check_ss_collisions (l, sprmasks, &ss_collmask);
	    if (ss_collmask)
	      {
#if defined (DEBUG_SS_COLLISIONS)
 		printf ("VIC: Sprite-Sprite collision, mask = $%02X ", ss_collmask);
 		print_binary (ss_collmask);
 		printf (" at line $%04X\n", v_vic_ii->rasterline);
#endif
		if (vic[0x1a] & 0x4 && zf)
		  {
		    setirq (I_RASTER, 1);
		    videoint |= 0x84;
		  }
	      }
	  }
	if (check_sb_coll)
	  {
	    int zf = sb_collmask == 0 ? 1 : 0;
	    check_sb_collisions (v_vic_ii, l, sprmasks, &sb_collmask);
	    if (sb_collmask)
	      {
#if defined (DEBUG_SB_COLLISIONS)
 		printf ("VIC: Sprite-Background collision! mask = $%02X ", sb_collmask);
 		print_binary (sb_collmask);
 		printf (" at line $%04X\n", v_vic_ii->rasterline);
#endif
		if (vic[0x1a] & 0x2 && zf)
		  {
		    setirq (I_RASTER, 1);
		    videoint |= 0x82;
		  }
	      }
	  }
      }

#if defined (C128) && defined (USE_VDC)
  Vdc_DoRaster ();
#endif
}


/*

   The following is needed to turn Vic-II emulation on and off in x128.

 */

void
VicII_SetStatus (int awake)
{
  set_window_mapping (v_vic_ii->display, v_vic_ii->win, !(v_vic_ii->asleep = !awake));
  SetSpeed (speed);
}
