/*
 * $Id: video_x.c,v 1.2 1996/07/29 23:32:03 ettore Exp ettore $
 *
 * This file is part of VICE, the Versatile Commodore Emulator
 * See README file for copyright notice.
 *
 * X11 video interfacing.
 *
 * Written by
 *   Ettore Perazzoli (ettore@comm2000.it)
 *
 *
 * $Log: video_x.c,v $
 * Revision 1.2  1996/07/29 23:32:03  ettore
 * set_window_mapping() and resize_window().
 * Catches WM_DelWin.
 *
 * Revision 1.1  1996/04/01 05:45:58  ettore
 * Initial revision
 *
 */

#include <stdio.h>
#include <limits.h>

#include "types.h"
#include "video_x.h"
#include "config.h"
#include "x64.h"

#include <X11/cursorfont.h>
#if defined (XPM)
#include <X11/xpm.h>
#endif


/*

   Internal display-specific variables

 */

Display *display;
int      screen;
Atom     WM_DelWin;


/*

   alloc_colors(): tries to allocate the colors for the video emulations.  If this can't be done
   with the default colormap, it allocates a private one. Returns 0 on success.

 */

int
alloc_colors (Display * disp, int scr_num, Window w, int n_colors, char *cnames[],
	      Colormap * cmap_return, PIXEL * pixels_return)
{
  int i, use_own_colormap = 0;
  XColor screen_def, exact_def;

  /* try with the default colormap first */

  *cmap_return = DefaultColormap (disp, scr_num);
  for (i = 0; i < n_colors; i++)
    {
      if (!XAllocNamedColor (disp, *cmap_return, cnames[i], &screen_def, &exact_def))
	{
	  use_own_colormap = 1;
	  break;
	}
      if (screen_def.pixel > PIXEL_MAX)
	{
	  use_own_colormap = 1;
	  break;
	}
      else
	pixels_return[i] = (PIXEL) screen_def.pixel;
    }
  if (!use_own_colormap)
    return 0;

  /* we can't use the default colormap - let's allocate a private one, then */

  *cmap_return = XCreateColormap (disp, w, visual, AllocNone);
  puts ("using private colormap...");
  for (i = 0; i < n_colors; i++)
    {
      if (!XAllocNamedColor (disp, *cmap_return, cnames[i], &screen_def, &exact_def))
	return -1;		/* argh... we can't do it */
      if (screen_def.pixel > PIXEL_MAX)
	return -1;
      else
	pixels_return[i] = (PIXEL) screen_def.pixel;
    }

  XSetWindowColormap (disp, w, *cmap_return);
  return 0;
}


/*

   create_window(): initializes a window and allocates the colors.

 */

Window
create_window (Display * disp, int scr_num, char *win_name, char *icon_name, char *res_class,
	       char *res_name, int width, int height, int minwidth, int minheight, int maxwidth,
	       int maxheight, char *icon_data[], int mapped, Atom *wm_delwin)
{
  int disp_width = DisplayWidth (disp, scr_num);
  int disp_height = DisplayHeight (disp, scr_num);
  XClassHint class_hints;
  XSizeHints size_hints;
  XSetWindowAttributes win_attr;
  Pixmap icon_pixmap;
  Window win;
  XTextProperty win_name_prop, icon_name_prop;
  XWMHints wm_hints;
  int x_pos, y_pos;

  x_pos = (disp_width - width) / 2;
  y_pos = (disp_height - height) / 2;

  win_attr.cursor = XCreateFontCursor (disp, XC_left_ptr);
  win_attr.backing_store = Always;
  win_attr.event_mask = ExposureMask | KeyReleaseMask | KeyPressMask |
    ButtonPressMask | ButtonReleaseMask | StructureNotifyMask | FocusChangeMask |
    EnterWindowMask | LeaveWindowMask;
  win = XCreateWindow (disp, RootWindow (disp, scr_num), x_pos, y_pos,
		       width, height, 0, depth, InputOutput, visual,
		       CWBackingStore | /* CWBackPixel | */ CWEventMask | CWCursor, &win_attr);
  if (!win)
    return 0;
  size_hints.width = width;
  size_hints.height = height;
  size_hints.min_width = minwidth;
  size_hints.min_height = minheight;
  size_hints.max_width = maxwidth;
  size_hints.max_height = maxheight;
  size_hints.x = x_pos;
  size_hints.y = y_pos;
  size_hints.flags = PSize | PMinSize | PMaxSize;
  XSetStandardProperties (disp, win, win_name, win_name, None, NULL, 0, &size_hints);
  if (!XStringListToTextProperty (&win_name, 1, &win_name_prop))
    {
      XDestroyWindow (disp, win);
      return 0;
    }
  if (!XStringListToTextProperty (&icon_name, 1, &icon_name_prop))
    {
      XDestroyWindow (disp, win);
      return 0;
    }
  class_hints.res_name = res_name;
  class_hints.res_class = res_class;
  XSetClassHint (disp, win, &class_hints);
  wm_hints.initial_state = NormalState;
  wm_hints.input = True;
#if defined (XPM)
  XCreatePixmapFromData (disp, RootWindow (disp, scr_num), icon_data, &icon_pixmap, NULL, NULL);
  wm_hints.icon_pixmap = icon_pixmap;
  wm_hints.flags = StateHint | InputHint | IconPixmapHint;
#else
  wm_hints.flags = StateHint | InputHint;
#endif
  XSetWMProperties (disp, win, &win_name_prop, &icon_name_prop, NULL, 0, &size_hints, &wm_hints,
		    &class_hints);
  XSetWMProtocols (disp, win, wm_delwin, 1);
  if (mapped)
    XMapWindow (disp, win);
  return (win);
}


/*

   resize_window(): resizes a window.

 */

void
resize_window (Display * disp, Window w, int winwidth, int winheight)
{
  XResizeWindow (disp, w, winwidth, winheight);
}


/*

   set_window_mapping(): maps/unmaps one window.

 */

void
set_window_mapping (Display * disp, Window w, int mapped)
{
  if (mapped)
    XMapWindow (disp, w);
  else
    XUnmapWindow (disp, w);
  XSync (disp, 1);
}
