/*
 * $Id: xcontrol.c,v 1.8 1996/04/01 09:01:41 jopi Exp jopi $
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README for copyright notice
 *
 * X64 V0.2 PL 2 Control Window
 *
 *
 * Written by
 *   Jarkko Sonninen (sonninen@lut.fi)
 *   Jouko Valta (jopi@stekt.oulu.fi)
 *
 * Patch by
 *   Richard Hable (Richard.Hable@jk.uni-linz.ac.at) /RH/
 *
 *
 * $Log: xcontrol.c,v $
 * Revision 1.8  1996/04/01  09:01:41  jopi
 * Minor updates
 * Bus Activity display
 *
 * Revision 1.7  1995/11/07  16:51:00  jopi
 * *** empty log message ***
 *
 * Revision 1.6  1995/04/01  07:54:09  jopi
 * X64 0.3 PL 0
 * Compability. Bug fixes.
 *
 * Revision 1.5  1994/01/26  16:21:38  jopi
 * X64 version 0.2 PL 1
 *
 * Revision 1.4  1993/11/10  01:55:34  jopi
 * *** empty log message ***
 *
 * Revision 1.3  93/06/21  13:39:52  jopi
 *  X64 version 0.2 PL 0
 *
 */


/* Xlib include files */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>

#include <X11/cursorfont.h>
#include <X11/keysym.h>

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include <sys/ioctl.h>

#ifdef __STDC__
#include <stdarg.h>
#else /* __STDC__ */
#include <varargs.h>
#endif

#include "cpu.h"	/* registers */
#include "macro.h"
#include "vmachine.h"	/* VIC definitions */
#include "extern.h"
#include "x64.h"

#define BORDER_WIDTH   1


/*
 * Local functions
 */

static void open_cntl_win ( void );
static void handle_debugwin_event ( XEvent * );

#ifndef __STDC__
static int  draw_string ( /* ... */ );
#else	       /* __STDC__ */

static int  draw_string ( Window w, int x, int y, char *s,... );
#endif	       /* __STDC__ */
static void redraw_cntl_win ( void );
static void update_cntl_win ( void );
static void update_asm ( void );


#ifdef LED_ARRAY

static void open_led_win ( void );
static int  toggle_ledbar ( void );
static void draw_ledbar ( void );
static void update_ledbar ( void );

#endif  /* LED_ARRAY */



/*
 * Display and screen are used as arguments to nearly every Xlib routine
 * so it simplifies routine calls to declare them global.
 */

extern Display *display;
extern int      screen;
extern Window   borderwin, VDCborderwin;


/* control win */

static GC      debuggc;
static Window  debug_stop, debug_step, debug_exit, debug_asm;
Window  debugwin;
int     debugstate;

static  int     halt, trace;
int     redraw;

int     ledflg = 0;

#ifdef LED_ARRAY
static Window ledwin = 0;
#endif


/* fonts */

static XFontStruct *font_info;	/* Structure containing font info */
static int     font_height;
static int     font_width;

static char   *font_list[] = {
    "-*-helvetica-bold-r-*-*-12-*",
    "-*-courier-bold-r-*-*-12-*",
    "-*-times-bold-r-*-*-12-*",
    "fixed",
    NULL,
};


/* ------------------------------------------------------------------------- */

void    debug_main(void)
{
    XGCValues values;
    Window    rootw = RootWindow(display, screen);
    char  **p = font_list;


    if (!display) {
	fprintf(stderr, "no X server !!!\nPlease, recompile x64.c\n");
	exit (-5);
    }


    while (p && *p && !(font_info = XLoadQueryFont(display, *p))) ++p;
    if (!p) {
	(void) fprintf(stderr, "%s: Fatal: No fonts available.\n", progname);
	exit(-1);
    }
    font_height = font_info -> ascent + font_info -> descent;
    font_width = font_info -> max_bounds.width;


    values.foreground = BlackPixel(display, screen);
    values.background = WhitePixel(display, screen);
    debuggc = XCreateGC(display, rootw, GCForeground | GCBackground, &values);

    XSetFont(display, debuggc, font_info -> fid);

    open_cntl_win();
#ifdef LED_ARRAY
    open_led_win();
#endif
}


static void open_cntl_win(void)
{
    Window    rootw;
    XSizeHints size_hints;
    XTextProperty windowName, iconName;
    XWMHints  wm_hints;
    XClassHint class_hints;

    char   *win_name = "C64 Control";
    char   *icon_name = "C64 CTRL";

    size_hints.min_width = size_hints.width = font_width * 27;
    size_hints.min_height = size_hints.height = font_height * 16;
    size_hints.flags = PSize | PMinSize;

    rootw = RootWindow(display, screen);
    debugwin = XCreateSimpleWindow(display, rootw, 0, 0,
	size_hints.width, size_hints.height,
	BORDER_WIDTH,
	BlackPixel(display, screen),
	WhitePixel(display, screen));

    if (debugflg)
	printf("Control Window open\n");

    XSetStandardProperties(display, debugwin, "VICE Control", "CTRL",
	None, NULL, 0, &size_hints);

    /* Select event types wanted */
    XSelectInput(display, debugwin,
	ExposureMask | KeyReleaseMask | KeyPressMask |
	ButtonPressMask | StructureNotifyMask |
	FocusChangeMask | EnterWindowMask | LeaveWindowMask);

    if (XStringListToTextProperty(&win_name, 1, &windowName) == 0) {
	(void) fprintf(stderr,
	    "%s: structure allocation for windowName failed.\n",
	    progname);
	exit(-1);
    }
    if (XStringListToTextProperty(&icon_name, 1, &iconName) == 0) {
	(void) fprintf(stderr,
	    "%s: structure allocation for iconName failed.\n",
	    progname);
	exit(-1);
    }
    wm_hints.initial_state = NormalState;

    wm_hints.input = True;
    wm_hints.flags = StateHint | InputHint;

    class_hints.res_name = progname;
    class_hints.res_class = "C64";

    XSetWMProperties(display, debugwin, &windowName, &iconName, NULL, 0,
	&size_hints, &wm_hints, &class_hints);


    debug_stop = XCreateSimpleWindow(display, debugwin,
	font_width, font_height * 10,
	font_width * 10, font_height,
	BORDER_WIDTH,
	BlackPixel(display, screen),
	WhitePixel(display, screen));
    XMapWindow(display, debug_stop);

    debug_step = XCreateSimpleWindow(display, debugwin,
	font_width, font_height * 12,
	font_width * 10, font_height,
	BORDER_WIDTH,
	BlackPixel(display, screen),
	WhitePixel(display, screen));
    XMapWindow(display, debug_step);

    debug_exit = XCreateSimpleWindow(display, debugwin,
	font_width, font_height * 14,
	font_width * 10, font_height,
	BORDER_WIDTH,
	BlackPixel(display, screen),
	WhitePixel(display, screen));
    XMapWindow(display, debug_exit);

    debug_asm = XCreateSimpleWindow(display, debugwin,
	font_width * 13, font_height * 1,
	font_width * 13, font_height * 14,
	BORDER_WIDTH,
	BlackPixel(display, screen),
	WhitePixel(display, screen));
    XMapWindow(display, debug_asm);

    /* Display window XMapWindow (display, debugwin); */
}


/* ------------------------------------------------------------------------- */

void toggle_cntl_win(void)
{
    if ((debugstate = !debugstate)) {
	XMapWindow(display, debugwin);
    } else {
	XUnmapWindow(display, debugwin);
    }
}

/* ------------------------------------------------------------------------- */

void    x_loop(void)
{
    XEvent  report;	/* Structure for event information */

#ifdef LED_ARRAY
    static CLOCK c;

    if (ledflg && (clk - c) > (CYCLES_PER_SEC >> 2)) {
	update_ledbar();
	c = clk;
    }
#endif


    if (debugstate)
	update_cntl_win();

/*
    if (redraw) {
	vic_redraw();
	redraw = 0;
    }

#ifndef SETOWN
    flush_queue();
#endif
*/

    while (XEventsQueued(display, QueuedAfterFlush) > 0 || (halt && !trace)) {
	XNextEvent(display, &report);

	/* Main Window Event */
	if (report.xany.window == borderwin ||
	    report.xany.window == VDCborderwin) {
	    handle_mainwin_event(&report);
	    continue;
	}
	if (report.xany.window == debugwin) {
	    handle_debugwin_event(&report);
	    continue;
	}
#ifdef LED_ARRAY
	if (report.xany.window == ledwin) {
	    if (report.type != Expose || report.xexpose.count != 0)
		continue;
	    draw_ledbar();
	    continue;
	}
#endif
    }		/* End while */

    if (trace) {
	int_clk[I_SPECX] = clk;
	find_next_int();
#ifdef NEW_TIMER_CODE
	update_intr_alarm();
#endif
	if (halt)
	    trace = 0;
    }
}


static void handle_debugwin_event(XEvent *report)
{
    static char buffer[20];
    KeySym  key;
    XComposeStatus compose;
    int     count;
    int     show_keyarr();


    /* printf("debugwin event\n"); */
    switch (report -> type) {
      case Expose:
	/*
	 * Unless this is the last contiquous expose, to place_text and
	 * place_notify in next Expose
	 */
	if (report -> xexpose.count != 0)
	    break;

	redraw_cntl_win();
	break;

      case ButtonPress:
	if (report -> xbutton.subwindow == debug_stop) {
	    halt = !halt;
	    redraw_cntl_win();

	} else if (report -> xbutton.subwindow == debug_step) {
	    trace = !trace;

	} else if (report -> xbutton.subwindow == debug_exit) {
	    end64();

	} else
	    switch (report -> xbutton.button) {
	      case Button1:
		mon (PC);
		break;
	      case Button2:
		ledflg = toggle_ledbar();	/* On the control win only */
		/*debugflg = !debugflg;*/
		break;
	      case Button3:
		/* open debugging window */
		toggle_cntl_win();
	    }
	break;
      case KeyPress:
	count = XLookupString(&report -> xkey, buffer, 20, &key, &compose);

	switch (key) {
	  case XK_Up:
	    joy1 |= 1;
	    break;
	  case XK_Down:
	    joy1 |= 2;
	    break;
	  case XK_Left:
	    joy1 |= 4;
	    break;
	  case XK_Right:
	    joy1 |= 8;
	    break;
	  case XK_Return:
	    joy1 |= 16;
	    break;
	}
	if (debugflg)
	    printf("press joy1 %x\n", joy1);
	break;

      case KeyRelease:
	count = XLookupString(&report -> xkey, buffer, 20, &key, &compose);

	switch (key) {
	  case XK_Up:
	    joy1 &= ~1;
	    break;
	  case XK_Down:
	    joy1 &= ~2;
	    break;
	  case XK_Left:
	    joy1 &= ~4;
	    break;
	  case XK_Right:
	    joy1 &= ~8;
	    break;
	  case XK_Return:
	    joy1 &= ~16;
	    break;
	}
	if (debugflg)
	    printf("release joy1 %x\n", joy1);
	break;

      case FocusIn:
      case EnterNotify:
	XAutoRepeatOff(display);   /* no repeat in debugwin */
	break;

      case FocusOut:
      case LeaveNotify:
	XAutoRepeatOn(display);
	break;

    }
}


/* ------------------------------------------------------------------------- */

#ifndef __STDC__
static int draw_string(va_alist) va_dcl
{
    va_list args;
    char    buf[1000];

    Window  w;
    int     x, y;
    char   *s;
    int     vsprintf();

    va_start(args);
    w = va_arg(args, Window);
    x = va_arg(args, int);
    y = va_arg(args, int);
    s = va_arg(args, char *);

    vsprintf(buf, s, args);
    va_end(args);

#else	       /* __STDC__ */

static int draw_string(Window w, int x, int y, char *s,...)
{
    va_list args;
    char    buf[1000];
    va_start(args, s);
    vsprintf(buf, s, args);
    va_end(args);
#endif	       /* __STDC__ */

    XDrawImageString(display, w, debuggc,
	x * font_width, font_info -> ascent + y * font_height,
	buf, strlen(buf));

    return (0);
}


static void    redraw_cntl_win(void)
{
    XClearWindow(display, debugwin);
    draw_string(debugwin, 1, 1, "AC:");
    draw_string(debugwin, 1, 2, "XR:");
    draw_string(debugwin, 1, 3, "YR:");
    draw_string(debugwin, 1, 4, "SP:");
    draw_string(debugwin, 1, 5, "PC:");
    draw_string(debugwin, 1, 6, "SR:");
    /* draw_string(debugwin, 1, 7, "MEM:");*/

    draw_string(debug_stop, 3, 0, halt ? "CONT " : "STOP ");
    draw_string(debug_step, 3, 0, halt ? "STEP  " : "TRACE ");
    draw_string(debug_exit, 3, 0, "EXIT");
    update_cntl_win();
    update_asm();
}


static void    update_cntl_win(void)
{
    draw_string(debugwin, 4, 1, " %02X ", AC);
    draw_string(debugwin, 4, 2, " %02X ", XR);
    draw_string(debugwin, 4, 3, " %02X ", YR);
    draw_string(debugwin, 4, 4, " %02X ", SP);
    draw_string(debugwin, 4, 5, " %04X ", PC);

    draw_string(debugwin, 4, 6, " %s  ", sprint_status());
   /* draw_string(debugwin, 4, 7, " %02X ", src); */
    draw_string(debugwin, 1, 8, "%-15s", sprint_opcode(PC, 1));

    if (halt || trace)
	update_asm();
}


static void    update_asm(void)
{
    ADDRESS p = PC;
    int     i;

    for (i = 0; i < 14; i++) {
	draw_string(debug_asm, 1, i, " %-15s", sprint_opcode(p, 1));
	p += clength[lookup[LOAD(p)].addr_mode];
    }
    draw_string(debug_asm, 0, 0, ">");
    draw_string(debugwin, 14, 0, " %04X  ", PC);
}


/* ------------------------------------------------------------------------- */

/*
 *  LED Bar - a simple toy.
 */

#ifdef LED_ARRAY

#define NUM_LEDS	16
#define LED_DIM		2	/* cycles */
#define LED_HALF	(CYCLES_PER_SEC >> 5)
#define LED_BRIGHT	(CYCLES_PER_SEC >> 4)

#define LED_X		10
#define LED_Y		20
#define LEDX_SPC	15
#define XOFF		20
#define YOFF		20
#define LED_WID		(LEDX_SPC * NUM_LEDS + 2 * XOFF)
#define LED_HEI		(LED_Y + 40 + 2 * YOFF)


static GC      led_gc[4];


static void open_led_win (void)
{
    Window    rootw;
    XSizeHints size_hints;
    XTextProperty windowName, iconName;
    XWMHints  wm_hints;
    XClassHint class_hints;
    XGCValues gcval;
    XColor    exact_def, rgb_def;
    Colormap  def_cmap = DefaultColormap(display, screen);
    int       i;

    char   *win_name = "Bus Activity";
    char   *icon_name = "LEDs";


    size_hints.min_width = size_hints.width = LED_WID;
    size_hints.min_height = size_hints.height = LED_HEI;
    size_hints.flags = PSize | PMinSize;

    rootw = RootWindow(display, screen);
    ledwin = XCreateSimpleWindow(display, rootw, 0, 0,
	size_hints.width, size_hints.height,
	BORDER_WIDTH,
	BlackPixel(display, screen),
	WhitePixel(display, screen));

    if (debugflg)
	printf("LED Window open\n");

    XSetStandardProperties(display, ledwin, win_name, icon_name,
	None, NULL, 0, &size_hints);

    /* Select event types wanted */
    XSelectInput(display, ledwin,
	ExposureMask | StructureNotifyMask | FocusChangeMask);

    if (XStringListToTextProperty(&win_name, 1, &windowName) == 0) {
	(void) fprintf(stderr,
	    "%s: structure allocation for windowName failed.\n",
	    progname);
	exit(-1);
    }
    if (XStringListToTextProperty(&icon_name, 1, &iconName) == 0) {
	(void) fprintf(stderr,
	    "%s: structure allocation for iconName failed.\n",
	    progname);
	exit(-1);
    }

    wm_hints.initial_state = NormalState;
    wm_hints.input = True;
    wm_hints.flags = StateHint | InputHint;

    class_hints.res_name = progname;
    class_hints.res_class = "C64";

    XSetWMProperties(display, ledwin, &windowName, &iconName, NULL, 0,
	&size_hints, &wm_hints, &class_hints);


    /* Allocate colors for images on screen */

    if (!led_gc[0]) {
	char *lc_names[] = { "red4", "red3", "red2", "red" };
	/*char *lc_names[] = { "blue", "red4", "yellow", "green" };*/

	if (DefaultDepth(display, screen) < 4) {
	    fprintf (stderr, "Cannot create LED bar.\n");
	    return;
	}

	/* Allocate read only colors for each LED brightness */

	for (i = 0; i < 4; ++i) {
	    if (!XAllocNamedColor(display, def_cmap,
				  lc_names[i], &exact_def, &rgb_def)) {
		fprintf(stderr, "Cannot allocate color %s\n", lc_names[i]);
		return;
	    }

	    gcval.foreground = exact_def.pixel;
	    gcval.background = WhitePixel(display, screen);
	    led_gc[i] =
		XCreateGC(display, rootw, GCForeground | GCBackground, &gcval);
	}  /* for */
    }
}


static int  toggle_ledbar (void)
{
    static int  st = 0;

    if (!st) XMapWindow (display, ledwin);
    else XUnmapWindow(display, ledwin);

    st = !st;
    return (st);
}


static void draw_ledbar (void)
{
    int  i;


    for (i = 0; i < NUM_LEDS; ++i)
	XDrawImageString (display, ledwin, debuggc,
			  XOFF + i * LEDX_SPC + 2,
			  YOFF + LED_Y + 15 /*font->ascent*/ +2,
			  "0123456789ABCDEF" + i, 1);

    XDrawImageString (display, ledwin, debuggc,
		      XOFF * 2, YOFF + LED_Y + 40 /*font->ascent*/ +2,
		      "VIC-74154 Virtual Bus Tester", 30);

    update_ledbar ();
}


static void update_ledbar (void)
{
    GC   gcl;
    int  i, x = XOFF;
    int *r;

    if (!ledwin)
	return;


    for (i = 0, r = &ledarr[0]; i < NUM_LEDS; ++i, ++r) {

	if (*r < LED_DIM)
	    gcl = led_gc[0];
	else if (*r < LED_HALF)
	    gcl = led_gc[1];
	else if (*r < LED_BRIGHT)
	    gcl = led_gc[2];
	else
	    gcl = led_gc[3];

	XFillRectangle (display, ledwin, gcl, x, YOFF, LED_X, LED_Y);

	x += LEDX_SPC;
	*r = 0;

    }  /* for */
}

#endif  /* LED_ARRAY */

