/*
 * $Id: xdebug.c,v 1.7 1996/07/30 08:26:20 fachat Exp ettore $
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README for copyright notice
 *
 * X11-based 65xx tracer/debugger interface.
 * Also the tree of X resources is built here.
 *
 *
 * Written by
 *   Jarkko Sonninen  (sonninen@lut.fi)
 *
 * Patches by
 *   Jouko Valta      (jopi@stekt.oulu.fi)
 *   Andre Fachat     (a.fachat@physik.tu-chemnitz.de)
 *   Ettore Perazzoli (ettore@comm2000.it)
 *
 *
 * $Log: xdebug.c,v $
 * Revision 1.7  1996/07/30 08:26:20  fachat
 * Miscellaneous bugs fixed.
 * Improved Trace Mode.
 * floppy9Name and tapeName resources, with `-9' and `-1' options.
 * Experimental popup menu.
 * hjoyPort resource for the Linux joystick.
 *
 * Revision 1.6  1996/04/01  09:01:41  jopi
 * New options
 * syntax() rerarranged
 * New debugger functions for Next
 *
 * Revision 1.5  1995/11/07  16:51:00  jopi
 * *** empty log message ***
 *
 * Revision 1.4  1995/04/01  07:54:09  jopi
 * X64 0.3 PL 0
 * Added new options: -notraps -brevision.
 * Debugger status display added.
 *
 * Revision 1.3  1994/12/12  16:59:44  jopi
 * Added display of Memory Address, Memory Data, and Clock Cycles.
 *
 * Revision 1.2  1994/06/16  17:19:26  jopi
 * Xrm replaced with XtResources
 *
 * Revision 1.1  1994/02/18  18:26:14  jopi
 * Initial revision
 *
 *
 */

/* 1996/04/11 a.fachat
 * 
 * We normally start in run mode. From "run" mode one can get to
 * "stop" or "trace" mode. From "stop", we should be able to execute
 * one opcode with "step". We can also get to "run" or "trace".
 * From "trace" we can get to "stop" or "run" mode. 
 * -> always have three buttons named "run", "trace" and "stop".
 * -> add a button "step" for stop mode.
 */

#ifdef __hpux
#ifndef _POSIX_SOURCE
#define _POSIX_SOURCE
#endif
#endif


#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>     /* Get standard string definitions. */
#include <X11/cursorfont.h>

/*
 * And then the same again by hand...
 * If you are unfortunate enough not to have XPointer etc defined by your
 * system, you have to define by hand all types that are missing:
 */

#ifdef X_NOT_STDC_ENV
typedef char * XPointer;
/*typedef void *XtPointer;*/
#define XtOffsetOf(s_type,field) XtOffset(s_type*,field)
#endif

#ifndef XtNjumpProc
#define XtNjumpProc "jumpProc"
#endif


/*** String definitions. ***/
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/Cardinals.h>

/*** Simple Widgets ***/
#include <X11/Xaw/Grip.h>
#include <X11/Xaw/List.h>
/* #include <X11/Xaw/Panner.h> */
/* #include <X11/Xaw/Repeater.h> */
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/StripChart.h>
#include <X11/Xaw/Toggle.h>

/*** Menus ***/
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/Sme.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>

/*** Text Widgets ***/
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Text.h>
#include <X11/Xaw/Paned.h>

#include "cpu.h"
#include "macro.h"
#include "vmachine.h"	/* I_SPECX */
#include "extern.h"
#include "misc.h"
#include "video_x.h"
#include "x64.h"
#include "interrupt.h"
#include "serial.h"
#include "popup.h"
#include "mshell.h"
#include "sid.h"

 
/*
 * Variables
 */


#ifdef LED_ARRAY
static Window ledwin = 0;
#endif

int  ledflg = 0;


XtAppContext    app_con;
extern Display *display;
extern Window   borderwin, VDCborderwin;

int      	halt = 0;
static int      trace = 0, next = 0;

static ADDRESS  current_point;
static char     buf[256];


enum Register {
    R_AC, R_XR, R_YR, R_SP, R_PC, R_SR, R_EFF, R_DL, R_CLK, NO_REGISTERS
};

static char *registerLabels[] = {
    "AC:", "XR:", "YR:", "SP:", "PC:", "SR:", "EA>", "DL>", "Clk"
};


enum DState { DS_Run, DS_Halt, DS_Trace, DS_Step, DS_Break, DS_Mon };
enum DState  state;

static char *StatusLabels[] = {
    "Running.", "Halted.", "Trace.", "Execute.", "Break.", "Monitor Break."
};



/*
 * Functions
 */

extern void  end64 ( void );
extern void  debug_main ( void );
extern void  toggle_cntl_win ( void );

static void  Syntax ( XtAppContext, char * );
extern void  handle_mainwin_event ( XEvent * );

static void  ThumbProc ( Widget bar, XtPointer client, XtPointer ptr );
static void  ScrollProc ( Widget bar, XtPointer client, XtPointer ptr );
static void  MenuBar ( Widget parent );
static void  RegisterBar ( Widget parent );

static void UpdateRegisters ( void );
static void UpdateStatus ( enum DState );

static void QuitCb  ( Widget w, XtPointer clientData, XtPointer callData );
static void StopCb  ( Widget w, XtPointer clientData, XtPointer callData );
static void RunCb   ( Widget w, XtPointer clientData, XtPointer callData );
static void TraceCb ( Widget w, XtPointer clientData, XtPointer callData );
static void NextCb  ( Widget w, XtPointer clientData, XtPointer callData );

#ifdef LED_ARRAY
static void actLedBar	(Widget w, XtPointer clientData, XtPointer callData);
#endif
static void actSoftReset(Widget w, XtPointer clientData, XtPointer callData);
static void actHardReset(Widget w, XtPointer clientData, XtPointer callData);
static void actPowerOnReset(Widget w, XtPointer clientData, XtPointer callData);
static void actQuit	(Widget w, XtPointer clientData, XtPointer callData);
static void actNMI	(Widget w, XtPointer clientData, XtPointer callData);
static void actMonitor	(Widget w, XtPointer clientData, XtPointer callData);
static void actDebug	(Widget w, XtPointer clientData, XtPointer callData);
static void actDisk8	(Widget w, XtPointer clientData, XtPointer callData);
static void actTape1	(Widget w, XtPointer clientData, XtPointer callData);
static void x_mon       (void);

void open_action_popup(Window root, XEvent * event_sent, int x, int y);
void open_toggle_popup ( Window root, int x, int y );

int  check_next ( void );
void loop_test ( int cc, BYTE src, int set );

static void  set_asm ( ADDRESS );
static void  draw_asm ( ADDRESS );

static void DiskCb  ( Widget w, XtPointer clientData, XtPointer callData );
#ifdef CBMTAPE
static void TapeCb  ( Widget w, XtPointer clientData, XtPointer callData );
#endif
#ifdef LED_ARRAY

static void LedCb  ( Widget w, XtPointer clientData, XtPointer callData );


static void open_led_win ( void );
static int  toggle_ledbar ( void );
static void draw_ledbar ( void );
static void update_ledbar ( void );

#endif  /* LED_ARRAY */


/*
 * Widgets
 */

Widget registers[NO_REGISTERS];
Widget runButton, stopButton, traceButton, nextButton, stepButton;
Widget view;
Widget ASMbar;
Widget status1;
Widget toplevel;


/* ------------------------------------------------------------------------- */

/*
 * All application resources must reside in a structure. The structure
 * contains space where the application resources will be stored.
 */

AppResources app_resources;

/*
 * The resources specification contains:
 *
 * The name and class of the resource.
 * The representation type of the resource.
 * The location in the structure to keep the value of this resource.
 * The default value for this resource.
 *
 * This information is passed into XtGetApplicationResources().
 */

static XtResource resources[] = {
 {"basicName",	"BasicName",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, basicName),	XtRString, (XPointer)BASICNAME},
 {"kernelName",	"KernelName",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, kernelName),	XtRString, (XPointer)KERNELNAME},
 {"charName",	"CharName",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, charName),	XtRString, (XPointer)CHARGENNAME},
 {"ramName",	"RamName",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, ramName),	XtRString, (XPointer)RAMNAME},
 {"biosName",	"BiosName",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, biosName),	XtRString, (XPointer)BIOSROM_NAME},
 {"exromName",	"ExromName",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, exromName),	XtRString, (XPointer)EXROMNAME},
 {"module",	"Module",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, module),	XtRString, (XPointer)0},
 {"directory",	"Directory",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, directory),	XtRString, (XPointer)0},
 {"reuName",	"ReuName",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, reuName),	XtRString, (XPointer)0},
 {"kernelRev",	"KernelRev",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, kernelRev),	XtRString, (XPointer)0},
 {"basicRev",	"BasicRev",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, basicRev),	XtRString, (XPointer)0},

 {"startAddr", 	"StartAddr",	XtRString,	  sizeof (XtRString),
      XtOffsetOf (AppResources, startAddr),	XtRImmediate, (XPointer)0},

 {"colors",	"Colors",	XtRBoolean, sizeof(Boolean),
      XtOffsetOf (AppResources, colors),	XtRImmediate, (XtPointer)True},
 {"asm", 	"Asm",		XtRBoolean, sizeof (Boolean),
      XtOffsetOf (AppResources, asmflag),	XtRImmediate, (XtPointer)False},
 {"hex", 	"Hex",		XtRBoolean, sizeof (Boolean),
      XtOffsetOf (AppResources, hexflag),	XtRImmediate, (XtPointer)True},
 {"debug", 	"Debug",	XtRBoolean, sizeof (Boolean),
      XtOffsetOf (AppResources, debugflag),	XtRImmediate, (XtPointer)False},
 {"verbose",	"Verbose",	XtRBoolean, sizeof (Boolean),
      XtOffsetOf (AppResources, verboseflag),	XtRImmediate,(XtPointer)False},
 {"test", 	"Test",		XtRBoolean, sizeof (Boolean),
      XtOffsetOf (AppResources, testflag),	XtRImmediate, (XtPointer)False},
 {"halt", 	"Halt",		XtRBoolean, sizeof (Boolean),
      XtOffsetOf (AppResources, haltflag),	XtRImmediate, (XtPointer)False},
 {"notraps", 	"NoTraps",	XtRBoolean, sizeof (Boolean),
      XtOffsetOf (AppResources, notraps),	XtRImmediate, (XtPointer)False},
 {"programName","ProgramName",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, programName),	XtRString, (XPointer)0},
 {"tapeName",	"TapeName",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, tapeName),	XtRString, (XPointer)0},
 {"floppyName",	"FloppyName",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, floppyName),	XtRString, (XPointer)0},
 {"floppy9Name","Floppy9Name",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, floppy9Name),	XtRString, (XPointer)0},
 {"printCommand","PrintCommand",XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, printCommand),	XtRString, (XPointer)"lp"},

 /* added 03-15-96 -- EP */
#ifndef PET 
 {"joyPort", 	"JoyPort", 	XtRInt, sizeof (XtRInt),
      XtOffsetOf (AppResources, joyPort), 	XtRImmediate, (XtPointer)2},
#endif
#if defined( CBM64) || defined( C128)
#if defined( CHECK_SS_COLL)
 {"checkSsColl", "CheckSsColl",	XtRBoolean, sizeof (Boolean),
      XtOffsetOf (AppResources, checkSsColl),    XtRImmediate, (XtPointer)True},
#else
 {"checkSsColl", "CheckSsColl",	XtRBoolean, sizeof (Boolean),
      XtOffsetOf (AppResources, checkSsColl),    XtRImmediate, (XtPointer)False},
#endif
#if defined( CHECK_SB_COLL)
 {"checkSbColl", "CheckSbColl",	XtRBoolean, sizeof (Boolean),
      XtOffsetOf (AppResources, checkSbColl),    XtRImmediate, (XtPointer)True},
#else
 {"checkSbColl", "CheckSbColl",	XtRBoolean, sizeof (Boolean),
      XtOffsetOf (AppResources, checkSbColl),    XtRImmediate, (XtPointer)False},
#endif
#endif
#if defined( C128) || defined( PET)
 {"video40", 	"Video40",	XtRBoolean, sizeof (Boolean),
      XtOffsetOf (AppResources, video40),	XtRImmediate, (XtPointer)False},
 {"video80", 	"Video80",	XtRBoolean, sizeof (Boolean),
      XtOffsetOf (AppResources, video80),	XtRImmediate, (XtPointer)False},
#endif
 {"speed",	"Speed",	XtRInt,	  sizeof (XtRInt),
#if defined( C128)
      XtOffsetOf (AppResources, speed),		XtRImmediate, (XtPointer)C128_CUSTOM_SPEED},
#elif defined( PET)
      XtOffsetOf (AppResources, speed),		XtRImmediate, (XtPointer)PET_CUSTOM_SPEED},
#elif defined( VIC20)
      XtOffsetOf (AppResources, speed),		XtRImmediate, (XtPointer)VIC20_CUSTOM_SPEED},
#else  /* defined( CBM64) */
      XtOffsetOf (AppResources, speed),		XtRImmediate, (XtPointer)PET_CUSTOM_SPEED},
#endif 
#if !defined( C128)
 {"refreshRate","RefreshRate",  XtRInt, sizeof (XtRInt),
      XtOffsetOf (AppResources, refreshRate),   XtRImmediate, (XtPointer)REFRESH_RATE},
#else
 {"vdcRefreshRate","VdcRefreshRate",  XtRInt, sizeof (XtRInt),
      XtOffsetOf (AppResources, vdcRefreshRate),   XtRImmediate, (XtPointer)VDC_REFRESH_RATE},
 {"vicRefreshRate","VicRefreshRate",  XtRInt, sizeof (XtRInt),
      XtOffsetOf (AppResources, vicRefreshRate),   XtRImmediate, (XtPointer)VIC_REFRESH_RATE},
#endif
 {"mitshm", 	"Mitshm",       XtRBoolean, sizeof (Boolean),
      XtOffsetOf (AppResources, mitshm), 	XtRImmediate, (XtPointer)True}
};


/*
 * Options
 */

static XrmOptionDescRec options[] = {
#if defined(PET) || defined (C128)
    {"+40",		"*video40",	XrmoptionNoArg,		(XPointer)"false"},
    {"-40",	 	"*video40",	XrmoptionNoArg, 	(XPointer)"true"},
    {"+80",		"*video80",	XrmoptionNoArg,         (XPointer)"false"},
    {"-80",	 	"*video80",	XrmoptionNoArg, 	(XPointer)"true"},
#endif
    {"-asm",	 	"*asm", 	XrmoptionNoArg, 	(XPointer)"true"},
    {"-basicName",	"basicName",	XrmoptionSepArg, 	(XPointer)0},
#if defined (C128)
    {"-biosName",	"biosName",	XrmoptionSepArg, 	(XPointer)0},
#endif
#ifdef PATCH_ROM
    {"-brevision",	"*basicRev",	XrmoptionSepArg, 	(XPointer)0},
#endif
#if defined( C128) || defined( CBM64)
    {"+checksb",        "*checkSbColl", XrmoptionNoArg,         (XPointer)"false"},
    {"-checksb",        "*checkSbColl", XrmoptionNoArg,         (XPointer)"true"},
    {"+checkss",        "*checkSsColl", XrmoptionNoArg,         (XPointer)"false"},
    {"-checkss",        "*checkSsColl", XrmoptionNoArg,         (XPointer)"true"},
#endif
/*     {"+colors",		"*colors", 	XrmoptionNoArg, 	(XPointer)"false"}, */
/*     {"-colors",		"*colors", 	XrmoptionNoArg, 	(XPointer)"true"}, */
    {"-debug",	 	"*debug", 	XrmoptionNoArg, 	(XPointer)"true"},
    {"-directory",	"*directory",	XrmoptionSepArg, 	(XPointer)0},
    {"-display",	".display", 	XrmoptionSepArg, 	(XPointer)0},
    {"-exromName",	"exromName",	XrmoptionSepArg, 	(XPointer)0},
    {"-floppy",		"*floppyName",	XrmoptionSepArg, 	(XPointer)0},
    {"-8",		"*floppyName",	XrmoptionSepArg, 	(XPointer)0},
    {"-9",	        "*floppy9Name",	XrmoptionSepArg, 	(XPointer)0},
    {"-fn", 		"*font", 	XrmoptionSepArg, 	(XPointer)0},
    {"-halt",	 	"*halt", 	XrmoptionNoArg, 	(XPointer)"true"},
    {"+hex",	 	"*hex", 	XrmoptionNoArg, 	(XPointer)"false"},
    {"-hex",	 	"*hex",		XrmoptionNoArg, 	(XPointer)"true"},
#ifndef PET
    {"-joyport", 	"*joyPort",	XrmoptionSepArg,        (XPointer)"2"},
    {"-hjoyport",       "*hjoyPort",    XrmoptionSepArg,        (XPointer)"1"},
#endif
    {"-kernelName",	"kernelName",	XrmoptionSepArg, 	(XPointer)0},
    {"-load",		"*programName",	XrmoptionSepArg, 	(XPointer)0},
    {"-module",		"*module",	XrmoptionSepArg, 	(XPointer)0},
    {"-game",		"*module",	XrmoptionSepArg, 	(XPointer)0},
#if defined( MITSHM)
    {"+mitshm",         "*mitshm",	XrmoptionNoArg,		(XPointer)"false"},
    {"-mitshm",         "*mitshm",	XrmoptionNoArg,		(XPointer)"true"},
#endif
/*    {"-name",		"*name",	XrmoptionSepArg, 	(XPointer)0}, */
    {"-printCommand",	"*printCommand",XrmoptionSepArg,	(XPointer)0},
    {"-ramName",	"ramName",	XrmoptionSepArg, 	(XPointer)0},
#if !defined( C128)
    {"-refresh",        "*refreshRate", XrmoptionSepArg,	(XPointer)0},
#endif
#ifdef REU
    {"-reuName",	"*reuName",	XrmoptionSepArg,	(XPointer)0},
#endif
#ifdef PATCH_ROM
    {"-revision",	"*kernelRev",	XrmoptionSepArg, 	(XPointer)0},
#endif
    {"-start",          "*startAddr",	XrmoptionSepArg,        (XPointer)0},
    {"-speed",          "*speed",	XrmoptionSepArg,        (XPointer)0},
    {"-tape",		"*tapeName", 	XrmoptionSepArg, 	(XPointer)0},
    {"-1",              "*tapeName",	XrmoptionSepArg,        (XPointer)0},
    {"-test",	 	"*test", 	XrmoptionNoArg, 	(XPointer)"true"},
    {"+traps",	        "*notraps", 	XrmoptionNoArg, 	(XPointer)"false"},
    {"-traps",		"*notraps", 	XrmoptionNoArg, 	(XPointer)"true"},
#if defined( C128)
    {"-vdcrefresh",     "*vdcRefreshRate", XrmoptionSepArg,     (XPointer)0},
#endif
    {"-verbose",	"*verbose", 	XrmoptionNoArg, 	(XPointer)"true"},
#if defined( C128)
    {"-vicrefresh",     "*vicRefreshRate", XrmoptionSepArg,     (XPointer)0},
#endif
};


/*
 * Xdebugger Resources
 */

String fallback_resources[] = {
    "*fileButton.label:      File",
    "*optionButton.label:    Options",

    "*run.Label:             Run",
    "*run.width:             150",
    "*stop.Label:            Stop",
    "*stop.width:            150",
    "*trace.Label:           Trace",
    "*trace.width:           150",
    "*next.Label:            Next",
    "*next.width:            150",
    "*step.Label:            Step",
    "*step.width:            150",

/*    "*status.width:          150",*/

    "*view.allowVert:        True",
    "*view.forceBars:        True",
    "*view.width:            200",
    "*view.height:           320", 
    "*hbox.orientation:      Horizontal",
    "*hbox.borderWidth:      0",
    "*hbox.min:              28",
    "*hbox.max:              28",
    "*vbox.orientation:      Vertical",
    "*vbox.borderWidth:      0",   /* Registers and buttons */
    "*smeBSB*borderWidth:    1",
    "*Label*borderWidth:     0",   /* Register name boxes */
    "*bar.orientation:       Vertical",
    "*bar.length:            320",
    "*bar.shown:             0.02",

    "*foreground:	     Black",
    "*background:	     Light Gray",
    "*Label.foreground       black",
    "*font:		     fixed",
    NULL
};


/* ------------------------------------------------------------------------- */

void    toggle_cntl_win(void)
{
    static int      debugstate = 0;

    if ((debugstate = !debugstate)) {
	XtRealizeWidget (toplevel);
	XtUnrealizeWidget (nextButton);
	XtUnrealizeWidget (stepButton);
	UpdateRegisters();
	set_asm(PC);
    } else {
	XtUnrealizeWidget (toplevel);
	halt = trace = 0;
    }
}


void    x_loop(void)
{
    XEvent  report;	/* Structure for event information */

#ifdef LED_ARRAY
    static CLOCK c;

    if (ledflg && (clk - c) > (CYCLES_PER_SEC >> 2)) {
	update_ledbar();
	c = clk;
    }
#endif

#if defined (X64SOUND) && defined (CBM64) && defined (LIONEL_SID)
    /* Catch up the sound output and write the filled buffer to the
       audio output */
    SIDemu(1);
#endif /* X64SOUND */

    if (halt || trace) {
	UpdateRegisters();
	set_asm(PC);
    }
/*
    if (redraw) {
	(*vic_redraw)();
	redraw = 0;
    }
#ifndef SETOWN
    flush_queue();
#endif
*/

    /*
     * Stay in this loop while there are events to be served,
     * or the execution is stopped.
     * Trace is also used to flag single-step execution of
     * instruction when halted.
     */

    while ((halt && !trace) || XtAppPending (app_con)) {
	XtAppNextEvent(app_con, &report);
	/* Main Window Event */
	if (report.xany.window == borderwin ||
	    report.xany.window == VDCborderwin) {
	    handle_mainwin_event(&report);
	    continue;
	}

	/* Xdebugger Event */
	if (XtDispatchEvent (&report)) {
	    state = (halt ? DS_Halt : trace ? DS_Trace : DS_Run);
	    UpdateStatus(state);
	    continue;
	}

#ifdef LED_ARRAY
	if (report.xany.window == ledwin) {
	    if (report.type != Expose || report.xexpose.count != 0)
		continue;
	    draw_ledbar();
	}
#endif
    }	/* End while */


    if (trace) {
	setalarm(I_SPECX, 1);	
/*	int_clk[I_SPECX] = clk +1;
	find_next_int();
#ifdef NEW_TIMER_CODE
	update_intr_alarm();
#endif
*/

	if (halt) {		/* halt + trace == Single step excution */
	    if (!next || check_next())
		trace = 0;
	}

	state = (halt ? DS_Halt : trace ? DS_Trace : DS_Run);
	UpdateStatus(state);
    }
}


static void UpdateRegisters (void)
{
    int eff;

    sprintf (buf, "$%02X", AC);
    XtVaSetValues (registers [R_AC],
		   XtNstring, buf, 
		   NULL);

    sprintf (buf, "$%02X", XR);
    XtVaSetValues (registers [R_XR],
                   XtNstring, buf, 
		   NULL);

    sprintf (buf, "$%02X", YR);
    XtVaSetValues (registers [R_YR],
                   XtNstring, buf, 
		   NULL);

    sprintf (buf, "$%02X", SP);
    XtVaSetValues (registers [R_SP],
                   XtNstring, buf, 
		   NULL);

    sprintf (buf, "$%04X", PC);
    XtVaSetValues (registers [R_PC],
                   XtNstring, buf, 
		   NULL);

    sprintf (buf, "$%02X", GET_SR());
    XtVaSetValues (registers [R_SR],
                   XtNstring, buf, 
		   NULL);

    /* Memory */

    eff = eff_address(PC, 1);
    if (eff >= 0)
	sprintf (buf, "$%04X", eff);
    else
	sprintf (buf, "----");

    XtVaSetValues (registers [R_EFF],
                   XtNstring, buf, 
		   NULL);

    if (eff >= 0)
	sprintf (buf, "$%02X", LOAD(eff));
    else
	sprintf (buf, "--");

    XtVaSetValues (registers [R_DL],
		   XtNstring, buf, 
		   NULL);

    /* Clock */

    sprintf (buf, "%ld", clk);
    XtVaSetValues (registers [R_CLK],
		   XtNstring, buf, 
		   NULL);
}


/* Debugger Status */

static void UpdateStatus (enum DState dst)
{
    static int s = -1;
    if (s == (int)dst)
	return;

    XtVaSetValues (status1,
		   XtNlabel, StatusLabels[dst],
		   NULL);
    s = dst;
}


/* ------------------------------------------------------------------------- */

static void QuitCb(Widget w, XtPointer clientData, XtPointer callData)
{
    end64 ();
}


/* ------------------------------------------------------------------------- */

/*
 * Files
 *  call fileselector window ...
 */

static void LoadCb(Widget w, XtPointer clientData, XtPointer callData)
{

}


static void DiskCb(Widget w, XtPointer clientData, XtPointer callData)
{

    if (serial_select_file(DT_DISK, 8, fileselector(0)) < 0) {
	fprintf (stderr, "No Disk.\n");
    }
}


static void TapeCb(Widget w, XtPointer clientData, XtPointer callData)
{

    if (serial_select_file(DT_TAPE, 1, fileselector(0)) < 0) {
	fprintf (stderr, "No Tape.\n");
    }
}


/* ------------------------------------------------------------------------- */

static void RunCb(Widget w, XtPointer clientData, XtPointer callData)
{
    halt = trace = next = 0;

    XtUnrealizeWidget(nextButton);
    XtUnrealizeWidget(stepButton);

    RESTART_VIDEO;
    
    UpdateRegisters();
    set_asm(PC);
}

static void StopCb(Widget w, XtPointer clientData, XtPointer callData)
{
    halt = 1;
    trace = next = 0;

    PAUSE_VIDEO;

    XtRealizeWidget(nextButton);
    XtManageChild(nextButton);
    XtRealizeWidget(stepButton);
    XtManageChild(stepButton);

    XFlush (display);
    
    UpdateRegisters();
    set_asm(PC);
}


/* FIXME: Instead of asm(), Trace should display the branches encountered. */

static void TraceCb(Widget w, XtPointer clientData, XtPointer callData)
{
    trace = 1;
    halt = next = 0;

    XtUnrealizeWidget(nextButton);
    XtUnrealizeWidget(stepButton);

    RESTART_VIDEO;
    
    UpdateRegisters();
    set_asm(PC);
}

static void StepCb(Widget w, XtPointer clientData, XtPointer callData) 
{
    trace = halt = 1;
    next = 0;

    UpdateRegisters();
    set_asm(PC);
}

static BYTE next_sp = 0xFF;

static void NextCb(Widget w, XtPointer clientData, XtPointer callData)
{
    BYTE  src  = LOAD(PC +1);


   /*
    * Halt, Trace: if "Next" is pressed when JSR is to be excuted, tracing
    *   is turned off until exution returns back, or the entry is removed
    *   from stack.
    * Halt: If a conditional branch backwards is encountered, turns on Run
    *   mode until ececution has passed the loop. Elsewhere, works like "Step".
    * Trace: Skip the rest of subroutine being traced.
    * This button has no use in Run mode.
    */

    if (halt) {					/* Halted */
	switch (LOAD(PC)) {
	  case 0x20:	/* JSR */
	    next_sp = SP;
	    next = 1;
	    break;


	    /* BRANCH instructions. */

	  case 0x10:	/* BPL */
	    loop_test (!IF_SIGN(), src, 1);
	    break;

	  case 0x30:	/* BMI */
	    loop_test (IF_SIGN(), src, 1);
	    break;

	  case 0x50:	/* BVC */
	    loop_test (!IF_OVERFLOW(), src, 1);
	    break;

	  case 0x70:	/* BVS */
	    loop_test (IF_OVERFLOW(), src, 1);
	    break;

	  case 0x90:	/* BCC */
	    loop_test (!IF_CARRY(), src, 1);
	    break;

	  case 0xB0:	/* BCS */
	    loop_test (IF_CARRY(), src, 1);
	    break;

	  case 0xD0:	/* BNE */
	    loop_test (!IF_ZERO(), src, 1);
	    break;

	  case 0xF0:	/* BEQ */
	    loop_test (IF_ZERO(), src, 1);
	    break;

	  default:
	    trace = 1;		/* Single Step */
#ifdef DEBUG
	    printf ("single-step\n");
#endif
	}  /* switch */

    }  /* halt */

    else {
	if (trace) {				/* Tracing */
	    /*trace = 0;*/
	}
	else {					/* Running */
	    printf("Run/Stop-Restore.\n");
	    restore();
	}
    }

    RESTART_VIDEO;
    
    UpdateRegisters();
    set_asm(PC);
}


/* ------------------------------------------------------------------------- */

/* Check if we've done the program structure
 * returns nonzero when the next instruction is reached.
 */

int  check_next (void)
{
    BYTE  src  = LOAD(PC +1);


    switch (LOAD(PC)) {

      case 0x40:	/* RTI */
      case 0x60:	/* RTS */
      case 0x68:	/* PLA */
	if (SP > next_sp)
	    next = 0;
	break;

      case 0x9A:	/* TXS */
	if (XR > next_sp)
	    next = 0;
	break;


	/* BRANCH instructions. */

      case 0x10:	/* BPL */
	loop_test (!IF_SIGN(), src, 0);
	break;

      case 0x30:	/* BMI */
	loop_test (IF_SIGN(), src, 0);
	break;

      case 0x50:	/* BVC */
	loop_test (!IF_OVERFLOW(), src, 0);
	break;

      case 0x70:	/* BVS */
	loop_test (IF_OVERFLOW(), src, 0);
	break;

      case 0x90:	/* BCC */
	loop_test (!IF_CARRY(), src, 0);
	break;

      case 0xB0:	/* BCS */
	loop_test (IF_CARRY(), src, 0);
	break;

      case 0xD0:	/* BNE */
	loop_test (!IF_ZERO(), src, 0);
	break;

      case 0xF0:	/* BEQ */
	loop_test (IF_ZERO(), src, 0);

    }  /* switch */


    return (!next);
}


void loop_test (int cc, BYTE src, int set)
{
    static ADDRESS  next_begin = 0, next_end  = 0;
    static int      next_cnt = 0;


    if (src & 0x80) {

	if (set) {
	    if (cc) {
		next_begin = REL_ADDR(PC, src);
		next_end   = PC;
		next_cnt = 0;
		++next;
#ifdef DEBUG
		printf ("pc %04x  cc %d  -- set %d loop %04x %04x  SET LOOP\n",
			PC, cc, set, next_begin, next_end);
#endif
	    }

	    trace = 1;			/* Excute instruction */
	}
	else if (PC == next_end) {	/* Don't set but test only */
	    if (cc) {
		++next_cnt;
	    }
	    else {
		printf ("Loop  %04x %04x  executed %d times.\n",
			next_begin, next_end, next_cnt);
		next_begin = 0;
		next_end   = 0;
		next_cnt = 0;
		next = 0;
	    }
	}
    }    /* offset */
    else
	trace = 1;			/* Do single-step instead */

}


/* ------------------------------------------------------------------------- */

/*
 * Assembly Window
 */

static void ThumbProc (Widget bar, XtPointer client, XtPointer ptr)
{
    float percent = *(float*) ptr;
    ADDRESS p = ((int) (percent * 0xffff)) & 0xffff; 
    /*printf ("thumb %f\n", percent);*/
    draw_asm (p);
    return;
}


static void ScrollProc (Widget bar, XtPointer client, XtPointer ptr)
{
    int position = (int) ptr;

    /*printf ("position %d\n", position);*/
    if (position < 0) {
	draw_asm (current_point + 0x10);
    } else {
	draw_asm (current_point - 0x10);
    }
    return;
}


void set_asm (/*Widget bar,*/ ADDRESS p)
{
    float percent = (p & 0xffff) / 65535.0;
    XawScrollbarSetThumb (ASMbar, percent, 0.02);
    draw_asm (p);
}


void draw_asm (ADDRESS p)
{
    Widget  sink;
    char    buf2[8192];
    short   lines, height;  /* NOT ints */
    int i;


    buf2[0] = '\0';
    current_point = p;

    XtVaGetValues (view, 
	 XtNheight, &height, 
	 XtNtextSink, &sink,
	 NULL);
    lines = XawTextSinkMaxLines (sink, height);


    for (i = 0; i <= lines; i++) {
	p &= 0xffff;
	sprintf (buf, "%04X %s  %-s\n", p, 
		 sprint_ophex (p), sprint_opcode(p, 1));
	strcat (buf2, buf);
	p += clength[lookup[LOAD(p)].addr_mode];
    }

    XtVaSetValues (view,
                   XtNstring, buf2,
		   NULL);
}


/* ------------------------------------------------------------------------- */

static void MenuBar (Widget parent)
{
    Widget hbox, fileMenu, optionMenu, shell, use;


    hbox = XtVaCreateManagedWidget 
	("hbox", boxWidgetClass, parent, 
	 NULL);
    
    fileMenu = XtVaCreateManagedWidget 
	("fileButton", menuButtonWidgetClass, hbox,
	 XtNmenuName,"fileMenu",
	 NULL);

    shell = XtVaCreatePopupShell 
	("fileMenu", simpleMenuWidgetClass, fileMenu, 
	 NULL);

    use = XtVaCreateManagedWidget
        ("load", smeBSBObjectClass, shell,
         NULL);
    /*XtAddCallback(use, XtNcallback, LoadCb, (XtPointer)0);*/

#ifndef NO_SERIAL
    use = XtVaCreateManagedWidget
        ("disk", smeBSBObjectClass, shell,
         NULL);
    XtAddCallback(use, XtNcallback, DiskCb, (XtPointer)0);
#endif
#ifdef CBMTAPE
    use = XtVaCreateManagedWidget
        ("tape", smeBSBObjectClass, shell,
         NULL);
    XtAddCallback(use, XtNcallback, TapeCb, (XtPointer)0);
#endif

    use = XtVaCreateManagedWidget
	("quit", smeBSBObjectClass, shell,
         NULL);
    XtAddCallback (use, XtNcallback, QuitCb, (XtPointer)0);


    optionMenu = XtVaCreateManagedWidget 
	("optionButton", menuButtonWidgetClass, hbox,
	 XtNmenuName,"optionMenu",
	 NULL);

    shell = XtVaCreatePopupShell 
	("optionMenu", simpleMenuWidgetClass, optionMenu, 
	 NULL);

    use = XtVaCreateManagedWidget
        ("attributes",smeBSBObjectClass, shell,
         NULL);
    /*XtAddCallback(use, XtNcallback, AttributesCb, (XtPointer)self);*/

    XtVaCreateManagedWidget
        ("refresh",smeBSBObjectClass, shell,
         NULL);
    /*XtAddCallback (refresh, XtNcallback, RefreshCb, (XtPointer)self);*/

#ifdef LED_ARRAY
    use = XtVaCreateManagedWidget
        ("activity",smeBSBObjectClass, shell,
         NULL);
    XtAddCallback (use, XtNcallback, LedCb, (XtPointer)0);
#endif
}


static void RegisterBar (Widget parent)
{
    int i;
    Widget form;
    Widget label, lastlabel;

    form = parent
	/* XtVaCreateManagedWidget("form", formWidgetClass, parent, NULL) */;

    /*
     * Registers
     */

    for (lastlabel = NULL, i = 0; i < NO_REGISTERS; i++, lastlabel = label) {
	label = XtVaCreateManagedWidget (
	     "register", labelWidgetClass, form, 
	     XtNlabel, registerLabels[i], 
	     XtNfromVert, lastlabel,
	     NULL);
	registers[i] = XtVaCreateManagedWidget (
	     "value",asciiTextWidgetClass, form,
	     XtNtype,XawAsciiString,
	     XtNeditType,XawtextEdit,
	     XtNfromHoriz, label,
	     XtNfromVert, lastlabel,
	     NULL);
    }

    /*
     * Debugger Status
     */

    status1 = XtVaCreateManagedWidget (
	"status", labelWidgetClass, form, 
	XtNlabel, StatusLabels[DS_Run],
	XtNfromVert, lastlabel,
	NULL);

    /*
     * Buttons
     */

    runButton = XtVaCreateManagedWidget (
	"run", commandWidgetClass, form,
	XtNfromVert, status1,
	NULL);
    XtAddCallback(runButton, XtNcallback, RunCb, NULL);

    stopButton = XtVaCreateManagedWidget (
	"stop", commandWidgetClass, form,
	 XtNfromVert, runButton,
	 NULL);
    XtAddCallback(stopButton, XtNcallback, StopCb, NULL);

    traceButton = XtVaCreateManagedWidget (
	"trace", commandWidgetClass, form,
	 XtNfromVert, stopButton,
	 NULL);
    XtAddCallback(traceButton, XtNcallback, TraceCb, NULL);

    nextButton = XtVaCreateManagedWidget (
	"next", commandWidgetClass, form,
	XtNfromVert, traceButton,
	NULL);
    XtAddCallback(nextButton, XtNcallback, NextCb, NULL);

    stepButton = XtVaCreateManagedWidget (
	"step", commandWidgetClass, form,
	XtNfromVert, nextButton,
	NULL);
    XtAddCallback(stepButton, XtNcallback, StepCb, NULL);
}


/* ------------------------------------------------------------------------- */

/*
 *  Parse arguments and X resources
 */

void   arg_parser(int argc, char **argv)
{
    toplevel = XtAppInitialize(&app_con, "X64", options, XtNumber(options),
			       &argc, argv, fallback_resources, NULL, ZERO);

    if (argc != 1)
	Syntax(app_con, argv[0]);

    XtGetApplicationResources(toplevel, (XtPointer) &app_resources,
			      resources, XtNumber(resources), NULL, ZERO );
}


void   debug_main(void)
{
    Widget form, pane;
    Widget vbox;


    /*
     * Vertical Pane of widgets 
     */
    pane = XtCreateManagedWidget("pane", panedWidgetClass, toplevel, NULL, 0);

    /*
     * Menu bar
     */
    MenuBar (pane);

    form = XtVaCreateManagedWidget("form", formWidgetClass, pane, NULL);

    /*
     * Assembly listing
     */

    ASMbar = XtVaCreateManagedWidget("bar", scrollbarWidgetClass, form, NULL);
    XtAddCallback(ASMbar, XtNjumpProc, ThumbProc, NULL);
    XtAddCallback(ASMbar, XtNscrollProc, ScrollProc, NULL);

    view = XtVaCreateManagedWidget
	("view", asciiTextWidgetClass, form, 
	 XtNfromHoriz, ASMbar,
	 NULL);

    vbox = XtVaCreateManagedWidget 
	("vbox", formWidgetClass, form,
	 XtNfromHoriz, view,
	 NULL);

    RegisterBar (vbox);

    display = XtDisplay (toplevel);
    /* XtRealizeWidget(toplevel); */
    /* XtAppMainLoop(app_con); */


#ifdef LED_ARRAY
    /*
     * Bus Activity LEDs Window
     */

    open_led_win();
    /*led = XtCreateManagedWidget("led", panedWidgetClass,
			      toplevel, NULL, 0);*/
#endif
}


static void  Syntax (XtAppContext appc, char *call)
{
    int i, ccount = 0, len = 0;
    char optstring [40];

    XtDestroyApplicationContext(appc);

    fprintf(stderr, "\nUsage: %s\n    ", call);
    ccount = 4;
    for (i = 0; i < XtNumber(options); i++) {
#if 0
        if (options[i].argKind == XrmoptionNoArg) {
	    fprintf(stderr, "    [ %s ]\n", options[i].option);
	} else {
	    fprintf(stderr, "    [ %s <value>]\n", options[i].option);
	}
#endif
	if (i < XtNumber(options))
	    if (options[i].option[0] == '+') {
	        len = sprintf (optstring, "[-/%s", options[i].option);
		i++;
	    }
	else
	    len = sprintf (optstring, "[%s", options[i].option);
	if (options[i].argKind == XrmoptionNoArg) {
	    strcat (optstring, "] ");
	    len += 2;
	} else {
	    strcat (optstring, " <value>] ");
	    len += 10;
	}
	if ((ccount += len) > 80) {
            fprintf (stderr, "\n    %s", optstring);
	    ccount = strlen (optstring) + 4;
	} else 
	    fprintf (stderr, "%s", optstring);
    }
    fprintf(stderr, "\n\nAll options may be abbreviated.\n\n");
    exit(1);
}


/* ------------------------------------------------------------------------- */

/*
 *  LED Bar - a simple toy.
 */

#ifdef LED_ARRAY

#define NUM_LEDS	16
#define LED_DIM		2	/* cycles */
#define LED_HALF	(CYCLES_PER_SEC >> 5)
#define LED_BRIGHT	(CYCLES_PER_SEC >> 4)

#define LED_X		10
#define LED_Y		20
#define LEDX_SPC	15
#define XOFF		20
#define YOFF		20
#define LED_WID		(LEDX_SPC * NUM_LEDS + 2 * XOFF)
#define LED_HEI		(LED_Y + 40 + 2 * YOFF)


static GC      led_gc[4];
static GC      debuggc;


/* fonts */

static XFontStruct *font_info;	/* Structure containing font info */
static int     font_height;
static int     font_width;

static char   *font_list[] = {
    "-*-helvetica-bold-r-*-*-12-*",
    "-*-courier-bold-r-*-*-12-*",
    "-*-times-bold-r-*-*-12-*",
    "fixed",
    NULL,
};


static void LedCb(Widget w, XtPointer clientData, XtPointer callData)
{
    ledflg = toggle_ledbar();
}

/* ------------------------------------------------------------------------- */

static void open_led_win (void)
{
    Window    rootw;
    XSizeHints size_hints;
    XTextProperty windowName, iconName;
    XWMHints  wm_hints;
    XClassHint class_hints;
    XGCValues gcval;
    XColor    exact_def, rgb_def;
    Colormap  def_cmap = DefaultColormap(display, screen);
    int       i;

    char  **p = font_list;
    char   *win_name = "Bus Activity";
    char   *icon_name = "LEDs";


    /* win */

    size_hints.min_width = size_hints.width = LED_WID;
    size_hints.min_height = size_hints.height = LED_HEI;
    size_hints.flags = PSize | PMinSize;

    rootw = RootWindow(display, screen);
    ledwin = XCreateSimpleWindow(display, rootw, 0, 0,
	size_hints.width, size_hints.height, 1,
	BlackPixel(display, screen),
	WhitePixel(display, screen));

    if (debugflg)
	printf("LED Window open\n");

    XSetStandardProperties(display, ledwin, win_name, icon_name,
	None, NULL, 0, &size_hints);

    /* Select event types wanted */
    XSelectInput(display, ledwin,
	ExposureMask | StructureNotifyMask | FocusChangeMask);

    if (XStringListToTextProperty(&win_name, 1, &windowName) == 0) {
	(void) fprintf(stderr,
	    "%s: structure allocation for windowName failed.\n",
	    progname);
	exit(-1);
    }
    if (XStringListToTextProperty(&icon_name, 1, &iconName) == 0) {
	(void) fprintf(stderr,
	    "%s: structure allocation for iconName failed.\n",
	    progname);
	exit(-1);
    }

    wm_hints.initial_state = NormalState;
    wm_hints.input = True;
    wm_hints.flags = StateHint | InputHint;

    class_hints.res_name = progname;
    class_hints.res_class = "C64";

    XSetWMProperties(display, ledwin, &windowName, &iconName, NULL, 0,
	&size_hints, &wm_hints, &class_hints);


    /* Allocate colors for images on screen */

    gcval.foreground = BlackPixel(display, screen);
    gcval.background = WhitePixel(display, screen);
    debuggc = XCreateGC(display, rootw, GCForeground | GCBackground, &gcval);


    if (!led_gc[0]) {
	char *lc_names[] = { "red4", "red3", "red2", "red" };
	/*char *lc_names[] = { "blue", "red4", "yellow", "green" };*/

	if (DefaultDepth(display, screen) < 4) {
	    fprintf (stderr, "Cannot create LED bar.\n");
	    return;
	}

	/* Allocate read only colors for each LED brightness */

	for (i = 0; i < 4; ++i) {
	    if (!XAllocNamedColor(display, def_cmap,
				  lc_names[i], &exact_def, &rgb_def)) {
		fprintf(stderr, "Cannot allocate color %s\n", lc_names[i]);
		return;
	    }

	    gcval.foreground = exact_def.pixel;
	    gcval.background = WhitePixel(display, screen);
	    led_gc[i] =
		XCreateGC(display, rootw, GCForeground | GCBackground, &gcval);
	}  /* for */
    }


    /* font */

    while (p && *p && !(font_info = XLoadQueryFont(display, *p))) ++p;
    if (!p) {
	(void) fprintf(stderr, "%s: Fatal: No fonts available.\n", progname);
	exit(-1);
    }
    font_height = font_info -> ascent + font_info -> descent;
    font_width = font_info -> max_bounds.width;


    XSetFont(display, debuggc, font_info -> fid);
}


static int  toggle_ledbar (void)
{
    static int  st = 0;

    if (!st) XMapWindow (display, ledwin);
    else XUnmapWindow(display, ledwin);

    st = !st;
    return (st);
}


static void draw_ledbar (void)
{
    int  i;


    for (i = 0; i < NUM_LEDS; ++i)
	XDrawImageString (display, ledwin, debuggc,
			  XOFF + i * LEDX_SPC + 2,
			  YOFF + LED_Y + 15 /*font->ascent*/ +2,
			  "0123456789ABCDEF" + i, 1);

    XDrawImageString (display, ledwin, debuggc,
		      XOFF * 2, YOFF + LED_Y + 40 /*font->ascent*/ +2,
		      "VIC-74154 Virtual Bus Tester", 30);

    update_ledbar ();
}


static void update_ledbar (void)
{
    GC   gcl;
    int  i, x = XOFF;
    int *r;


    if (!ledwin)
	return;


    for (i = 0, r = &ledarr[0]; i < NUM_LEDS; ++i, ++r) {

	if (*r < LED_DIM)
	    gcl = led_gc[0];
	else if (*r < LED_HALF)
	    gcl = led_gc[1];
	else if (*r < LED_BRIGHT)
	    gcl = led_gc[2];
	else
	    gcl = led_gc[3];

	XFillRectangle (display, ledwin, gcl, x, YOFF, LED_X, LED_Y);

	x += LEDX_SPC;
	*r = 0;

    }  /* for */
}
#endif  /* LED_ARRAY */


/*****************************************************************************
 * Here are the new popup menus
 */

Widget  shell;
int     ended;		/* 1 = end, 2 = mon() */


/* Action Popup Menu **********************************************/

typedef struct actionList {
	void (*func)(Widget w, XtPointer clientData, XtPointer callData);
	void *misc;
	char *text;
} actionList;

void actSoftReset(Widget w, XtPointer clientData, XtPointer callData) {
	PC = LOAD_ADDR(0xfffc);
	ended = 1;
}

void actHardReset(Widget w, XtPointer clientData, XtPointer callData) {
	PC = intr(I_RESET, PC);
	ended = 1;
}

void actPowerOnReset(Widget w, XtPointer clientData, XtPointer callData) {
#ifdef CBM64
	ram[0x8002] = 0;
#endif	
	PC = intr(I_RESET, PC);
	ended = 1;
}

void actNMI(Widget w, XtPointer clientData, XtPointer callData) {
	cpu_do_nmi();
	ended = 1;
}

void actQuit(Widget w, XtPointer clientData, XtPointer callData) {
	end64();
}

void actMonitor(Widget w, XtPointer clientData, XtPointer callData) {
	ended = 2;	/* Monitor */
}


void x_mon(void)
{
        XAutoRepeatOn(display);
        /* -- drop to monitor; continue with x or C-d */
        PAUSE_VIDEO;
        XSync(display, 1);

        mon(PC);
        RESTART_VIDEO;
#ifndef AUTOREPEAT
        XAutoRepeatOff(display);
#endif
}

void actDebug(Widget w, XtPointer clientData, XtPointer callData) {
	toggle_cntl_win();
	ended = 1;
}

#ifdef LED_ARRAY
void actLedBar(Widget w, XtPointer clientData, XtPointer callData) {
	toggle_ledbar();
	ended = 1;
}
#endif

void actDisk8(Widget w, XtPointer clientData, XtPointer callData) {
	DiskCb(w, clientData, callData);
	ended = 1;
}

void actTape1(Widget w, XtPointer clientData, XtPointer callData) {
	TapeCb(w, clientData, callData);
	ended = 1;
}

actionList acts[] = {
	{ actTape1, NULL, "Attach tapeimage (device 1)" },
#ifdef PET
	{ NULL, NULL, "Attach tapeimage (device 2)" },
#endif
	{ actDisk8, NULL, "Attach diskimage (device 8)" },
	{ NULL, NULL, "Attach diskimage (device 9)" },
	{ NULL, NULL, "Change fs dir (device 11)" },
	{ NULL, NULL, "---------------------"},
	{ actMonitor, NULL, "Activate monitor" },
	{ actDebug, NULL, "xdebugger" },
#ifdef LED_ARRAY
	{ actLedBar, NULL, "LED bar" },
#endif
	{ NULL, NULL, "---------------------"},
	{ actSoftReset, NULL, "SoftReset" },
	{ actHardReset, NULL, "HardReset" },
	{ actPowerOnReset, NULL, "PowerOn" },
	{ actNMI, NULL, "NMI" },
	{ NULL, NULL, "---------------------"},
	{ actQuit, NULL, "Exit" },
	{ NULL }
};

String PopupTranslationTable =
{
#if 0
  "<BtnMotion>:		highlight()\n\
   <BtnUp>:		notify() unhighlight()\n"
#else
  "<EnterWindow>:	highlight()\n\
   <LeaveWindow>:	unhighlight()\n\
   <Motion>:		highlight()\n\
   <BtnDown>:		notify() unhighlight()\n"
#endif
};


void open_action_popup(Window root, XEvent * event_sent, int x, int y) {
    XEvent event;
    Widget shell;
    Widget use;
    actionList *tab;
    Cursor cursor = XCreateFontCursor (XtDisplay (toplevel), XC_left_ptr);

    ended = 0;

    do
	XtAppNextEvent (app_con, &event);
    while (event.type != ButtonRelease);
    
    shell = XtVaAppCreateShell
  	("actionPopup", "X64", simpleMenuWidgetClass, XtDisplay (toplevel),
 	  XtNx, x-30, XtNy, y - 5, XtNcursor, cursor,
 	  NULL); 

    tab = acts;

    while(tab->text) {
        if (*tab->text != '-')
	  use = XtVaCreateManagedWidget
	    (tab->text, smeBSBObjectClass, shell,
	     NULL);
	else
	  use = XtVaCreateManagedWidget
	    (tab->text, smeLineObjectClass, shell,
	     NULL);
        if(tab->func) 
	  XtAddCallback(use, XtNcallback, tab->func, (XtPointer)tab->misc);

        tab++;
    }

    XtVaSetValues (shell, XtNtranslations, XtParseTranslationTable (PopupTranslationTable), NULL);
    XtRealizeWidget (shell);
    
    while(!ended) {
      XtAppNextEvent(app_con, &event);

      if (event.type == ButtonPress && event.xany.window != XtWindow (shell))
	ended = 1;
      else
	XtDispatchEvent(&event);
    }

    XtDestroyWidget(shell);

    if (ended == 2)
	x_mon();
}


/* Toggle Popup Menu **********************************************/

static void toggleCb(Widget w, XtPointer clientData, XtPointer callData)
{
	struct ms_vartab *tab = (struct ms_vartab*) clientData;
#if 0
	printf("toggleCb callback(clientData=%p, callData=%p)\n",
						clientData, callData);
#endif
	if(tab->type == TOGT_BOOL) 
	  *(tab->variable) = ! *(tab->variable);
	else if(tab->type == TOGT_INT) {
	  if((tab->max_val - tab->min_val)==1) {
	    *(tab->variable) += 1 - 2*( *(tab->variable) - tab->min_val );
	  } else {
	    printf("int change not implemented!\n");
	  }
	}
	if(tab->set_func) tab->set_func(*tab->variable);
	ended = 1;
}


void open_toggle_popup(Window root, int x, int y) {
    XEvent event;
    Widget shell;
    Widget use;
    struct ms_vartab *tab;
    char buf[100];
    Cursor cursor = XCreateFontCursor (XtDisplay(toplevel), XC_left_ptr);

    ended = 0;

    do
	XtAppNextEvent (app_con, &event);
    while (event.type != ButtonRelease);

    shell = XtVaAppCreateShell 
	("togglePopup", "X64", simpleMenuWidgetClass, XtDisplay(toplevel), 
	  XtNx, x-30, XtNy, y - 5, XtNcursor, cursor, NULL);

    tab = mon_vars;

    while(tab->name) {
      if(tab->type) {
	if(tab->type==TOGT_BOOL) sprintf(buf,"%-20s=%s",tab->name,*(tab->variable)?"On":"Off");
	if(tab->type==TOGT_INT) sprintf(buf,"%-20s=%d",tab->name,*(tab->variable));
        use = XtVaCreateManagedWidget
          (buf, smeBSBObjectClass, shell,
           NULL);
        XtAddCallback(use, XtNcallback, toggleCb, (XtPointer)tab);

        tab++;
      }
    }

    XtVaSetValues (shell, XtNtranslations, XtParseTranslationTable (PopupTranslationTable), NULL);
    XtRealizeWidget (shell);

    while(!ended) {
      XtAppNextEvent(app_con, &event);

      if( ((event.type == ButtonPress) || (event.type == ButtonRelease))
		&& (event.xany.window == root) )
	ended = 1;

      XtDispatchEvent(&event);
    }
    XtDestroyWidget(shell);
}

