/*
 * $Id: zasm.c,v 1.2 1996/07/29 23:40:42 jopi Exp ettore $
 *
 * This file is part of Commodore 128 emulator
 *	and Program Development System.
 *
 *   Copyright (C) 1995-1996, Jouko Valta
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *
 * Zilog Z80 machine language disassembler.
 *
 * Original instruction table generator in MSX BASIC by Tuomo Sajaniemi, 1986.
 * Ported to CBM BASIC V7.0, MCMLXXXIX
 * Why to make yet another Z80 disassembler, one might ask. Well, the answer
 * is simple: none of the available ones is perfect. Also, we need on-line
 * interpreter as well as two-processor support.
 *
 *
 * Written by
 *   Jouko Valta  (jopi@stekt.oulu.fi)
 *
 *
 * $Log: zasm.c,v $
 * Revision 1.2  1996/07/29 23:40:42  jopi
 * Small bug fixed.
 *
 * Revision 1.1  1996/04/01  08:25:07  jopi
 * Initial revision
 *
 *
 */


#include <stdio.h>
#include <string.h>
#include <ctype.h>

#ifndef NO_ASM
#include "asm.h"
#include "mshell.h"
#endif

#define MAXADDR  0xffff


typedef unsigned char BYTE;
typedef unsigned short  ADDRESS;


extern void initialize ( int test );
extern int  disassemble_line ( int counter, BYTE *inbuf );
extern int  sprint_instr( int counter, BYTE *inbuf, char *buf );
extern int  parse_z80_instr ( char *line, ADDRESS adr, char *arg, int mode );

#ifndef NO_ASM
static int  parse_arg ( char *p, char *q, ADDRESS adr, int code, BYTE *arg);
#endif

/* ------------------------------------------------------------------------- */

struct raw1 {
    char  mnem[13];
    short code;
};

struct raw2 {
    char  mnem[10];
    char  rgr;
    char  post;
    short code;
    char  b;		/* Instruction spacing */
};


/*
 * Operands
 *
 * Shift Opcodes DD and FD:
 *	h   ->	IXh	IYh
 *	l   ->	IXl	IYl
 *	hl  ->	IX	IY
 *     (hl) ->	IXh+d	IYh+d
 * Note: if both HL and H or L are present, only HL is replaced.
 */

#define C_SPACE		' '
#define C_BYTE		'%'
#define C_WORD		'#'
#define C_REL		'$'

#define C_HXY		'>'
#define C_LXY		'<'
#define C_HLXY		'*'
#define C_MHLXY		'@'


enum { OP_R, OP_1, OP_2, OP_3, OP_4, OP_5, OP_6, OP_7, OP_8 };

static int   id[]= { 7, 7, 3, 3, 1, 7, 3, 7, 7 };

static char *regn[][8] = {
    { "b", "c", "d", "e", ">", "<", "",  "a" },		/* r  */
    { "b", "c", "d", "e", ">", "<", "(@)","a" },	/* s  */
    { "bc","de","*","af","",  "",  "",  ""  },		/* ss */
    { "bc","de","*","sp","",  "",  "",  ""  },		/* dd */
    { "bc","de","",  "",  "",  "",  "",  ""  },		/*    */
    { "nz","z", "nc","c", "po","pe","p", "m" },		/* cc */
    { "nz","z", "nc","c", "",  "",  "",  ""  },		/*    */
    { "0", "8", "10","18","20","28","30","38" },	/* L  */
    { "0.","1.","2.","3.","4.","5.","6.","7." }		/* sb */
};


static char *hxy1[] = { "h",  "ixh", "iyh" };
static char *hxy2[] = { "l",  "ixl", "iyl" };
static char *hxy3[] = { "hl", "ix",  "iy" };


/*
 * Instructions
 */


#define N_CMDS1		66
#define N_CMDS2		38


/* Single Commands */

static struct raw1 data1[] = {
    { "adc  a,%",	206 },
    { "add  a,%",	198 },
    { "and  %",		230 },
    { "call #",		205 },
    { "ccf",		 63 },
    { "cp   %",		254 },
    { "cpd",		681 },
    { "cpdr",		697 },
    { "cpi",		673 },
    { "cpir",		689 },
    { "cpl",		 47 },
    { "daa",		 39 },
    { "di",		243 },
    { "djnz $",		 16 },
    { "ei",		251 },
    { "ex   (sp),*",	227 },		/* ex (sp),hl) , ex hl,(sp) */
    { "ex   af,af'",	  8 },
    { "ex   de,*",	235 },		/* ex de,hl) */
    { "exx",		217 },
    { "halt",		118 },
    { "im   0",		582 },
    { "im   1",		598 },
    { "im   2",		606 },
    { "in   a,(%)",	219 },
    { "ind",		682 },
    { "indr",		698 },
    { "ini",		674 },
    { "inir",		690 },
    { "jp   #",		195 },
    { "jp   (*)",	233 },		/* Not indirect. Alias  ld pc,hl */
    { "jr   $",		 24 },
    { "ld   (#),a",	 50 },
    { "ld   (#),hl",	 34 },
    { "ld   (@),%",	 54 },		/* ld (hl),n */
    { "ld   a,(#)",	 58 },
    { "ld   a,i",	599 },
    { "ld   a,r",	607 },
    { "ld   *,(#)",	 42 },
    { "ld   i,a",	583 },
    { "ld   r,a",	591 },
    { "ld   sp,*",	249 },
    { "ldd",		680 },
    { "lddr",		696 },
    { "ldi",		672 },
    { "ldir",		688 },
    { "neg",		580 },
    { "nop",		  0 },
    { "or   %",		246 },
    { "otdr",		699 },
    { "otir",		691 },
    { "out  (%),a",	211 },
    { "outd",		683 },
    { "outi",		675 },
    { "ret",		201 },
    { "reti",		589 },
    { "retn",		581 },
    { "rla",		 23 },
    { "rlca",		  7 },
    { "rld",		623 },
    { "rra",		 31 },
    { "rrca",		 15 },
    { "rrd",		615 },
    { "sbc  a,%",	222 },
    { "scf",		 55 },
    { "sub  %",		214 },
    { "xor  %",		238 }
};



/* Multiple Commands */

static char *oper[] = { "", ",%", ",#", ",$", ",(#)", ",(c)", "),a", ")" };

static struct raw2 data2[] = {
    { "adc  a,",	OP_1,0,	136, 1 },
    { "adc  *,",	OP_3,0,	586,16 },
    { "add  a,",	OP_1,0,	128, 1 },
    { "add  *,",	OP_3,0,	  9,16 },
    { "and  ",		OP_1,0,	160, 1 },
    { "call ",		OP_5,2,	196, 8 },
    { "cp   ",		OP_1,0,	184, 1 },
    { "dec  ",		OP_1,0,	  5, 8 },
    { "dec  ",		OP_3,0,	 11,16 },
    { "in   ",		OP_R,5,	576, 8 },
    { "inc  ",		OP_1,0,	  4, 8 },
    { "inc  ",		OP_3,0,	  3,16 },
    { "jp   ",		OP_5,2,	194, 8 },
    { "jr   ",		OP_6,3,	 32, 8 },
    { "ld   ",		OP_3,2,	  1,16 },
    { "ld   ",		OP_1,1,	  6, 8 },
    { "ld   ",		OP_3,4,	587,16 },
    { "ld   (",		OP_4,6,	  2,16 },
    { "ld   (#),",	OP_3,0,	579,16 },
    { "ld   a,(",	OP_4,7,	 10,16 },
    { "or   ",		OP_1,0,	176, 1 },
    { "out  (c),",	OP_R,0,	577, 8 },
    { "pop  ",		OP_2,0,	193,16 },
    { "push ",		OP_2,0,	197,16 },
    { "ret  ",		OP_5,0,	192, 8 },
    { "rl   ",		OP_1,0,	272, 1 },
    { "rlc  ",		OP_1,0,	256, 1 },
    { "rr   ",		OP_1,0,	280, 1 },
    { "rrc  ",		OP_1,0,	264, 1 },
    { "rst  ",		OP_7,0,	199, 8 },
    { "sbc  a,",	OP_1,0,	152, 1 },
    { "sbc  *,",	OP_3,0,	578,16 },
    { "sla  ",		OP_1,0,	288, 1 },
    { "sll  ",		OP_1,0,	304, 1 },	/* Unofficial */
    { "sra  ",		OP_1,0,	296, 1 },
    { "srl  ",		OP_1,0,	312, 1 },
    { "sub  ",		OP_1,0,	144, 1 },
    { "xor  ",		OP_1,0,	168, 1 }
};



static char  t[766][16];	/* 3 x 256 commands */


/*
 * Initialize
 *
 * This routine copies all predefined mnemonics, and then creates
 * itself LD r,r as well as BIT, SET and RES.
 * H, L, HL, and (HL) are coded with special characters for IX/IY.
 */

void initialize(test)
     int test;
{
    int  i, j, n;


    if (test)
	printf("Initializing mnemonic tables\n");


    for (i=0; i < N_CMDS1; i++)
	strcpy (t[data1[i].code], data1[i].mnem);

    for (i=0; i < N_CMDS2; i++) {
	n = data2[i].rgr;

	for (j=0; j <= id[n]; j++)
	    if (*regn[n][j]) {
		sprintf (t[data2[i].code + j * data2[i].b], "%s%s%s",
			 data2[i].mnem, regn[n][j], oper[(int)data2[i].post]);
	    }
    }


    /*  ld r,r  (except (hl),(hl)) */
    /*  If either i or j equals 6, H and L are not replaced with IX/IY. */

    for (i=0; i <= id[OP_1]; i++)
	for (j=0; j <= id[OP_1]; j++) {

	    if (i != 6 || j != 6) {
		sprintf (t[64 + i * 8 + j],
			 "ld   %s,%s", regn[OP_1][i], regn[OP_1][j]);
	    }
	}

    strcpy (t[0x66], "ld   h,(@)");
    strcpy (t[0x6e], "ld   l,(@)");
    strcpy (t[0x74], "ld   (@),h");
    strcpy (t[0x75], "ld   (@),l");


    /*  bit  res  set */

    for (i=0; i <= id[OP_8]; i++)
	for (j=0; j <= id[OP_1]; j++) {

	    n = 320 + i * 8 +j;
	    sprintf (t[n],      "bit  %s%s", regn[OP_8][i], regn[OP_1][j]);
	    sprintf (t[n + 64], "res  %s%s", regn[OP_8][i], regn[OP_1][j]);
	    sprintf (t[n +128], "set  %s%s", regn[OP_8][i], regn[OP_1][j]);
	}

    if (test)
	for (i=0; i <= 699; i++)
	    printf ("%03X\t%s\n", i, t[i]);
}



/* ------------------------------------------------------------------------- */

static char  txt[32];
FILE *infile;


/*
 * Disassemble and print one line.
 * If command following a prefix is not prefixable, the prefix is
 * printed as unknown command.
 *
 */

int  disassemble_line(counter, inbuf)
     int counter;
     BYTE *inbuf;
{
    char  buf[20];
    BYTE *p;
    int   i, len = 0;


    printf ("%04X\t", counter);


    len = sprint_instr(counter, inbuf, buf);

    for (i = 0, p = inbuf; i < len; ++i, ++p) {
	printf ("%02X ", *p & 0xff);
	txt[i] = (isprint(*p) ? *p : '.');
    }

    txt[len] = 0;

    if (len <3)
	printf ("\t");
    printf ("\t%s\t", buf);

    printf ( ((strlen(buf) < 8) ? "\t%s\n" : "%s\n"), txt);

    return (len);
}


/*
 * Disassemble one line
 */

int  sprint_instr(counter, inbuf, buf)
     int   counter;
     BYTE *inbuf;
     char *buf;
{
    char  tmp[20], *ke;
    BYTE  op, ce, xy;
    int   disp = 0;
    int   len = 0, done = 0, was_hl = 0;

    ce = xy = 0;


    /* Read Opcode */

    do {
	op = inbuf[len++];

	switch (op) {
	  case 0xdd: xy = 1; break;
	  case 0xfd: xy = 2; break;
	  case 0xcb: ce = 1; op = inbuf[len++]; ++done; break;
	  case 0xed: ce = 2; op = inbuf[len++]; ++done; break;
	  default:  done++;
	}
    } while (!done && len < 4);


    if (xy && ce) {
	disp = (int)op;		/* disp for XY+CE */
	op = inbuf[len++];	/* opcode */
    }


    /* Find Mnemonic */

    if (!*t[op + ce * 256]) {
	sprintf (buf, "???");
	return (1);		/* len */
    }

    strcpy (buf, t[op + ce * 256]);


    /* Build the Operand */

    while ((ke = strpbrk(buf, "%#$><*@")) != NULL) {
	switch (*ke) {			/* first special character */

	  case C_BYTE:
	    *(ke++) = 0;
	    sprintf (tmp, "%s%02X%s", buf, inbuf[len++], ke);
	    break;

	  case C_WORD:
	    {
		BYTE lobyte = inbuf[len++];

		*(ke++) = 0;
		sprintf (tmp, "%s%02X%02X%s", buf, inbuf[len++], lobyte, ke);
	    }
	    break;

	  case C_REL:		/*  relative branch */
	    disp = (int)inbuf[len++];

	    if (0x80 & disp)
		disp -= 256;
	    disp += (counter +2);

	    *(ke++) = 0;
	    sprintf (tmp, "%s%04X%s", buf, (disp & MAXADDR), ke);
	    break;


	    /* IX and IY:  HL -> IX (DD) or HL -> IY (FD) */
	    /* For the rest commands prefix is ignored. */

	  case C_HXY:
	    ++was_hl;
	    *(ke++) = 0;
	    sprintf (tmp, "%s%s%s", buf, hxy1[xy], ke);
	    break;

	  case C_LXY:
	    ++was_hl;
	    *(ke++) = 0;
	    sprintf (tmp, "%s%s%s", buf, hxy2[xy], ke);
	    break;

	  case C_HLXY:
	    ++was_hl;
	    *(ke++) = 0;
	    sprintf (tmp, "%s%s%s", buf, hxy3[xy], ke);
	    break;

	  case C_MHLXY:			/* signed disp */
	    ++was_hl;
	    *(ke++) = 0;

	    if (!xy)
		sprintf (tmp, "%s%s%s", buf, hxy3[0], ke);
	    else {
		/* If xy, do indexed with byte displacement */
		/* JP (HL) is not indirect */

		if (!xy || !ce)		/* We already have it for XY+CE */
		    disp = (int)inbuf[len++];

		sprintf (tmp, "%s%s+%02X%s", buf, hxy3[xy], disp, ke);
	    }
	}  /* switch */

	strcpy (buf, tmp);

    }  /* while (special) */


    /* For the rest commands prefix is ignored. */

    if (/*strict &&*/ xy && !was_hl) {
	sprintf (buf, "???");
	return (1);			 /* len (prefix only) */
    }

    return (len);
}


/* ------------------------------------------------------------------------- */

/*
 *  z 8 0 - a s s e m b l e r
 */

/*
 * This routine interprets Z80 assembly lines.
 * On success, places the instruction code in *args and returns the length.
 *
 */

#ifndef NO_ASM

int  parse_z80_instr(line, adr, memory, mode)
     char *line, *memory;
     ADDRESS adr;
     int mode;
{
    int   i, len = 0;
    char *p, *q;


    while (*line && isspace(*line))
	line++;

    if (!*line)
	return 0; /* line doesn't contain any mnemonic */


    /* Parse */

    for (i = 0; i < 760; i++) {
	p = t[i];
	q = line;

	if (!*p) continue;

	if (0 != (len = parse_arg(p, q, adr, i, (BYTE *)memory)))
	    return (len);
    }

    if (0 != (len = parse_arg("ld pc,*", line, adr, 233, (BYTE *)memory)))
	return (len);		/* LD PC,HL  is alias for JP (HL) */

    return (E_SYNTAX);
}


/*
 * Token Parser.
 * On success, builds the instruction code with possible prefix in *arg and
 * returns the length. Return value 0 means no match, whereas negative values
 * are error codes.
 */

static int  parse_arg(p, q, adr, code, arg)
     char *p, *q;
     ADDRESS adr;
     int code;
     BYTE *arg;
{
    int  len = 1, mode = 1, xy = 0;
    int  ival, disp, i;


	/*
	 * Since spesific addressing modes are not used, this routine parses
	 * possible parameters according to special characters in the search
	 * pattern.
	 */

	while (*p && *q) {

	    switch (*p) {			/* Do special characters */

	      case C_SPACE:
		if (*q && !isspace(*q))
		    return (0);  /* Fail */

		while (*(++p) == C_SPACE);
		while (isspace(*q)) ++q;
		break;

	      case C_BYTE:
		arg[1] = (BYTE) sconv(q, 0, mode);
		while (isxdigit(*q)) ++q;
		++len;
		++p;
		break;


	      case C_WORD:
		ival = (int)sconv(q, 0, mode);
		arg[1] = (BYTE) ival & 0xFF;
		arg[2] = (BYTE) (ival >> 8) & 0xFF;
		while (isxdigit(*q)) ++q;
		len = +2;
		++p;
		break;

	      case C_REL:
		ival = (int)sconv(q, 0, mode);
		while (isxdigit(*q)) ++q;
		++len;
		++p;
/*
		if (tstflg) {
		    printf("addressing mode RELATIVE\n");
		    printf("rel disp %d  %04x\n", ival - adr - 2, ival - adr - 2);
		}
*/
		if ((ival -= (adr + 2)) <= 127 && ival >= -128) {
		    arg[1] = (BYTE) ((ival) & 0xFF);
		    while (isxdigit(*q)) ++q;
		    return i;   /* len */
		}
		return E_LONG_BRANCH;
		break;

		/* Check HL, IX or IY */
		/* HL -> IX (DD) or IY (FD) */
	      case C_HXY:
		for (i = 0; i < 3; ++i) {
		    if (!strncmp(q, hxy1[i], strlen(hxy1[i]))) xy = i;
		    break;
		}
		if (i > 2)
		    return (E_BAD_INDEX);
		++p;
		break;

	      case C_LXY:
	      case C_HLXY:
		++p;
		break;


	      case C_MHLXY:
		;
		++p;
		break;

	      case '.':
		if (*q == '.' || *q == ',') { ++p; ++q; }
		break;

	      case '(':
	      case ')':  /* work out any permittable spaces */
	      case ',':
		while (isspace(*q)) ++q;
		if (*p == *q) { ++p; ++q; }
		while (isspace(*q)) ++q;
		break;

	      default:
		if (*p == tolower(*q)) {
		    ++p; ++q;
		    continue;
		}
		return (0);  /* Fail */
	    }
	}  /* while */


printf("%3d >%s< :: %d (%s)\n", code, p, len, q);
   if (!*p && (!*q || isspace(*q))) {

       if (xy) {
	   *arg++ = (xy == 1 ? 0xDD : 0xFD);
	   ++len;
       }
       if (code > 255) {
	       *arg++ = ((code >= 512) ? 0xED : 0xCB);
	       ++len;

	       if (xy) {  /* AND: (HL) !!! */
		   *arg++ = (BYTE)disp;
	           ++len;
	       }
       }

       *arg = (BYTE)code;
       return (len);
   }

    return (0);   /* Fail */

}
#endif  /* NO_ASM */


/* ------------------------------------------------------------------------- */

#ifdef STANDALONE_ZDIS

int main (argc, argv)
     int argc;
     char **argv;
{
    BYTE buf[6];
    int  counter = 0, bytes, left;
    int  debugflg=0;

    char *progname = argv[0];
    while ( --argc && '-' == (*++argv)[0]) {
 
	/* Flags */
 
	if (!strcmp (*argv, "-debug")) {
	    ++debugflg;
	    continue;
	}

	fprintf(stderr,
		"\n%s: Bad argument near '%s'.\n", progname, argv[0]);

    }  /* while */


    initialize(debugflg);


    if (!argc)
	infile = stdin;
    else
	if ((infile = fopen (argv[0], "rb")) == NULL) {
	    fprintf (stderr, "\n%s: Cannot open '%s'\n", progname, argv[0]);
	    exit (1);
	}


    bytes = left = 0;
    while ((left += fread(buf + left, 1, 4-left, infile)) > 0) {
	bytes = disassemble_line(counter, buf);
	counter += bytes;
	if ((left -= bytes) > 0)
	    memcpy(buf, buf + 4-left, left);
    }

    return 0;
}

#endif  /* STANDALONE_ZDIS */

